/*******************************************************************************

NAME            fpgaBbuPsDiag.c
SUMMARY         Zebulon fpga BBU and power supply function diagnostics
VERSION         %version: 1 %
UPDATE DATE     %date_modified: Dev 30 14:20 2009 %
PROGRAMMER      %created_by:    Jim Tu %

        Copyright 2009 Quanta Corporation. All Rights Reserved.

DESCRIPTION:
        This file has code to test BBU and power supply function of Zebulon fpga

NOTES:
		2009-12-30, Jim Tu: remove PS_CONTROL diagnostics

REFERENCE:

*******************************************************************************/
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/pci_regs.h>

#include <commTypes.h>
#include <commRpt.h>
#include <commMEC.h>
#include <commNvsramApi.h>
#include <commSysInfo.h>
#include <commIpcApi.h>
#include <commI2cApi.h>

#include "fpgaMidplaneI2cLib.h"
#include "fpgaRegLib.h"

#include "fpgaIntMaskDiag.h"
#include "fpgaEppDiag.h"
#include "fpgaResetDiag.h"
#include "fpgaBbuPsDiag.h"

/** EXTERN Function and Value ******************************/
EXTERN VOID 	fpgaHostCardReset(BOOLEAN status);
EXTERN VOID 	fpgaBoardReset(BOOLEAN status);
EXTERN VOID 	fpgaNvsramUnlockAll(VOID);
EXTERN VOID 	ctrlBootCountCompareSet(VOID);
EXTERN UINT8 	ctrlBootCountGet(VOID);
EXTERN UINT8 	ctrlBootCountCompareGet(VOID);

EXTERN INT32 	gManualDrive;	//global value to indicate the test is manual or automatic


int i2cEeprom64KBitRead(struct i2c_adapter *adap, uint16_t i2c_addr,
                     uint16_t offset, uint8_t *buf, int count)
{
    uint8_t __offset[2];
    struct i2c_msg msg[2];
    int ret;

    __offset[0] = (uint8_t) (offset >> 8); //MSB of word address
    __offset[1] = (uint8_t) (offset & 0xFF); //LSB of word address

    /* dummy write */
    msg[0].addr = i2c_addr;
    msg[0].flags = 0;
    msg[0].len = 2;
    msg[0].buf = __offset;

    /* read op */
    msg[1].addr = i2c_addr;
    msg[1].flags = I2C_M_RD;
    msg[1].len = count;
    msg[1].buf = buf;

    if ((ret = i2c_transfer(adap, msg, 2)) != 2)
        return (ret);

    return (0);
}


/****  BBU status  ****/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBatteryDrawingStatus
*
* SUMMARY:      verify if controller is powered by BBU
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from HOTSWAP_PGOOD bit
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE:	powered by BBU
*				FALSE:	powered by power supply
*/
INT32 fpgaBatteryDrawingStatus(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_3) & HOTSWAT_PGOOD_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBatteryInplace
*
* SUMMARY:      verify if BBU is attached
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from BBU_INPL bit
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE:	BBU is attached
*				FALSE:	BBU is not attached
*/
INT32 fpgaBatteryInplace(VOID)
{
	if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_3) & BBU_INPL_BIT) {
		return TRUE;
    	} else {
		return FALSE;
    	}
}


/****  BBU Control  ****/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuEnable
*
* SUMMARY:      enable/disable BBU
*
* SCOPE:        private
*
* DESCRIPTION:  enable/disable BBU
*
* RESTRICTIONS:
* 
* RETURNS:     
*/
VOID fpgaBbuEnable
 (
 BOOLEAN status
 )
{
    INT32 result;
    
    if(status)
	{	//turn on BBU
        result = fpgaRegisterWrite(FPGA_DISABLE_BATTERIES, 0x00);
        result = fpgaRegisterWrite(FPGA_ENABLE_BATTERIES, 0x01);
    }
	else
	{	//turn off BBU
        result = fpgaRegisterWrite(FPGA_ENABLE_BATTERIES, 0x00);
        result = fpgaRegisterWrite(FPGA_DISABLE_BATTERIES, 0x01);
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuDischarge
*
* SUMMARY:      enable/disable BBU discharge
*
* SCOPE:        private
*
* DESCRIPTION:  enable/disable BBU discharge
*
* RESTRICTIONS:
* 
* RETURNS:     
*/
VOID fpgaBbuDischarge
 (
 BOOLEAN status
 )
{
    INT32 result;
    UINT8 temp;
    
    temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_2);
    if(status)
	{	//start discharge BBU
        temp = temp | EN_DISCH_BIT;
    }
	else
	{	//stop discharge BBU
        temp = temp & (~EN_DISCH_BIT);
    }
	
	fpgaEncodedRegisterWrite();
    result = fpgaRegisterWrite(FPGA_PROTECTED_WRITE_2, temp);
}


/**** Power Supply Status ****/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrFailLatch
*
* SUMMARY:      verify if power had failed PWR_FAILn bit
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from 
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power had failed
*				FALSE:	power is good
*/
INT32 fpgaPwrFailLatch(VOID)
{
    if(fpgaRegisterRead(FPGA_LEGACY_MISCELLANEOUS_INFORMATION_1) & PWR_FAIL_BIT)
	{
        return FALSE;
    }
	else
	{
        return TRUE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaMainPwrFailLatch
*
* SUMMARY:      verify if power is failed
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from MAIN_PWR_FAIL bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is fail
*				FALSE:	power is good
*/
INT32 fpgaMainPwrFailLatch(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_1) & MAIN_PWR_FAIL_LATCH_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaMainPwrFail
*
* SUMMARY:      verify if power is failed
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from MAIN_PWR_FAIL bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is fail
*				FALSE:	power is good
*/
INT32 fpgaMainPwrFail(VOID)
{
    if(fpgaRegisterRead(FPGA_LEGACY_MISCELLANEOUS_INFORMATION_1) & MAIN_PWR_FAIL_BIT)
	{
        return FALSE;
    }
	else
	{
        return TRUE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPsAcGoodInt
*
* SUMMARY:      verify if AC line is good
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS_AC_GOOD_INT bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	AC power is good
*				FALSE:	AC power is not good
*/
INT32 fpgaPsAcGoodInt(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_1) & PS_AC_GOOD_INT_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPs0FaultInt
*
* SUMMARY:      verify if power supply 0 is fail
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS0_FAULT_INT bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is fail
*				FALSE:	power is good
*/
INT32 fpgaPs0FaultInt(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_2) & PS0_FAULT_INT_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPs1FaultInt
*
* SUMMARY:      verify if power supply 1 is fail
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS1_FAULT_INT bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is fail
*				FALSE:	power is good
*/
INT32 fpgaPs1FaultInt(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_2) & PS1_FAULT_INT_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPs0Inplace
*
* SUMMARY:      verify if power supply 0 is attached
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS0_INPL bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is attached
*				FALSE:	power is not attached
*/
INT32 fpgaPs0Inplace(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_3) & PS0_INPL_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPs1Inplace
*
* SUMMARY:      verify if power supply 1 is attached
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS1_INPL bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is attached
*				FALSE:	power is not attached
*/
INT32 fpgaPs1Inplace(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_3) & PS1_INPL_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPs0OverTemp
*
* SUMMARY:      verify if power supply 0 is over temp
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS0_OVER_TEMP bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is over temp
*				FALSE:	power is good
*/
INT32 fpgaPs0OverTemp(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_2) & PS0_OVER_TEMP_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPs1OverTemp
*
* SUMMARY:      verify if power supply 1 is over temp
*
* SCOPE:        private
*
* DESCRIPTION:  return the reading from PS1_OVER_TEMP bit
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	power is over temp
*				FALSE:	power is good
*/
INT32 fpgaPs1OverTemp(VOID)
{
    if(fpgaRegisterRead(FPGA_SUBSYSTEM_STATUS_LINES_2) & PS1_OVER_TEMP_BIT)
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/**** Power Supply Control ****/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaHostCardPsDisable
*
* SUMMARY:     	Disable host card power
*
* SCOPE:        private
*
* DESCRIPTION:  Disable host card power
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaHostCardPsDisable(VOID)
{
    UINT8 	temp;
    
    //set disbale bit
    temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_3);
    fpgaEncodedRegisterWrite();
    fpgaRegisterWrite(FPGA_PROTECTED_WRITE_3, temp | HOST_CARD_PS_DISABLE_BIT);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaHostCardPsEnable
*
* SUMMARY:     	Enable host card power
*
* SCOPE:        private
*
* DESCRIPTION:  Enable host card power
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaHostCardPsEnable(VOID)
{
    UINT8 	temp;

    //release disable bit
    temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_3);
    fpgaEncodedRegisterWrite();
    fpgaRegisterWrite(FPGA_PROTECTED_WRITE_3, temp & (~HOST_CARD_PS_DISABLE_BIT));
}


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrEngagement
*
* SUMMARY:     	switch on/off the power source
*
* SCOPE:        private
*
* DESCRIPTION:  switch the power source by the circuit on the midplane
*	            through I2C-GPIO
*
*				vccPs  : turn on the VCC12_BP from midplane 
*               vccBbu : turn on the VCC_BATT from midplane
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaPwrEngagement
 (
 BOOLEAN vccPs, 
 BOOLEAN vccBbu
 ) 
{
    //VCC12_BP switch
    if(vccPs)
	{
        midplaneGpioWrite(&VCC_PS_ENABLE, TRUE);  
    }
	else
	{
        midplaneGpioWrite(&VCC_PS_ENABLE, FALSE);  
    }
    
    //VCC_BATT switch
    if(vccBbu)
	{
        midplaneGpioWrite(&VCC_BBU_ENABLE, TRUE);  
    }
	else
	{
        midplaneGpioWrite(&VCC_BBU_ENABLE, FALSE); 
    }
}


VOID fpgaPsuCtl
 (
 BOOLEAN vccPs
 )
{
    //VCC12_BP switch
    if(vccPs)
    {
        midplaneGpioWrite(&VCC_PS_ENABLE, TRUE);
    }
    else
    {
        midplaneGpioWrite(&VCC_PS_ENABLE, FALSE);
    }
}


VOID fpgaBbuCtl
 (
 BOOLEAN vccBbu
 )
{
    //VCC_BATT switch
    if(vccBbu)
    {
        midplaneGpioWrite(&VCC_BBU_ENABLE, TRUE);
    }
    else
    {
        midplaneGpioWrite(&VCC_BBU_ENABLE, FALSE);
    }
}


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrAltEngagement
*
* SUMMARY:     	switch on/off the power source for alternate controller
*
* SCOPE:        private
*
* DESCRIPTION:  switch the power source of alternate controller by the 
*				circuit on the midplane through I2C-GPIO
*
*				vccPsAlt  : turn on the VCC12_BP of alternate controller
*							from midplane
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaPwrAltEngagement
 (
 BOOLEAN vccPsAlt
 )
{
	//VCC12_BP switch
    if(vccPsAlt){
        midplaneGpioWrite(&VCC_PS_ALT_ENABLE, TRUE);  
    }else{
        midplaneGpioWrite(&VCC_PS_ALT_ENABLE, FALSE);  
    }
}

#if 0
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPsControl
*
* SUMMARY:     	Enable to shut down power supply
*
* SCOPE:        private
*
* DESCRIPTION:  Enable to shut down power supply
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaPsControl
 (
 BOOLEAN status
 )
{
    UINT8 	temp;

    //release disable bit
    temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_1);
	if(status)
	{
		temp |= PS_CONTROL_BIT;
	}
	else
	{
		temp &= (~PS_CONTROL_BIT);
	}
    fpgaEncodedRegisterWrite();
    fpgaRegisterWrite(FPGA_PROTECTED_WRITE_1, temp);
}
#endif

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaShutOffPs
*
* SUMMARY:     	Enable shut off power through CPLD
*
* SCOPE:        private
*
* DESCRIPTION:  Enable shut off power through CPLD
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaShutOffPs
 (
 BOOLEAN status
 )
{
    UINT8 	temp;

    //release disable bit
    temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_2);
	if(status)
	{
		temp |= FPGA_SHUT_OFF_PS_BIT;
	}
	else
	{
		temp &= (~FPGA_SHUT_OFF_PS_BIT);
	}
    fpgaEncodedRegisterWrite();
    fpgaRegisterWrite(FPGA_PROTECTED_WRITE_2, temp);
}


/********************************/
/**  BBU Diagnostics Function  **/
/********************************/


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuInplaceDiag
*
* SUMMARY:     	verify BBU_INPLACE_L conntector
*
* SCOPE:        private
*
* DESCRIPTION:  verify BBU_INPLACE_L conntector
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaBbuInplaceDiag(VOID)
{
	INT32 result = TRUE;
	
	//for LSI midplane, do nothing
    
	//for Quanta midplane
	if(midplaneTypeVerification())
	{
//Modify for TestMidplane Changes
//		midplaneGpioWrite(&BBU_INPLACE_L, 1);  //emulate BBU has been removed
//		mdelay(10);
//		if(fpgaBatteryInplace()) //Pull low
//		{     
//			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_INPLACE_ERROR,
//							MSG_FPGA_BBU_INPLACE_ERROR, 0, 0, 0, 0);
//			printk("[FPGA BBU&PS] BBU Remove.\n");
//			result = FALSE;
//		}
	
//		midplaneGpioWrite(&BBU_INPLACE_L, 0);  //emulate BBU is attached
//		mdelay(10);
		if(!fpgaBatteryInplace())	//Pull high
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_INPLACE_ERROR,
							MSG_FPGA_BBU_INPLACE_ERROR, 0, 0, 0, 0);
			printk("[FPGA BBU&PS] BBU Inplace.\n");
			result = FALSE;
		}
	
        printk("[FPGA BBU&PS] BBU Inplace Test Done.\n");
	}
    
	return result;
}

VOID fpgaBbuInplace(INT32 status)
{
	if(status)
	{
		midplaneGpioWrite(&BBU_INPLACE_L, 0);
	}
//	else
//	{
//		midplaneGpioWrite(&BBU_INPLACE_L, 1);
//	}
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbu12vDiag
*
* SUMMARY:     	verify BBU 12V charging pin 
*
* SCOPE:        private
*
* DESCRIPTION:  verify BBU 12V charging pin
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaBbu12vDiag(VOID)
{
	INT32 result = TRUE;
	
	//for LSI midplane, do nothing
    
	//for Quanta midplane
	if(midplaneTypeVerification())
	{
		if(!midplaneGpioRead(&VCC12_INPUT))
		{
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_VCC12_ERROR,
							MSG_FPGA_BBU_VCC12_ERROR, 0, 0, 0, 0);
			result = FALSE;
		} 
		
		printk("[FPGA BBU&PS] BBU VCC12_INPUT Test Done.\n");
    }
	return result;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuDischargeDiag
*
* SUMMARY:     	verify BBU discharge pin
*
* SCOPE:        private
*
* DESCRIPTION:  verify BBU discharge pin
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaBbuDischargeDiag(VOID)
{
	INT32 result = TRUE;

	//for LSI midplane, do nothing
    
	//for Quanta midplane
	if(midplaneTypeVerification())
	{
		//enable BBU discharge function
		fpgaBbuDischarge(TRUE);
		
		if(!midplaneGpioRead(&BBU_DISCHARGE))
		{
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_DISCHARGE_ERROR,
							MSG_FPGA_BBU_DISCHARGE_ERROR, 0, 0, 0, 0);
			result = FALSE;
			printk("[FPGA BBU&PS] BBU Discharge On Fail. \n");
		} 
		else 
		{
			//printk("[FPGA BBU&PS] BBU Discharge On Good. \n");
		}
		
		//disable BBU discharge function
		fpgaBbuDischarge(FALSE);
		
		if(midplaneGpioRead(&BBU_DISCHARGE))
		{
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_DISCHARGE_ERROR,
							MSG_FPGA_BBU_DISCHARGE_ERROR, 0, 0, 0, 0);
			result = FALSE;
			printk("[FPGA BBU&PS] BBU Discharge Off Fail. \n");
		} 
		else 
		{
			//tprintk("[FPGA BBU&PS] BBU Discharge Off Good. \n");
		}
    }
	return result;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuEnableDiag
*
* SUMMARY:     	verify BBU enable pin
*
* SCOPE:        private
*
* DESCRIPTION:  verify BBU enable pin
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaBbuEnableDiag(VOID)
{
	INT32 result = TRUE;

	if(midplaneTypeVerification()){
		//enable BBU function
		fpgaBbuEnable(TRUE);
		
		if(!midplaneGpioRead(&BBU_ENABLE))
		{
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_ENABLE_ERROR,
							MSG_FPGA_BBU_ENABLE_ERROR, 0, 0, 0, 0);
			result = FALSE;
			printk("[FPGA BBU&PS] BBU Enable Fail. \n");
		} 
		else 
		{
			printk("[FPGA BBU&PS] BBU Enable Good. \n");
		}
		
		//disable BBU function
		fpgaBbuEnable(FALSE);
		
		if(midplaneGpioRead(&BBU_ENABLE))
		{
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_ENABLE_ERROR,
							MSG_FPGA_BBU_ENABLE_ERROR, 0, 0, 0, 0);
			result = FALSE;
			printk("[FPGA BBU&PS] BBU Disable Fail. \n");
		} 
		else 
		{
			printk("[FPGA BBU&PS] BBU Disable Good. \n");
		}
    }
	
	return result;
}

/****************************************************************************************/

/*****************************************/
/**  Power Supply Diagnostics Function  **/
/*****************************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrAltColdReboot
*
* SUMMARY:     	power off the peer controller and power up again
*
* SCOPE:        private
*
* DESCRIPTION:  cut of the alternate controller power supply and resume again
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaPwrAltColdReboot(VOID)
{
	printk("\n");
	printk("\n");
	printk("== Controller Power Cycle Request Received!! (through IPC Protocol) == \n");
	printk("\n");
	fpgaPwrAltEngagement(FALSE);	//shut down peer controller
	
	mdelay(10000);
	
	fpgaPwrAltEngagement(TRUE);		//power up peer controller
}
EXPORT_SYMBOL(fpgaPwrAltColdReboot);


VOID controllerPowerCycle(VOID)
{
	IPC_MSG msg;

	ipcEthernetRegister(IPC_APP_REQUEST_POWER_CYCLE, NULL);

	/* send request */
	msg.hdr.app = IPC_APP_REQUEST_POWER_CYCLE;
	msg.hdr.flag = IPC_APP_REQUEST;
	msg.hdr.len = 0;
	ipcEthernetSend((EIPC_MSG*) &msg);

	ipcEthernetDeregister(IPC_APP_REQUEST_POWER_CYCLE);
	
	printk("\n");
	printk("== Controller Power Cycle Request Send!! (through IPC Protocol) == \n");
	printk("\n");
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrDiagLsi
*
* SUMMARY:     	power supply diagnostics with LSI midplane
*
* SCOPE:        private
*
* DESCRIPTION:  power supply diagnostics with LSI midplane
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaPwrDiagLsi(VOID)
{
    INT32 result = TRUE;
    
	/***   Verify default pin status with LSI midplane   ***/
    if(fpgaPwrFailLatch())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_PWR_FAIL_LATCH_ERROR,
						MSG_FPGA_PS_PWR_FAIL_LATCH_ERROR, 0, 0, 0, 0);
        result = FALSE;
    }  
    
    if(fpgaMainPwrFail())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_MAIN_PWR_FAIL_ERROR,
						MSG_FPGA_PS_MAIN_PWR_FAIL_ERROR, 0, 0, 0, 0);
        result = FALSE;
    }
    
	if(fpgaPsAcGoodInt())
	{     
		commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_AC_GOOD_ERROR,
            	        MSG_FPGA_PS_AC_GOOD_ERROR, 0, 0, 0, 0);
		result = FALSE;
	}
	
	if(fpgaPs0OverTemp())
	{
		commReportError(MEC_FPGA, FPGA_MISCELLANEOUS_MEC, FPGA_PS0_OVER_TEMP_ERROR,
						MSG_FPGA_PS0_OVER_TEMP_ERROR, 0, 0, 0, 0);
		result = FALSE;
	}
	if(fpgaPs1OverTemp())
	{
		commReportError(MEC_FPGA, FPGA_MISCELLANEOUS_MEC, FPGA_PS1_OVER_TEMP_ERROR,
						MSG_FPGA_PS1_OVER_TEMP_ERROR, 0, 0, 0, 0);
		result = FALSE;
	}
    
    /***   Verify PWR_FLT_IN_L with LSI midplane   ***/
	
#if 0	//for LSI firmware
    fpgaPsuI2cTest(0, 0xFC);    //emulate power supply 0 failes    
    if(!fpgaPs0FaultInt())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_FAULT_ERROR, 
						MSG_FPGA_PS_0_FAULT_ERROR, 0, 0, 0, 0);
        result = FALSE;
    } 
    if(fpgaPs1FaultInt())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_FAULT_ERROR,
						MSG_FPGA_PS_1_FAULT_ERROR, 0, 0, 0, 0);
        result = FALSE;
    } 
    
    fpgaPsuI2cTest(0, 0xFF);
    fpgaPsuI2cTest(1, 0xFC);    //emulate power supply 1 failes
    if(fpgaPs0FaultInt())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_FAULT_ERROR, 
						MSG_FPGA_PS_0_FAULT_ERROR, 0, 0, 0, 0);
        result = FALSE;
    } 
    if(!fpgaPs1FaultInt())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_FAULT_ERROR,
						MSG_FPGA_PS_1_FAULT_ERROR, 0, 0, 0, 0);
        result = FALSE;
    }  
        
    fpgaPsuI2cTest(0, 0xFC);	//emulate both power back fail
    if(!fpgaPsAcGoodInt())
	{     
        commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_AC_GOOD_ERROR,
						MSG_FPGA_PS_AC_GOOD_ERROR, 0, 0, 0, 0);
        result = FALSE;
    }
	
	//let both power supplies back to normal
    fpgaPsuI2cTest(0, 0xFF);			
    fpgaPsuI2cTest(1, 0xFF);
#endif

	printk("[FPGA BBU&PS] PWR Test Done. (LSI)\n");
    
    return result;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrMonitorDiag
*
* SUMMARY:     	power supply diagnostics
*
* SCOPE:        private
*
* DESCRIPTION:  power supply diagnostics
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaPwrMonitorDiag(VOID)
{
    INT32 	result = TRUE;
	INT32 	powerCycle = 0;
	VOID* 	diagSeqRecord;
	UINT16 	bbuCurrent[8];
     
    if(!midplaneTypeVerification())
	{     
		//LSI midplane, run fpgaPwrDiagLsi()
		result = fpgaPwrDiagLsi(); 
		
    } 
	else if(sysMidplaneType != TEST_MIDPLANE_WO_BBU_CABLE) 
	{
		/***  BBU Cable is connected  ***/ 	
		printk("[FPGA BBU&PS] Power & BBU Switch Diag Start......\n");
		
		/****************/
		/* Set Syndrome */
		/****************/
		fpgaNvsramUnlockAll();
		
		for(powerCycle = 0; powerCycle < 5; powerCycle++)
		{
	
			diagSeqRecord = (gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_STATUS_LOCATION);

			//verify if this the first time
			if( (*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_ENCODE_UINT8_0_LOCATION) 
																		!= BBU_SWITCH_ENCODE_PATTERN_0) ||    
 				(*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_ENCODE_UINT8_1_LOCATION) 
																		!= BBU_SWITCH_ENCODE_PATTERN_1) )
			{
				//set checkpoint
				*(volatile UINT8*)diagSeqRecord = BBU_SWITCH_STATUS_NORMAL;
		
				//set syndrome
				*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_ENCODE_UINT8_0_LOCATION) 
																		= BBU_SWITCH_ENCODE_PATTERN_0;    
				*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_ENCODE_UINT8_1_LOCATION) 
																		= BBU_SWITCH_ENCODE_PATTERN_1;	
			}
		
			/***************************/
			/* Run BBU/PS Switch Diag. */
			/***************************/
			if(*(volatile UINT8*)diagSeqRecord == BBU_SWITCH_STATUS_NORMAL)
			{
				*(volatile UINT8*)diagSeqRecord = BBU_SWITCH_STATUS_TEST;
			
				ctrlBootCountCompareSet();
				
				//printk("[FPGA BBU&PS] MIS reg: 0x%02X\n", fpgaRegisterRead(0x24));
				//printk("[FPGA BBU&PS] Sub reg: 0x%02X\n", fpgaRegisterRead(0x64));
				
				//clear PWR_FAILn latch
				fpgaRegisterWrite(FPGA_CLEAR_POWER_FAIL_LATCH, 0x01);	

				//clean main power fail latch
				fpgaRegisterWrite(FPGA_CLEAR_MAIN_POWER_FAIL_LATCH, 0x01);

				//verify PWR_FAIL 
				if(fpgaPwrFailLatch())
				{     
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_CLEAR_PWR_FAIL_LATCH_ERROR,
									MSG_FPGA_PS_CLEAR_PWR_FAIL_LATCH_ERROR, 0, 0, 0, 0);
					result = FALSE;
				}
				
				//verify MAIN_PWR_FAIL latch
				if(fpgaMainPwrFailLatch())
				{     
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_CLEAR_MAIN_PWR_FAIL_LATCH_ERROR,
									MSG_FPGA_PS_CLEAR_MAIN_PWR_FAIL_LATCH_ERROR, 0, 0, 0, 0);
					result = FALSE;
				}
				
				//verify MAIN_PWR_FAIL 
				if(fpgaMainPwrFail())
				{     
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_MAIN_PWR_FAIL_ERROR,
									MSG_FPGA_PS_MAIN_PWR_FAIL_ERROR, 0, 0, 0, 0);
					result = FALSE;
				}
				
				//printk("[FPGA BBU&PS] MIS reg: 0x%02X\n", fpgaRegisterRead(0x24));
				//printk("[FPGA BBU&PS] Sub reg: 0x%02X\n", fpgaRegisterRead(0x64));
				
				/***   Verify Power Supply Monitor with Power Supplies Are Off   ***/
				//emulate power supplies off
				printk("[FPGA BBU&PS] Turn off AC\n");
				fpgaPsuCtl( TRUE );
				fpgaBbuCtl( TRUE );
				mdelay( 50 );

				fpgaPsuCtl( FALSE );
                mdelay( 50 );

				//verify MAIN_PWR_FAIL latch
				if(!fpgaMainPwrFailLatch())
				{     
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_MAIN_PWR_FAIL_LATCH_ERROR,
									MSG_FPGA_PS_MAIN_PWR_FAIL_LATCH_ERROR, 0, 0, 0, 0);
					result = FALSE;
				}

//LSI FPGA issue
#if 0				
				if(!fpgaMainPwrFail())
				{     
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_MAIN_PWR_FAIL_ERROR,
									MSG_FPGA_PS_MAIN_PWR_FAIL_ERROR, 0, 0, 0, 0);
					result = FALSE;
				}
#endif			
				//printk("[FPGA BBU&PS] MIS reg: 0x%02X\n", fpgaRegisterRead(0x24));
				//printk("[FPGA BBU&PS] Sub reg: 0x%02X\n", fpgaRegisterRead(0x64));

				/***************************/
				/* Verify BBU BATT current */
				/***************************/
			
				//dummy read
				bbuCurrent[0] = ina209CurrentMonitor(MID_INA209_1_ADDRESS);	
				bbuCurrent[1] = ina209CurrentMonitor(MID_INA209_2_ADDRESS);
				bbuCurrent[2] = ina209CurrentMonitor(MID_INA209_3_ADDRESS);
				bbuCurrent[3] = ina209CurrentMonitor(MID_INA209_4_ADDRESS);
				bbuCurrent[4] = ina209CurrentMonitor(MID_INA209_5_ADDRESS);
				bbuCurrent[5] = ina209CurrentMonitor(MID_INA209_6_ADDRESS);
				bbuCurrent[6] = ina209CurrentMonitor(MID_INA209_7_ADDRESS);
				bbuCurrent[7] = ina209CurrentMonitor(MID_INA209_8_ADDRESS);
				
				//printk("[FPGA BBU&PS] (Supporting Dummy). Current: 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X \n",
				//		bbuCurrent[0], bbuCurrent[1], bbuCurrent[2], bbuCurrent[3], bbuCurrent[4], bbuCurrent[5], bbuCurrent[6], bbuCurrent[7]);
				
				mdelay(20);
				
				//read BBU supporting current
				bbuCurrent[0] = (0x10000 - ina209CurrentMonitor(MID_INA209_1_ADDRESS)) & MID_CURRENT_MASK;	
				bbuCurrent[1] = (0x10000 - ina209CurrentMonitor(MID_INA209_2_ADDRESS)) & MID_CURRENT_MASK;
				bbuCurrent[2] = (0x10000 - ina209CurrentMonitor(MID_INA209_3_ADDRESS)) & MID_CURRENT_MASK;
				bbuCurrent[3] = (0x10000 - ina209CurrentMonitor(MID_INA209_4_ADDRESS)) & MID_CURRENT_MASK;
				bbuCurrent[4] = (0x10000 - ina209CurrentMonitor(MID_INA209_5_ADDRESS)) & MID_CURRENT_MASK;
				bbuCurrent[5] = (0x10000 - ina209CurrentMonitor(MID_INA209_6_ADDRESS)) & MID_CURRENT_MASK;
				bbuCurrent[6] = (0x10000 - ina209CurrentMonitor(MID_INA209_7_ADDRESS)) & MID_CURRENT_MASK;
				bbuCurrent[7] = (0x10000 - ina209CurrentMonitor(MID_INA209_7_ADDRESS)) & MID_CURRENT_MASK;
				
				//verify each BBU pin current
				if((bbuCurrent[0] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[0] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_1 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[0]);
					result = FALSE;
				}
			
				if((bbuCurrent[1] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[1] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_2 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[1]);
					result = FALSE;
				}
			
				if((bbuCurrent[2] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[2] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_3 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[2]);
					result = FALSE;
				}
		
				if((bbuCurrent[3] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[3] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_4 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[3]);
					result = FALSE;
				}
		
				if((bbuCurrent[4] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[4] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_5 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[4]);
					result = FALSE;
				}
				
				if((bbuCurrent[5] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[5] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_6 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[5]);
					result = FALSE;
				}
		
				if((bbuCurrent[6] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[6] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_7 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[6]);
					result = FALSE;
				}
				
				if((bbuCurrent[7] < MID_NORMAL_CURRENT_LB) || (bbuCurrent[7] > MID_NORMAL_CURRENT_UB))
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_8 Failed. (Supporting) Current: 0x%04X \n", bbuCurrent[7]);
					result = FALSE;
				}
			
				printk("[FPGA BBU&PS] (Supporting). Current: 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X \n",
						bbuCurrent[0], bbuCurrent[1], bbuCurrent[2], bbuCurrent[3], bbuCurrent[4], bbuCurrent[5], bbuCurrent[6], bbuCurrent[7]);


				//switch VCC12_BP back to narmal
				printk("[FPGA BBU&PS] Turn on AC\n");
				fpgaPsuCtl( TRUE );
				mdelay(50);
				fpgaBbuCtl( FALSE );
				mdelay(50);

   		
				/*************************************/
				/* Some boards could be reset here!! */
				/*************************************/
				//printk("[FPGA BBU&PS] MIS reg: 0x%02X\n", fpgaRegisterRead(0x24));
				//printk("[FPGA BBU&PS] Sub reg: 0x%02X\n", fpgaRegisterRead(0x64));
				//verify MAIN_PWR_FAIL
				if(fpgaMainPwrFail())
				{     
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_MAIN_PWR_FAIL_ERROR,
									MSG_FPGA_PS_MAIN_PWR_FAIL_ERROR, 0, 0, 0, 0);
					result = FALSE;
				}
	
				/***************************/
				/* Verify BBU BATT current */
				/***************************/
					
				//dummy read
				bbuCurrent[0] = ina209CurrentMonitor(MID_INA209_1_ADDRESS);
				bbuCurrent[1] = ina209CurrentMonitor(MID_INA209_2_ADDRESS);
				bbuCurrent[2] = ina209CurrentMonitor(MID_INA209_3_ADDRESS);
				bbuCurrent[3] = ina209CurrentMonitor(MID_INA209_4_ADDRESS);	
				bbuCurrent[4] = ina209CurrentMonitor(MID_INA209_5_ADDRESS);
				bbuCurrent[5] = ina209CurrentMonitor(MID_INA209_6_ADDRESS);
				bbuCurrent[6] = ina209CurrentMonitor(MID_INA209_7_ADDRESS);
				bbuCurrent[7] = ina209CurrentMonitor(MID_INA209_8_ADDRESS);
				
				//printk("[FPGA BBU&PS] (Normal dummy). Current: 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X \n",
				//		bbuCurrent[0], bbuCurrent[1], bbuCurrent[2], bbuCurrent[3], bbuCurrent[4], bbuCurrent[5], bbuCurrent[6], bbuCurrent[7]);
				
				mdelay(20);
				
				//read BBU supporting current
				bbuCurrent[0] = ina209CurrentMonitor(MID_INA209_1_ADDRESS) & MID_CURRENT_MASK;
				bbuCurrent[1] = ina209CurrentMonitor(MID_INA209_2_ADDRESS) & MID_CURRENT_MASK;
				bbuCurrent[2] = ina209CurrentMonitor(MID_INA209_3_ADDRESS) & MID_CURRENT_MASK;
				bbuCurrent[3] = ina209CurrentMonitor(MID_INA209_4_ADDRESS) & MID_CURRENT_MASK;	
				bbuCurrent[4] = ina209CurrentMonitor(MID_INA209_5_ADDRESS) & MID_CURRENT_MASK;
				bbuCurrent[5] = ina209CurrentMonitor(MID_INA209_6_ADDRESS) & MID_CURRENT_MASK;
				bbuCurrent[6] = ina209CurrentMonitor(MID_INA209_7_ADDRESS) & MID_CURRENT_MASK;	
				bbuCurrent[7] = ina209CurrentMonitor(MID_INA209_8_ADDRESS) & MID_CURRENT_MASK;

			
				//verify each BBU pin current
				if(bbuCurrent[0] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_1 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[0]);
					result = FALSE;
				}
		
				if(bbuCurrent[1] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_2 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[1]);
					result = FALSE;
				}
			
				if(bbuCurrent[2] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_3 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[2]);
					result = FALSE;
				}
		
				if(bbuCurrent[3] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_4 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[3]);
					result = FALSE;
				}

				if(bbuCurrent[4] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_5 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[4]);
					result = FALSE;
				}
			
				if(bbuCurrent[5] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_6 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[5]);
					result = FALSE;
				}

				if(bbuCurrent[6] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_7 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[6]);
					result = FALSE;
				}
				
				if(bbuCurrent[7] > MID_OFF_CURRENT_UB)
				{
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
					printk("[FPGA BBU&PS] BBU BATT_PIN_8 Failed (Noraml). Current: 0x%04X \n", bbuCurrent[7]);
					result = FALSE;
				}
			
				printk("[FPGA BBU&PS] (Normal). Current: 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X \n",
						bbuCurrent[0], bbuCurrent[1], bbuCurrent[2], bbuCurrent[3], bbuCurrent[4], bbuCurrent[5], bbuCurrent[6], bbuCurrent[7]);
			
				fpgaPsuCtl( TRUE );
                fpgaBbuCtl( TRUE );
			}


			/***************************************/
			/* Verify the power fail latch cleaner */
			/***************************************/
			
			if(!fpgaMainPwrFailLatch())
			{		//the latch should be active     
				commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_MAIN_PWR_FAIL_LATCH_ERROR,
								MSG_FPGA_PS_MAIN_PWR_FAIL_LATCH_ERROR, 0, 0, 0, 0);
				result = FALSE;
			}
			//printk("[FPGA BBU&PS] MIS reg: 0x%02X\n", fpgaRegisterRead(0x24));
			//printk("[FPGA BBU&PS] Sub reg: 0x%02X\n", fpgaRegisterRead(0x64));

			//clean main power fail latch
			fpgaRegisterWrite(FPGA_CLEAR_MAIN_POWER_FAIL_LATCH, 0x01);	
			
			if(fpgaMainPwrFailLatch())
			{	//the latch should be cleaned
				commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_CLEAR_MAIN_PWR_FAIL_LATCH_ERROR,
								MSG_FPGA_PS_CLEAR_MAIN_PWR_FAIL_LATCH_ERROR, 0, 0, 0, 0);
				result = FALSE;
			}
			//printk("[FPGA BBU&PS] MIS reg: 0x%02X\n", fpgaRegisterRead(0x24));
			//printk("[FPGA BBU&PS] Sub reg: 0x%02X\n", fpgaRegisterRead(0x64));

			/***************************/
			/* Verify the power switch */
			/***************************/
			if(*(volatile UINT8*)diagSeqRecord == BBU_SWITCH_STATUS_TEST)
			{
				*(volatile UINT8*)diagSeqRecord = BBU_SWITCH_STATUS_NORMAL;
			
				/* Should not be reboot */
			
				if(ctrlBootCountGet() == ctrlBootCountCompareGet())
				{	//boot count doesn't increase.
					printk("[FPGA BBU&PS] Power: BBU -> PS: Good. \n");
				} 
				else 
				{	//if controller is reboot, the boot count increase.
					commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_POWER_SWITCH_ERROR, MSG_FPGA_BBU_POWER_SWITCH_ERROR, 0, 0, 0, 0);
				}
			}
		
			//clean syndrome and checkpoint
			*(volatile UINT8*)diagSeqRecord = BBU_SWITCH_STATUS_NORMAL;
			*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_ENCODE_UINT8_0_LOCATION) = 0;    
			*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + BBU_SWITCH_ENCODE_UINT8_1_LOCATION) = 0;
		}
#if 0
		//turn on host card
		fpgaHostCardPsEnable();     
		mdelay(300);				//delay for asserting HOST_PWR_GOOD (150ms at least)
		fpgaHostCardReset(FALSE);MAINBOARD_DIAG
#endif	
		printk("[FPGA BBU&PS] Power & BBU Switch Diag Done.\n");

		/*******************************/
		/***   Verify PWR_FLT_IN_L   ***/
		/*******************************/
    
		midplaneGpioWrite(&PS0_FAULT_L, 0);    //emulate power supply 0 failes
		midplaneGpioWrite(&PS1_FAULT_L, 1);    
		if(!fpgaPs0FaultInt())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_FAULT_ERROR, 
							MSG_FPGA_PS_0_FAULT_ERROR, 0, 0, 0, 0);
			result = FALSE;
		} 
		if(fpgaPs1FaultInt())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_FAULT_ERROR,
							MSG_FPGA_PS_1_FAULT_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}    
    
		midplaneGpioWrite(&PS0_FAULT_L, 1);
		midplaneGpioWrite(&PS1_FAULT_L, 0);    //emulate power supply 1 failes
		if(fpgaPs0FaultInt())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_FAULT_ERROR, 
							MSG_FPGA_PS_0_FAULT_ERROR, 0, 0, 0, 0);
			result = FALSE;
		} 
		if(!fpgaPs1FaultInt())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_FAULT_ERROR,
							MSG_FPGA_PS_1_FAULT_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}    
        
		midplaneGpioWrite(&PS0_FAULT_L, 1);
		midplaneGpioWrite(&PS1_FAULT_L, 1);
		
		printk("[FPGA BBU&PS] Power Fault Diag Done.\n");
		
		/**************************/
		/***   Verify AC_GOOD   ***/
		/**************************/
		
		midplaneGpioWrite(&AC_GOOD, 1);
		if(fpgaPsAcGoodInt())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_AC_GOOD_ERROR,
            	        	MSG_FPGA_PS_AC_GOOD_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		midplaneGpioWrite(&AC_GOOD, 0);
		if(!fpgaPsAcGoodInt())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_AC_GOOD_ERROR,
            	        	MSG_FPGA_PS_AC_GOOD_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
			
		midplaneGpioWrite(&AC_GOOD, 1);
		printk("[FPGA BBU&PS] AC GOOD Diag Done.\n");
		
#if 0	//Modified by Jim Tu, 2009-12-30
		//Unused in FPGA REV.08
		/*****************************/
		/***   Verify PS_CONTROL   ***/
		/*****************************/
		fpgaPsControl(TRUE);
		if(midplaneGpioRead(&PS0_CONTROL_L_BP1))
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_CONTROL_ERROR,
            	        	MSG_FPGA_PS_CONTROL_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		fpgaPsControl(FALSE);
		if(!midplaneGpioRead(&PS0_CONTROL_L_BP1))
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_CONTROL_ERROR,
            	        	MSG_FPGA_PS_CONTROL_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		printk("[FPGA BBU&PS] PS CONTROL Diag Done.\n");
#endif		
		/*******************************/
		/***   Verify PS_OVER_TEMP   ***/
		/*******************************/
		
		if(fpgaPs0OverTemp())
		{
			commReportError(MEC_FPGA, FPGA_MISCELLANEOUS_MEC, FPGA_PS0_OVER_TEMP_ERROR,
							MSG_FPGA_PS0_OVER_TEMP_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
			
		if(fpgaPs1OverTemp())
		{
			commReportError(MEC_FPGA, FPGA_MISCELLANEOUS_MEC, FPGA_PS1_OVER_TEMP_ERROR,
							MSG_FPGA_PS1_OVER_TEMP_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
    
		printk("[FPGA BBU&PS] Power Over Temp Diag Done.\n");
    } 
	else 
	{
		/***  BBU cable is not connected.  ***/
		printk("[FPGA BBU&PS] BBU cable is not connected.\n");
		commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_BBU_PIN_ERROR, MSG_FPGA_BBU_PIN_ERROR, 0, 0, 0, 0);
	}
	
	return result;  
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaPwrInplaceDiag
*
* SUMMARY:     	verify power supply inplace pin
*
* SCOPE:        private
*
* DESCRIPTION:  verify power supply inplace pin
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaPwrInplaceDiag(VOID)
{
	INT32 result = TRUE;
    
	if(!midplaneTypeVerification())
	{
		//LSI midplane
		//verify Power Inplace, default power supplies are attached
		if(!fpgaPs0Inplace()) {     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_INPLACE_ERROR, 
							MSG_FPGA_PS_0_INPLACE_ERROR, 0, 0, 0, 0);
            		result = FALSE;
    		} 
    		if(!fpgaPs1Inplace()) {     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_INPLACE_ERROR,
							MSG_FPGA_PS_1_INPLACE_ERROR, 0, 0, 0, 0);
            		result = FALSE;
    		}
    	} else {
		//Quanta test midplane
		
		midplaneGpioWrite(&PWR_INPL_0_L, 1);   //emulate power supply 0 is not attached
		midplaneGpioWrite(&PWR_INPL_1_L, 1);   //emulate power supply 1 is not attached
    
		//verify Power Inplace 
		if(fpgaPs0Inplace())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_INPLACE_ERROR, 
							MSG_FPGA_PS_0_INPLACE_ERROR, 0, 0, 0, 0);
			result = FALSE;
		} 
		if(fpgaPs1Inplace())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_INPLACE_ERROR,
							MSG_FPGA_PS_1_INPLACE_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}   
    
		midplaneGpioWrite(&PWR_INPL_0_L, 0);   //emulate power supply 0 is attached
		midplaneGpioWrite(&PWR_INPL_1_L, 0);   //emulate power supply 1 is attached
		
		//verify Power Inplace 
		if(!fpgaPs0Inplace())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_0_INPLACE_ERROR, 
							MSG_FPGA_PS_0_INPLACE_ERROR, 0, 0, 0, 0);
			result = FALSE;
		} 
		if(!fpgaPs1Inplace())
		{     
			commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_1_INPLACE_ERROR,
							MSG_FPGA_PS_1_INPLACE_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}  
		
		midplaneGpioWrite(&PWR_INPL_0_L, 1);
		midplaneGpioWrite(&PWR_INPL_1_L, 1);
		
		printk("[FPGA BBU&PS] Power Inplace Diag Done.\n");
	}
	
    return result;
}

/*********************************************************************************************/

/*********************************/
/**  Main Diagnostics Function  **/
/*********************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuPsDiag
*
* SUMMARY:     	verify power supply and BBU functions
*
* SCOPE:        public
*
* DESCRIPTION:  verify power supply and BBU functions
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaBbuPsDiag(VOID)
{   
	//fpgaMaskIntAll(TRUE);     //mask possible INT32errupts
	if( fpgaPwrMonitorDiag() & fpgaPwrInplaceDiag() & 
		fpgaBbuInplaceDiag() & fpgaBbu12vDiag() & 
		fpgaBbuDischargeDiag() & fpgaBbuEnableDiag() )
	{
		fpgaClearSubsystemInt();
		//fpgaMaskIntAll(FALSE);    //release masks
		return TRUE;
	}
	else
	{
		fpgaClearSubsystemInt();
		//fpgaMaskIntAll(FALSE);    //release masks
		return FALSE;
	}
}

/**********************************************************************************************/

/********************************************/
/**  Host Card Power Diagnostics Function  **/
/********************************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaHostCardPsDisableDiag
*
* SUMMARY:     	verify host card low power mode functions
*
* SCOPE:        private
*
* DESCRIPTION:  verify host card low power mode functions
*
* RESTRICTIONS:
* 
* RETURNS:     	TRUE:	Pass
*				FALSE:	Fail
*/
INT32 fpgaHostCardPsDisableDiag(VOID)
{
	INT32 			result = TRUE;
	VOID* 			diagSeqRecord;
	struct i2c_adapter *adap = NULL;
	uint8_t 		data;

	if( !adap )
		return FALSE;
	
	fpgaNvsramUnlockAll();
	
	diagSeqRecord = (gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_DIAG_SEQ_RECORD_OFFSET);

	//verify if this the first time in host card low power test
	if( (*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_0_LOCATION) != LOW_RESET_ENCODED_PATTERN_0) ||    
 		(*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_1_LOCATION) != LOW_RESET_ENCODED_PATTERN_1) ||    
		(*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_2_LOCATION) != LOW_RESET_ENCODED_PATTERN_2) )
	{
		//set checkpoINT32
		*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_NORMAL;
	
		//set syndrome
		*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_0_LOCATION) = LOW_RESET_ENCODED_PATTERN_0;    
		*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_1_LOCATION) = LOW_RESET_ENCODED_PATTERN_1;    
		*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_2_LOCATION) = LOW_RESET_ENCODED_PATTERN_2;	
	}
	
	if(*(volatile UINT8*)diagSeqRecord == LOW_RESET_SEQ_NORMAL)
	{
		printk("[HOST CARD LOW POWER] Low Power Mode Test start....\n");
		*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_FAIL;
	
		ctrlBootCountCompareSet();
    	
		//shut down host card
		fpgaHostCardReset(TRUE);
		udelay(100);
		fpgaHostCardPsDisable();
		mdelay(10000);
		
		//define HIC chip vendor ID & device ID
    	if(sysHostCardType == HOST_CARD_NONE)		//no host card
		{	
			printk("[HOST CARD LOW POWER] HIC is not attached yet!\n");
    	}
		else if(sysHostCardType == HOST_CARD_IB)		//IB host card
		{		
    		if(i2cEeprom64KBitRead(adap, VPD_EEPROM_ADDR, 0, &data, 1))
			{
				//printk("[HOST CARD LOW POWER] IB HIC passed\n");
				*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_GOOD;
       		}
			else
			{
				*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_FAIL;
				result = FALSE;
				printk("[HOST CARD LOW POWER] VPD can be accessed!!\n");
			}
    	}
		else if(sysHostCardType == HOST_CARD_FC)		//FC host card
		{		
    		if( /*hic9548ChSet(I2C_HIC_VPD_CHANNEL, 1) != 1*/ 0 )
			{
				if (i2cEeprom64KBitRead(adap, VPD_EEPROM_ADDR, 0, &data, 1))
				{
					//printk("[HOST CARD LOW POWER] FC HIC passed\n");
					*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_GOOD;
				}
				else
				{
					*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_FAIL;
					result = FALSE;
					printk("[HOST CARD LOW POWER] VPD can be accessed!!\n");
				}
			}
			else
			{
				*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_FAIL;
				result = FALSE;
				printk("[HOST CARD LOW POWER] I2C MUX can be accessed!!\n");
        	}
    	}
		else if(sysHostCardType == HOST_CARD_ISCSI)		//iSCSI host card
		{		
    		if(/*hic9548ChSet(I2C_HIC_VPD_CHANNEL, 1) != 1*/ 0)
			{
				if (i2cEeprom64KBitRead(adap, VPD_EEPROM_ADDR, 0, &data, 1))
				{
					//printk("[HOST CARD LOW POWER] ISCSI HIC passed\n");
					*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_GOOD;
				}
				else
				{
					*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_FAIL;
					result = FALSE;
					printk("[HOST CARD LOW POWER] VPD can be accessed!!\n");
				}
			}
			else
			{
				*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_HOST_CARD_FAIL;
				result = FALSE;
				printk("[HOST CARD LOW POWER] I2C MUX can be accessed!!\n");
       		}
    
		}
    
		fpgaHostCardPsEnable();     //turn on host card
		mdelay(300);				//delay for asserting HOST_PWR_GOOD (150ms at least)
		fpgaHostCardReset(FALSE);
		
		/*
		if(gManualDrive == 0)
		{	//manual mode, need reset
			mdelay(5);
			//reset UUT
			fpgaBoardReset(TRUE);
		}
		*/
		
		printk("[HOST CARD LOW POWER] Controller has not reset yet!! Please reset controller manually.\n");
	}
    
	//manual mode, verify if the controller has been reset.
	/*
	if(gManualDrive == 0)
	{
		if(ctrlBootCountGet() == ctrlBootCountCompareGet())
		{
			printk("[HOST CARD LOW POWER] Controller has not reset yet!! Please reset controller manually.\n");
		}
	}
	*/
    
	if(*(volatile UINT8*)diagSeqRecord == LOW_RESET_SEQ_HOST_CARD_FAIL)
	{
		commReportError(MEC_FPGA, FPGA_BBU_PS_MEC, FPGA_PS_HOST_CARD_DISABLE_ERROR, 
						MSG_FPGA_PS_HOST_CARD_DISABLE_ERROR, 0, 0, 0, 0);
		result = FALSE;
	}
 
	if(*(volatile UINT8*)diagSeqRecord == LOW_RESET_SEQ_HOST_CARD_GOOD)
	{
		result = TRUE;
	}
    
	//clear syndrome and checkpoint
	*(volatile UINT8*)diagSeqRecord = LOW_RESET_SEQ_NORMAL;   
	*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_0_LOCATION) = 0;    
	*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_1_LOCATION) = 0;    
	*(volatile UINT8*)(gFpgaBar2VirtualP + NVSRAM_ADDRESS_BASE + LOW_RESET_ENCODED_UINT8_2_LOCATION) = 0;
    
	return result;
}

/************************************************************************************************/

/**********************************/
/**  Midplane BBU Test function  **/
/**********************************/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuPsTest
*
* SUMMARY:     	Midplane power switch stress test
*
* SCOPE:        public
*
* DESCRIPTION:  Midplane power switch stress test
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaBbuPsTest(INT32 cycle)
{
    INT32 i;
	
	//fpgaMaskIntAll(TRUE);     //mask possible interrupts
	for(i = 0; i < cycle; i++){
		//fpgaBbuInplaceDiag();
		//fpgaBbuMonitorDiag(); 
		fpgaBbuDischargeDiag();
		fpgaPwrMonitorDiag();
		fpgaPwrInplaceDiag();
		printk("\n");
		printk("Test cycle: %d \n", i+1);
		printk("================================================== \n");
		mdelay(10000);
	}
	fpgaClearSubsystemInt();
	//fpgaMaskIntAll(FALSE);    //release masks
}

