/****************************************************************************
*                                                                           *
* Copyright 2004-2011 LSI Corporation. All rights reserved. This file is    *
* confidential and a trade secret of LSI Corporation. The receipt or        *
* possession of this file does not convey any rights to reproduce or        *
* disclose its contents or to manufacture, use, or sell anything it may     *
* describe, in whole, or in part, without the specific written consent of   *
* LSI Corporation.                                                          *
*                                                                           *
****************************************************************************/
/****************************************************************************
*                                                                           *
* NAME            fpgaCmdLib.c                                              *
* SUMMARY                                                                   *
* VERSION         %version: %                                               *
* UPDATE DATE     %date_modified: %                                         *
* PROGRAMMER      %created_by:    Shawn Wang %                              *
*                                                                           *
*        Copyright 2002-2011 LSI Corporation. All Rights Reserved.          *
*                                                                           *
* DESCRIPTION:                                                              *
*                                                                           *
****************************************************************************/

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include "commTypes.h"
#include "fpgaCmdLib.h"
#include "fpgaRegLib.h"
#include "fpgaNvsram.h"

#include "fpgaEppDiag.h"

#include "ssFun.h"

EXPORT_SYMBOL(fpgaNvsramLockAll);
EXPORT_SYMBOL(fpgaNvsramUnlockAll);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaNvsramLockControl
*
* SUMMARY:      Set the NVSRAM lock blocks
*
* SCOPE:        private
*
* DESCRIPTION:  FPGA can lock the NVSRAM content in 4kB increasement.
*				Values between 0x80 to 0xFF will also lock the entire NVSRAM.
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaNvsramLockControl
 (
 UINT8 lockBlocks	//the number of blocks needed to be locked
 )
{
	//the maximum block number is 128
	if(lockBlocks > 128)
	{
		lockBlocks = 128;    
	}
    
	//lock NVSRAM
	fpgaRegisterWrite(FPGA_NVSRAM_LOCK, lockBlocks);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaNvsramLockAll
*
* SUMMARY:      Lock the entire NVSRAM blocks
*
* SCOPE:        public
*
* DESCRIPTION:  Set block number as 128 to lock the entire NVSRAM blocks
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaNvsramLockAll(VOID)
{
	//lock entire NVSRAM
	//fpgaNvsramLockControl(FPGA_NVSRAM_LOCK_ENTIRE);
	fpgaRegisterWrite(FPGA_NVSRAM_LOCK, FPGA_NVSRAM_LOCK_MAX);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaNvsramUnlockAll
*
* SUMMARY:      Unlock the entire NVSRAM blocks
*
* SCOPE:        public
*
* DESCRIPTION:  Set block number as 0 to unlock the entire NVSRAM blocks
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaNvsramUnlockAll(VOID)
{
	//unlock entire NVSRAM
	fpgaNvsramLockControl(FPGA_NVSRAM_LOCK_NONE);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEnableQuickSwitch
*
* SUMMARY:      enable/diable all quick switch
*
* SCOPE:        public
*
* DESCRIPTION:  enable/diable all quick switch
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEnableQuickSwitch
(
 BOOLEAN status
)
{
    INT32 	result;
	UINT8 	temp;
    
	temp = fpgaRegisterRead(FPGA_ENABLE_QUICK_SWITCH);
    if(status)
	{
        temp = EN_LS_INTER_CON_BIT | EN_LS_CROSS_BIT | EN_PAIRING_BIT;
    }
	else
	{
        temp = 0x00;
    }

    result = fpgaRegisterWrite(FPGA_ENABLE_QUICK_SWITCH, temp); 
}

EXPORT_SYMBOL(fpgaEnableQuickSwitch);

INT32 fpgaCheckEppTimeout(VOID);
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppReadAddr
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  
*
* RESTRICTIONS:
* 
* RETURNS:      address
*/
UINT8 fpgaEppReadAddr(VOID)
{
    UINT8 	val;
    INT32 	timeout; 
    INT32 	t = 0;

    // sanity check, we MUST read from the FPGA once to initiate the transaction
    // then again for the data.
    if (fpgaRegisterRead(FPGA_EPP_STATUS_PORT) & EPP_FOLLOW_DR_BIT)
	{ 
    	printk("[FPGA EPP] OUT OF ORDER READS,read address\n");
    }

	//printk("[FPGA EPP] status: 0x%02X \n", fpgaRegisterRead(FPGA_EPP_STATUS_PORT));

    // this generates the read transaction
    val = fpgaRegisterRead(FPGA_EPP_ADDRESS_PORT);	//dummy
	//printk("[FPGA EPP] address read value: 0x%02X \n", val);

	//printk("[FPGA EPP] status: 0x%02X \n", fpgaRegisterRead(FPGA_EPP_STATUS_PORT));

    while(!(fpgaRegisterRead(FPGA_EPP_STATUS_PORT) & EPP_DONE_BIT))
    {
        if(++t > 0x80000)
        {
            fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, EPP_RESET_BIT);
            udelay(5);
            fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, 0x00);
			printk("[FPGA EPP]EPP Error - Read Address \n");
            break;
        }
        udelay(2);
    }

    udelay(5);

    // now we can read valid data
    val = fpgaRegisterRead(FPGA_EPP_ADDRESS_PORT);
	//printk("[FPGA EPP] address read value: 0x%02X \n", val);
	
    timeout = fpgaCheckEppTimeout();

    return val;
}

EXPORT_SYMBOL(fpgaEppReadAddr);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppReadData
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  
*
* RESTRICTIONS:
* 
* RETURNS:      data
*/
UINT8 fpgaEppReadData(VOID)
{
    UINT8 	val;
    INT32	timeout; 
    INT32 	t = 0;

    // sanity check, we MUST read from the FPGA once to initiate the transaction
    // then again for the data.
    if (fpgaRegisterRead(FPGA_EPP_STATUS_PORT) & EPP_FOLLOW_DR_BIT)
	{ 
    	printk("[FPGA EPP]OUT OF ORDER READS,read data\n");
    }

	//printk("[FPGA EPP] status: 0x%02X \n", fpgaRegisterRead(FPGA_EPP_STATUS_PORT));

    // this generates the read transaction
    val = fpgaRegisterRead(FPGA_EPP_DATA_PORT);
	
	//printk("[FPGA EPP] data read value: 0x%02X \n", val);

	//printk("[FPGA EPP] status: 0x%02X \n", fpgaRegisterRead(FPGA_EPP_STATUS_PORT));

    while(!(fpgaRegisterRead(FPGA_EPP_STATUS_PORT) & EPP_DONE_BIT))
    {
        if(++t > 0x80000)
        {
            fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, EPP_RESET_BIT);
            udelay(5);
            fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, 0x00);
			printk("[FPGA EPP]EPP Error - Read Data \n");
            break;
        }
        udelay(2);
    }
    udelay(5);
    // now we can read valid data
    val = fpgaRegisterRead(FPGA_EPP_DATA_PORT);

    timeout = fpgaCheckEppTimeout();

	//printk("[FPGA EPP] data read value: 0x%02X \n", val);

    return val;
}
EXPORT_SYMBOL(fpgaEppReadData);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppWriteAddr
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEppWriteAddr
 (
 UINT8 val
 )
{
    INT32 result;
    INT32 timeout;
    
	result = fpgaRegisterWrite(FPGA_EPP_ADDRESS_PORT, val);
	//printk("[FPGA EPP] status: 0x%02X \n", fpgaRegisterRead(FPGA_EPP_STATUS_PORT));
	
	timeout = fpgaCheckEppTimeout();
	
	//temp = fpgaRegisterRead(FPGA_EPP_STATUS_PORT);
	//if(temp & EPP_FOLLOW_AR_BIT)
	//{
	//	fpgaRegisterRead(FPGA_EPP_ADDRESS_PORT);
	//	printk("[FPGA EPP] AR follow-up read \n");
	//}
}

EXPORT_SYMBOL(fpgaEppWriteAddr);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppWriteData
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEppWriteData
 (
 UINT8 val
 )
{
    INT32 result;
    INT32 timeout; 
	
    result = fpgaRegisterWrite(FPGA_EPP_DATA_PORT, val);
	//printk("[FPGA EPP] status: 0x%02X \n", fpgaRegisterRead(FPGA_EPP_STATUS_PORT));
	
	timeout = fpgaCheckEppTimeout();
	
	//temp = fpgaRegisterRead(FPGA_EPP_STATUS_PORT);
	//if(temp & EPP_FOLLOW_DR_BIT)
	//{
	//	fpgaRegisterRead(FPGA_EPP_DATA_PORT);
	//	printk("[FPGA EPP] DR follow-up read \n");
	//}
}


EXPORT_SYMBOL(fpgaEppWriteData);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppWriteByte
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  write byte INT32o AHB bus of SAS2x36 through EPP INT32erface
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEppWriteByte
 (
 UINT32 	addr, 
 UINT8 		data
 )
{
    // 1st byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_0);
    fpgaEppWriteData((UINT8) (addr & 0x000000FF));

    // 2nd byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_1);
    fpgaEppWriteData((UINT8) ((addr & 0x0000FF00) >> 8));

    // 3rd byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_2);
    fpgaEppWriteData((UINT8) ((addr & 0x00FF0000) >> 16));

    // 4th byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_3);
    fpgaEppWriteData((UINT8) ((addr & 0xFF000000) >> 24));

    // send the 8 bits of data
    fpgaEppWriteAddr(AHB_DATA);
    fpgaEppWriteData(data);
}

EXPORT_SYMBOL(fpgaEppWriteByte);
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppWriteByte
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  read byte from AHB bus of SAS2x36 through EPP INT32erface
*
* RESTRICTIONS:
* 
* RETURNS:      data
*/
UINT8 fpgaEppReadByte
 (
 UINT32 addr
 )
{
    // 1st byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_0);
    fpgaEppWriteData((UINT8) (addr & 0x000000FF));

    // 2nd byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_1);
    fpgaEppWriteData((UINT8) ((addr & 0x0000FF00) >> 8));

    // 3rd byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_2);
    fpgaEppWriteData((UINT8) ((addr & 0x00FF0000) >> 16));

    // 4th byte of address
    fpgaEppWriteAddr(AHB_ADDRESS_3);
    fpgaEppWriteData((UINT8) ((addr & 0xFF000000) >> 24));

    // read the byte of data
    fpgaEppWriteAddr(AHB_DATA);
	
	// read the data from EPP data register in FPGA
    return fpgaEppReadData();
}

EXPORT_SYMBOL(fpgaEppReadByte);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppWriteDword
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  write word to AHB bus of SAS2x36 through EPP INT32erface
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEppWriteDword
 (
 UINT32 addr, 
 UINT32 data
 )
{   
    // 1st byte of Dword
    fpgaEppWriteByte(addr, (UINT8) (data & 0x000000FF)); 
    
    // 2nd byte of Dword
    fpgaEppWriteByte(addr + 1, (UINT8) ((data & 0x0000FF00) >> 8)); 

    // 3rd byte of Dword
    fpgaEppWriteByte(addr + 2, (UINT8) ((data & 0x00FF0000) >> 16)); 
    
    // 4th byte of Dword
    fpgaEppWriteByte(addr + 3, (UINT8) ((data & 0xFF000000) >> 24)); 
}

EXPORT_SYMBOL(fpgaEppWriteDword);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppReadDword
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  read word from AHB bus of SAS2x36 through EPP INT32erface
*
* RESTRICTIONS:
* 
* RETURNS:      data
*/
UINT32 fpgaEppReadDword
 (
 UINT32 addr
 )
{
    UINT32 data = 0;
    // 1st byte of Dword
    data = fpgaEppReadByte(addr) & 0x000000FF; 
    
    // 2nd byte of Dword
    data |= ((fpgaEppReadByte(addr + 1)) << 8) & 0x0000FF00; 
    
    // 3rd byte of Dword
    data |= ((fpgaEppReadByte(addr + 2)) << 16) & 0x00FF0000; 
    
    // 4th byte of Dword
    data |= ((fpgaEppReadByte(addr + 3)) << 24) & 0xFF000000;

    return data;
}

EXPORT_SYMBOL(fpgaEppReadDword);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaCheckEppTimeout
*
* SUMMARY:      Access AHB bus in SAS2x36 through FPGA EPP INT32erface
*
* SCOPE:        private
*
* DESCRIPTION:  check if EPP INT32erface is timeout
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE: 	normal
*				FALSE: 	timeout
*/
INT32 fpgaCheckEppTimeout(VOID)
{
    UINT8 temp;

    temp = fpgaRegisterRead(FPGA_EPP_STATUS_PORT);

    if (temp & EPP_TIMEOUT_BIT)
    {
        //EPP timeout detected, now try and clear it
        temp = fpgaRegisterRead(FPGA_EPP_STATUS_PORT);
        if ((temp & EPP_TIMEOUT_BIT) == 0x00)
        {
           // reset EPP to clear timeout
           fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, 0x04);
           udelay(5);
           fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, 0x00);
           return TRUE;
        }
		printk("[FPGA EPP] EPP Timeout!! Reset Fail \n");
        return FALSE;

    }
	
	//printk("EPP has no timeout \n");
    return TRUE;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEnableEppCrc
*
* SUMMARY:      Enable EPP CRC checking in SAS2x36
*
* SCOPE:        private
*
* DESCRIPTION:  Enable EPP CRC checking in SAS2x36 
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEnableEppCrc(VOID)
{
    UINT8 temp;

    // Make sure DISABLE CRC is off in the Status Register
    fpgaEppWriteAddr(AHB_STATUS);
    temp = fpgaEppReadAddr();
	temp = temp & 0x20;  // see debug port spec for status register
    
    //fpgaEppWriteAddr(AHB_STATUS);
    fpgaEppWriteData(temp);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaDisableEppCrc
*
* SUMMARY:      Disable EPP CRC checking in SAS2x36
*
* SCOPE:        private
*
* DESCRIPTION:  Disable EPP CRC checking in SAS2x36 
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaDisableEppCrc(VOID)
{
    UINT8 temp;

    // Make sure DISABLE CRC is on in the Status Register
    fpgaEppWriteAddr(AHB_STATUS);
    temp = fpgaEppReadAddr();
	temp = 0x10 | (temp & 0xC);  // see debug port spec for status register
    
    //fpgaEppWriteAddr(AHB_STATUS);
    fpgaEppWriteData(temp);
}
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBobcatUartMode
*
* SUMMARY:      switch Bobcat debug mode
*
* SCOPE:        private
*
* DESCRIPTION:  switch Bobcat debug mode
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaBobcatUartMode
(
 BOOLEAN status
)
{   
    INT32 	result;
    UINT8 	temp;
    
	temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_3);
    if(status)
	{
        temp |= BOBCAT_DBGM_BIT;
    }
	else
	{
        temp &= (~BOBCAT_DBGM_BIT);
    }
	
	fpgaEncodedRegisterWrite();
    result = fpgaRegisterWrite(FPGA_PROTECTED_WRITE_3, temp);
}
EXPORT_SYMBOL(fpgaBobcatUartMode);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEppInit
*
* SUMMARY:      EPP interface initialization
*
* SCOPE:        private
*
* DESCRIPTION:  EPP interface initialization
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaEppInit(VOID)
{
    UINT8  temp;

	//force Bobcat to EPP debug mode
	fpgaBobcatUartMode(FALSE);

	//reset EPP INT32erface
    fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, 0x04);
    mdelay(5);
    fpgaRegisterWrite(FPGA_EPP_CONTROL_PORT, 0x00);
    
    mdelay(10);

	//disable EPP CRC
    fpgaDisableEppCrc();  // WARNING: this generates transactions to the EXP's EPP INT32erface

	//modify AHB bus control register
    fpgaEppWriteAddr(AHB_CONTROL);
    temp = fpgaEppReadAddr();
    temp = 0x01;  // set up EXP EPP for byte transactions, no INTs
    fpgaEppWriteData(temp);

}


EXPORT_SYMBOL(fpgaEppInit);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaSas2x36PhyEnable
*
* SUMMARY:     	Enable SAS2x36 PHY
*
* SCOPE:        private
*
* DESCRIPTION:  Enable SAS2x36 PHY
*
* RESTRICTIONS:
* 
* RETURNS:     	
*/
VOID fpgaSas2x36PhyEnable
 (
 BOOLEAN status
 )
{
    UINT8 	temp;

    //release disable bit
    temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_2);
	if(status)
	{
		temp &= (~DISABLE_PHY_BIT);
	}
	else
	{
		temp |= DISABLE_PHY_BIT;
	}
    fpgaEncodedRegisterWrite();
    fpgaRegisterWrite(FPGA_PROTECTED_WRITE_2, temp);
}

EXPORT_SYMBOL(fpgaSas2x36PhyEnable);

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaSas2x36Reset
*
* SUMMARY:      reset Bobcat SAS2x36
*
* SCOPE:        private
*
* DESCRIPTION:  reset Bobcat SAS2x36
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaSas2x36Reset
(
 BOOLEAN status
)
{   
    INT32 	result;
    UINT8 	temp;
    
	temp = fpgaRegisterRead(FPGA_PROTECTED_WRITE_3);
    if(status)
	{
        temp |= SAS2X36_RST_BIT;
    }
	else
	{
        temp &= (~SAS2X36_RST_BIT);
    }
	
	fpgaEncodedRegisterWrite();
    result = fpgaRegisterWrite(FPGA_PROTECTED_WRITE_3, temp);
}

EXPORT_SYMBOL(fpgaSas2x36Reset);

int ssDoesCtlRoleSwap(void)
{
	volatile UINT8	*idSwapDataPtr = (volatile UINT8 *)(gFpgaBar2VirtualP+SSFUN_LOG_HEAD+ssBYTE_LogHeadMaxSize+ssBYTE_CTLROLESWAP);

	return (*idSwapDataPtr == SS_CTLROLESWAP)? 1:0;
}
EXPORT_SYMBOL(ssDoesCtlRoleSwap);

UINT8 fpgaCtlRoleRead(VOID)
{
#ifdef ALANWANG		// 2010/07/13
	UINT8	id;
	id=fpgaRegisterRead(FPGA_BOARD_ID);
	// FPGA_SUBSYSTEM_ID (MIDPLANE) ??
	#if defined(CTRLER_SOYUZ)
	return (id&0x80)?1:0;  // 0: UUT ctl, 1:Fixture ctl (soyuz)
	#else
	return (id&0x80)?0:1;  // 0: UUT ctl, 1:Fixture ctl
	#endif

#else // ALANWANG
 #define SSFUN_LOG_HEAD		0x2000
 #define ssBYTE_LogHeadMaxSize	64
 #define	ssBYTE_CTLROLESWAP	0
 //#define ssBYTE_InNvram	((volatile UINT8 *)(gFpgaBar2VirtualP+SSFUN_LOG_HEAD+ssBYTE_LogHeadMaxSize))
 #define	SS_CTLROLESWAP	BIT(7)
 	UINT8	id, idSwapData;
 	volatile UINT8	*idSwapDataPtr = (volatile UINT8 *)(gFpgaBar2VirtualP+SSFUN_LOG_HEAD+ssBYTE_LogHeadMaxSize+ssBYTE_CTLROLESWAP);
 	
 	idSwapData = *idSwapDataPtr;
 	
 	if(idSwapData & 0x7F) {
 		idSwapData=0;
 		*idSwapDataPtr=0;
 	}
 
 	id=fpgaRegisterRead(FPGA_BOARD_ID);
// 	if(id != 0x9e && id != 0x1e)
// 		return 2;	// Unknown ctl

//	printk("%s: SwapData=0x%02x(*DataPtr=0x%02x, id=0x%02x)\n", __FUNCTION__, idSwapData, *idSwapDataPtr, id);
 	id ^= (idSwapData & SS_CTLROLESWAP);

	#if defined(CTRLER_SOYUZ)
 	return (id&0x80)?1:0;	// 0: UUT ctl, 1:Fixture ctl (soyuz)
	#else
	return (id&0x80)?0:1;   // 0: UUT ctl, 1:Fixture ctl
	#endif
#endif // ALANWANG
}
EXPORT_SYMBOL(fpgaCtlRoleRead);


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaLedSync
*
* SUMMARY:      Set FPGA LED sync
*
* SCOPE:        private
*
* DESCRIPTION:  Set LED sync, then LED will operate as the pattern defined
*
* RESTRICTIONS:
*
* RETURNS:
*/
VOID fpgaLedSync
 (
 BOOLEAN status
 )
{
    INT32   result;
    UINT8   temp;

    //read the original content
    temp = fpgaRegisterRead(FPGA_MISCELLANEOUS_CONTROL_3);

    if(status){
        //enable LED_SYNC bit
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_3, (temp | LED_SYNC_BIT));
    }else{
        //disable LED_SYNC bit
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_3, (temp & (~LED_SYNC_BIT)));
    }
}
EXPORT_SYMBOL(fpgaLedSync);


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaLedAssign
*
* SUMMARY:      Assing LED pattern to specific LED
*
* SCOPE:        private
*
* DESCRIPTION:  Assign the LED pattern to specific LED index, then related LED
*               will operate as the the pattern defined.
*
* RESTRICTIONS:
*
* RETURNS:
*/
VOID fpgaLedAssign
 (
 UINT8 ledIndex,
 UINT8 ledPatternIndex
 )
{
    INT32 result;

    //Write LED index INT32o LED Index Register
    result = fpgaRegisterWrite(FPGA_LED_INDEX, ledIndex);

    //Write LED Pattern Index INT32o pattern selection
    result = fpgaRegisterWrite(FPGA_LED_PATTERN_SELECTION, ledPatternIndex);

    //reset the LED
    fpgaLedSync(TRUE);
    mdelay(1); //delay 1ms
    fpgaLedSync(FALSE);
}
EXPORT_SYMBOL(fpgaLedAssign);


