/*******************************************************************************

NAME            fpgaDriveInplDiag.c
SUMMARY         Zebulon FPGA drive inplace pin diagnostics function
VERSION         %version: 1 %
UPDATE DATE     %date_modified: Apr 28 14:20 2009 %
PROGRAMMER      %created_by:    Jim Tu %

        Copyright 2009 Quanta Corporation. All Rights Reserved.

DESCRIPTION:
        This file has code to test drive inplace function of Zebulon FPGA

NOTES:

REFERENCE:

*******************************************************************************/

#include <linux/delay.h>

#include <commTypes.h>
#include <commRpt.h>
#include <commMEC.h>

#include "fpgaMidplaneI2cLib.h"
#include "fpgaRegLib.h"

#define FPGA_DRIVE_INPL_MASK    0x00000001
#if defined(CTRLER_PIKESPEAK)
#define DRIVE_IN_PLACE_NUM      24
#else
#define DRIVE_IN_PLACE_NUM      20
#endif

/***************************************/
/**  Drive In-place Related Function  **/
/***************************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaDriveInPlace
*
* SUMMARY:      return drive inplace status
*
* SCOPE:        private
*
* DESCRIPTION:  read the drive inplace register and return
*
* RESTRICTIONS:
* 
* RETURNS:      Drive inplace statue form drive 1 (bit 0) 
*				to drive 24 (bit 23)
*/
UINT32 fpgaDriveInPlace(VOID)
{
    //Read register
    UINT32 driveInpl0 = fpgaRegisterRead(FPGA_DRIVE_INPLACE_0);
    UINT32 driveInpl1 = fpgaRegisterRead(FPGA_DRIVE_INPLACE_1);
    UINT32 driveInpl2 = fpgaRegisterRead(FPGA_DRIVE_INPLACE_2);
    
    //Set the highest byte as 0xff
    return (0xFF000000 | (driveInpl2 << 16) | (driveInpl1 << 8) | driveInpl0);
}

/*********************************/
/**  Main Diagnostics Function  **/
/*********************************/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaDriveInPlaceDiag
*
* SUMMARY:      diagnostics FPGA drive inplace function
*
* SCOPE:        public
*
* DESCRIPTION:  diagnostics FPGA drive inplace function by pulling low each drive
*				inplace pin and verify if the related register bit is asserted
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE: 	Pass
*				FALSE:	Fail
*/
INT32 fpgaDriveInplaceDiag(VOID)
{
	INT32 	result = TRUE;
	UINT32 	driveStatus, count;
    
	if(midplaneTypeVerification())
	{	//Quanta Midplane												
		for(count = 0; count < DRIVE_IN_PLACE_NUM; count++)
		{
			//verify drive inplace pin low
			midplaneGpioWrite(&D_INPL_L[count], 0);
						
			//Read drive status
			driveStatus = fpgaDriveInPlace();
			
			//if the bit status is high
			if(!(driveStatus & (FPGA_DRIVE_INPL_MASK << count)))
			{
				commReportError(MEC_FPGA, FPGA_DRIVE_INPLACE_MEC, (UINT16)(FPGA_DRIVE_INPLACE_ERROR_BASE + count),
								MSG_FPGA_DRIVE_INPLACE_ERROR, count+1, 1, 0, DG_FPGA_DRIVE_FORMAT);
				result = FALSE;
			}
		
			//verify drive inplace pin high
			midplaneGpioWrite(&D_INPL_L[count], 1);
				
			//Read drive status
			driveStatus = fpgaDriveInPlace();

			//if the bit status is low
			if((driveStatus & (FPGA_DRIVE_INPL_MASK << count)))
			{
				commReportError(MEC_FPGA, FPGA_DRIVE_INPLACE_MEC, (UINT16)(FPGA_DRIVE_INPLACE_ERROR_BASE + count),
								MSG_FPGA_DRIVE_INPLACE_ERROR, count+1, 2, 0, DG_FPGA_DRIVE_FORMAT);
				result = FALSE;
			}
			
			commPostProgress("[FPGA DRIVE INPL] Drive Inplace Test Progress", count, DRIVE_IN_PLACE_NUM);
		}
		printk("[FPGA DRIVE INPL] Drive Inplace Test Progress: Done. \n");
	}
	else
	{
		printk("[FPGA DRIVE INPL] Please use manufactury test midplane. \n");
	}

    return result;
}
