/*******************************************************************************

NAME            fpgaLedDiag.c
SUMMARY         Zebulon FPGA LED function diagnostics
VERSION         %version: 1 %
UPDATE DATE     %date_modified: Apr 27 20:30 2009 %
PROGRAMMER      %created_by:    Jim Tu %

        Copyright 2009 Quanta Corporation. All Rights Reserved.

DESCRIPTION:
        This file has code to test LED function of Zebulon FPGA

NOTES:

REFERENCE:

*******************************************************************************/

#include <linux/kernel.h>
#include <linux/delay.h>

#include <commTypes.h>
#include <commRpt.h>
#include <commMEC.h>

#include "fpgaMidplaneI2cLib.h"
#include "fpgaLedDiag.h"
#include "fpgaRegLib.h"
#include "fpgaCmdLib.h"

EXTERN VOID fpgaEnableQuickSwitch(BOOLEAN status);
EXTERN VOID fpgaEnableLsCross(BOOLEAN status);

UINT8	gSevenSegDisplayStatus;
/******************************/
/***  LED Control Function  ***/
/******************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaLedPatternConfig
*
* SUMMARY:      Set FPGA LED pattern
*
* SCOPE:        private
*
* DESCRIPTION:  Set LED pattern index, seq, highest state and blinking duration
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaLedForProcessorBoot
 (
 UINT8 status
 )
{
	INT32 	result;
	UINT8	temp;
	
	//read the original content
	temp = fpgaRegisterRead(FPGA_MISCELLANEOUS_CONTROL_3);
	
    if(status){
		//enable LED_SYNC bit
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_3, (temp | EN_UP_BOOT_BIT));
    }else{
		//disable LED_SYNC bit
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_3, (temp & (~EN_UP_BOOT_BIT)));
    }	       
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaLedPatternConfig
*
* SUMMARY:      Set FPGA LED pattern
*
* SCOPE:        private
*
* DESCRIPTION:  Set LED pattern index, seq, highest state and blinking duration
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaLedPatternConfig
 (
 UINT8 patternIndex, 
 UINT8 patternSeq, 
 UINT8 patternHighestState, 
 UINT8 patternDuration
 )
{
	INT32 result;
	
	//set index
	if((patternIndex != 0x00) && (patternIndex != 0x01)) //0 & 1 are pre-defined
	{
		if(patternIndex <= NUMBER_LED_PATTERN_PIKESPEAK)
        {
            result = fpgaRegisterWrite(FPGA_LED_PATTERN_INDEX, patternIndex);
            //else Error_Report();
        }
    }
    
    //set sequence
    patternSeq &= 0x01;                                 //only two sequence
    result = fpgaRegisterWrite(FPGA_LED_PATTERN_SEQUENCE, patternSeq);
    
    //set highest state
    patternHighestState &= 0x07;
    result = fpgaRegisterWrite(FPGA_LED_PATTERN_HIGHEST_STATE, patternHighestState);
    
    //set duration
    if(!patternHighestState)
    {
        result = fpgaRegisterWrite(FPGA_LED_PATTERN_DURATION, patternDuration);
    }                       
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaLedPatternSimpleBlink
*
* SUMMARY:      Set FPGA LED blink pattern
*
* SCOPE:        private
*
* DESCRIPTION:  Set LED pattern index 3 as blink every second
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaLedPatternSimpleBlink(VOID)
{
    fpgaLedPatternConfig(0x03, 0x01, 0x07, 0x1F);           //blink every 1 second
	//index: 3
	//sequence: 1
	//hightest state: 7
	//blink duration: 31
}


/****************************************/
/***  Seven-segment Control Function  ***/
/****************************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaSevenSegmentDisplaySII
*
* SUMMARY:      Control 7-segment display in system information mode
*
* SCOPE:        public
*
* DESCRIPTION:  1. Blink the following sequence:
*				Category Code: SE   =>fpgaSevenSegmentDisplaySII(1,0,'S','E')
*				Specific Code: 65   =>fpgaSevenSegmentDisplaySII(1,1,'6','5')
*				Clear Space: 00     =>fpgaSevenSegmentDisplaySII(1,2,'0','0')
*				Category Code: CL   =>fpgaSevenSegmentDisplaySII(1,3,'C','L')
*				Specific Code: D1   =>fpgaSevenSegmentDisplaySII(1,4,'D','L')
*
*				2. No blink display
*				Display CL          =>fpgaSevenSegmentDisplaySII(0,0,'C','L')        
*				(until any other registers are written)
*
*				Input Variables : 	blink: set 7-segment display in blink mode
*                      				index: set blink index
*                   				SevenSegDisplayL: left letter
*                 				    SevenSegDisplayR: right letter
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaSevenSegmentDisplaySet
 (
 UINT8 blink, 
 UINT8 index, 
 UINT8 SevenSegDisplayL, 
 UINT8 SevenSegDisplayR
 )
{
    INT32 	result;
    
	fpgaLedForProcessorBoot(FALSE);
	
    if(blink)
	{
        result = fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_BLINK_CONTROL, (index | 0x80));
    }
	else
	{
        result = fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_BLINK_CONTROL, 0x01);             //non-blink mode
    }
    
    switch(SevenSegDisplayL)
	{
        case '0':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_0);
           break;
        case '1':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_1);
           break;
        case '2':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_2);
           break;
        case '3':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_3);
           break;
        case '4':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_4);
           break;
        case '5':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_5);
           break;
        case '6':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_6);
           break;
        case '7':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_7);
           break;
        case '8':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_8);
           break;
        case '9':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_9);
           break;
        case 'A':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_A);
           break;
        case 'B':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_B);
           break;
        case 'C':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_C);
           break;
        case 'D':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_D);
           break;
        case 'E':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_E);
           break;
        case 'F':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_F);
           break;
        case 'I':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_I);
           break;
        case 'P':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_P);
           break;
        case 'S':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_S);
           break;
        case 'L':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_L);
           break;
        default:
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, DISPLAY_OFF);
           break;
    }
    
    switch(SevenSegDisplayR)
	{
        case '0':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_0);
           break;
        case '1':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_1);
           break;
        case '2':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_2);
           break;
        case '3':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_3);
           break;
        case '4':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_4);
           break;
        case '5':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_5);
           break;
        case '6':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_6);
           break;
        case '7':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_7);
           break;
        case '8':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_8);
           break;
        case '9':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_9);
           break;
        case 'A':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_A);
           break;
        case 'B':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_B);
           break;
        case 'C':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_C);
           break;
        case 'D':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_D);
           break;
        case 'E':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_E);
           break;
        case 'F':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_F);
           break;
        case 'I':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_I);
           break;
        case 'P':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_P);
           break;
        case 'S':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_S);
           break;
        case 'L':
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_L);
           break;
        default:
           result |= fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, DISPLAY_OFF);
           break;
    }
}

/***************************************/
/***  Specific LED Control Function  ***/
/***************************************/

/***  Front Panel LED  ***/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaControllerFaultLedEnable
*
* SUMMARY:      enable/disable controller fault LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to controller fault LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaControllerFaultLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(CONTROLLER_FAULT_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(CONTROLLER_FAULT_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaControllerRtrLedEnable
*
* SUMMARY:      enable/disable controller ready-to-remove LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to controller ready-to-remove LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaControllerRtrLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(CONTROLLER_RTR_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(CONTROLLER_RTR_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaCacheActLedEnable
*
* SUMMARY:      enable/disable controller cache active LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to controller cache active LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaCacheActLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(CACHE_ACTIVE_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(CACHE_ACTIVE_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaInterCacheActLedEnable
*
* SUMMARY:      enable/disable controller INT32ernal cache active LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to controller INT32ernal cache active LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaInterCacheActLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(INTERNAL_CACHE_ACTIVE_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(INTERNAL_CACHE_ACTIVE_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuFaultLedEnable
*
* SUMMARY:      enable/disable BBU fault LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to BBU fault LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaBbuFaultLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(BBU_FAULT_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(BBU_FAULT_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaBbuChargeLedEnable
*
* SUMMARY:      enable/disable BBU charge LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to BBU charge LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaBbuChargeLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(BBU_CHARGING_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(BBU_CHARGING_LED, LED_PATTERN_OFF);		//OFF
    }
}


/***  Midplane LED  ***/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaAltRtrLedEnable
*
* SUMMARY:      enable/disable alternate controller ready-to-remove LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to alternate controller ready-to-remove LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaAltRtrLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(ALT_RTR_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(ALT_RTR_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaSummaryFaultLedEnable
*
* SUMMARY:      enable/disable system summary fault LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to system summary fault LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaSummaryFaultLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(SUMMARY_FAULT_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(SUMMARY_FAULT_LED, LED_PATTERN_OFF);		//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaTrayIdentifyLedEnable
*
* SUMMARY:      enable/disable system tray identify LED
*
* SCOPE:        public
*
* DESCRIPTION:  set the "ON"/"OFF" pattern to system tray identify LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaTrayIdentifyLedEnable
 (
 BOOLEAN status
 )
{
    if(status)
	{
        fpgaLedAssign(TRAY_IDENTIFY_LED, LED_PATTERN_ON);		//ON
    }
	else
	{
        fpgaLedAssign(TRAY_IDENTIFY_LED, LED_PATTERN_OFF);		//OFF
	}
}


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaRtrPs0LedEnable
*
* SUMMARY:      enable/disable power supply 0 ready-to-remove LED
*
* SCOPE:        public
*
* DESCRIPTION:  set "ON"/"OFF" to power supply 0 ready-to-remove  LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaRtrPs0LedEnable
 (
 BOOLEAN status
 )
{
    INT32 	result;
    UINT8	temp;
	
	temp = fpgaRegisterRead(FPGA_MISCELLANEOUS_CONTROL_2);
    if(status)
	{
		result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_2, (temp | RTR_PS_0_LED_BIT));	//ON
    }
	else
	{
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_2, (temp & (~RTR_PS_0_LED_BIT)));	//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaRtrPs1LedEnable
*
* SUMMARY:      enable/disable power supply 1 ready-to-remove LED
*
* SCOPE:        public
*
* DESCRIPTION:  set "ON"/"OFF" to power supply 1 ready-to-remove  LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaRtrPs1LedEnable
 (
 BOOLEAN status
 )
{
    INT32 result;
    UINT8	temp;
	
	temp = fpgaRegisterRead(FPGA_MISCELLANEOUS_CONTROL_2);
    if(status)
	{
		result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_2, (temp | RTR_PS_1_LED_BIT));	//ON
    }
	else
	{
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_2, (temp & (~RTR_PS_1_LED_BIT)));	//OFF
    }
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaOverTempLedEnable
*
* SUMMARY:      enable/disable over temp. LED
*
* SCOPE:        public
*
* DESCRIPTION:  set "ON"/"OFF" to over temp. LED
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaOverTempLedEnable
 (
 BOOLEAN status
 )
{
    INT32 result;
    UINT8	temp;
	
	temp = fpgaRegisterRead(FPGA_MISCELLANEOUS_CONTROL_2);
    if(status)
	{
		result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_2, (temp | OVER_TEMP_LED_BIT));	//ON
    }
	else
	{
        result = fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_2, (temp & (~OVER_TEMP_LED_BIT)));	//OFF
    }
}


/*********************************/
/***  LED Diagnostic Function  ***/
/*********************************/

/***  Drive LEDs  ***/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaAllDriveLedOn
*
* SUMMARY:      enable/disable all drive LEDs
*
* SCOPE:        public
*
* DESCRIPTION:  set "ON"/"OFF" pattern to drive LEDs
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaAllDriveLedOn
 (
 BOOLEAN status
 )
{   
    UINT8 i;
	
	if(status)
	{
		for(i = 0x00; i < 0x48; i++)
		{										//for all drive LED index
			fpgaLedAssign(i, LED_PATTERN_ON);	//Drive LEDs constant on
		}
	}
	else
	{
		for(i = 0x00; i < 0x48; i++)
		{										//for all drive LED index
			fpgaLedAssign(i, LED_PATTERN_OFF);	//Drive LEDs constant off
		}
	}
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaDriveLedBlink
*
* SUMMARY:      blink all drive LEDs every second
*
* SCOPE:        public
*
* DESCRIPTION:  set blink pattern to drive LEDs
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaDriveLedBlink(VOID)
{   
    UINT8 i;
    
    fpgaLedPatternSimpleBlink();	//set pattern 3 as blinking
    for(i = 0x00; i < 0x46; i++)
	{								//for all drive LED index
        fpgaLedAssign(i, 0x03);		//Drive LED blinks
    }
}    


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaDriveLedDiag
*
* SUMMARY:      diagnostics the drive LEDs
*
* SCOPE:        public
*
* DESCRIPTION:  Turn the drive LED and verify the LED pins through I2C-GPIO device
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE: 	Pass
*				FALSE:	Fail
*/
INT32 fpgaDriveLedDiag(VOID)
{
	INT32 	result = TRUE;
	UINT8 	i;

	if(midplaneTypeVerification())
	{	//Quanta Midplane
		//verify off-state
		fpgaAllDriveLedOn(FALSE);
		for(i = 0; i < NUMBER_LED_PIKESPEAK; i++)
		{
			//verify FAULT LED, the pin should be high
			if(!midplaneGpioRead(&D_FAULT_LED_L[i]))
			{          
				commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_DRIVE_LED_ERROR_BASE + i,
								MSG_FPGA_DRIVE_LED_ERROR, i, 1, 0, DG_FPGA_DRIVE_FORMAT);
				result = FALSE;
			}
	
			//verify REPLACE LED, the pin should be high
			if(!midplaneGpioRead(&D_REPL_LED_L[i]))
			{         
				commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_DRIVE_LED_ERROR_BASE + i,
								MSG_FPGA_DRIVE_LED_ERROR, i, 2, 0, DG_FPGA_DRIVE_FORMAT);
				result = FALSE;
			}
			
			commPostProgress("[FPGA LED] Drive LED Test Progress", i, (NUMBER_LED_PIKESPEAK * 2));
		}
    
		//verify on-state
		fpgaAllDriveLedOn(TRUE);
		for(i = 0; i < NUMBER_LED_PIKESPEAK; i++)
		{
			//verify FAULT LED, the pin should be low
			if(midplaneGpioRead(&D_FAULT_LED_L[i]))
			{          
				commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_DRIVE_LED_ERROR_BASE + i,
								MSG_FPGA_DRIVE_LED_ERROR, i, 1, 0, DG_FPGA_DRIVE_FORMAT);
			result = FALSE;
			}
	
			//verify REPLACE LED, the pin should be low
			if(midplaneGpioRead(&D_REPL_LED_L[i]))
			{          
				commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_DRIVE_LED_ERROR_BASE + i,
								MSG_FPGA_DRIVE_LED_ERROR, i, 2, 0, DG_FPGA_DRIVE_FORMAT);
				result = FALSE;
			}
			
			commPostProgress("[FPGA LED] Drive LED Test Progress", (i + NUMBER_LED_PIKESPEAK), 
							 (NUMBER_LED_PIKESPEAK * 2));
    	}	
		
		printk("[FPGA LED] Drive LED Test Progress: Done. \n");
    }
	else
	{
		//LSI Midplane
		//Do nothing, need operator INT32ervention
	}
	
	fpgaAllDriveLedOn(FALSE);
	return result;
}

/***  LED on the midplane  ***/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaSystemLedDiag
*
* SUMMARY:      diagnostics the midplane LEDs (except drive LED)
*
* SCOPE:        public
*
* DESCRIPTION:  Turn the midplane LED and verify the LED pins 
*				through I2C-GPIO device
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE: 	Pass
*				FALSE:	Fail
*/
INT32 fpgaSystemLedDiag(VOID)
{
    INT32 result = TRUE;
	
	fpgaEnableQuickSwitch(TRUE);

	if(midplaneTypeVerification())
	{	//Quanta Midplane

		//verify ALT_RTR_LED_L on
		fpgaAltRtrLedEnable(TRUE);
		//if(midplaneGpioRead(&ALT_RTR_LED_L))
		//{ 
		//	commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_ALT_RTR_LED_ERROR,
		//					MSG_FPGA_ALT_RTR_LED_ERROR, 0, 0, 0, 0);
		//	printk("[FPGA LED] ALT_RTR on \n");
		//	result = FALSE;
		//}
		
		//verify ALT_RTR_LED_L off
		fpgaAltRtrLedEnable(FALSE);
		if(!midplaneGpioRead(&ALT_RTR_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_ALT_RTR_LED_ERROR,
							MSG_FPGA_ALT_RTR_LED_ERROR, 0, 0, 0, 0);
			printk("[FPGA LED] ALT_RTR off \n");
			result = FALSE;
		}
		
		//verify SUMMARY_FLT_LED_L on
		fpgaSummaryFaultLedEnable(TRUE);
		if(midplaneGpioRead(&SUMMARY_FLT_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_SUMMARY_FAULT_LED_ERROR,
							MSG_FPGA_SUMMARY_FAULT_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		//verify SUMMARY_FLT_LED_L off
		fpgaSummaryFaultLedEnable(FALSE);
		if(!midplaneGpioRead(&SUMMARY_FLT_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_SUMMARY_FAULT_LED_ERROR,
							MSG_FPGA_SUMMARY_FAULT_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
    
		//verify BOX_ID_LED on
		fpgaTrayIdentifyLedEnable(TRUE);
		if(midplaneGpioRead(&BOX_ID_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_TRAY_IDENTIFY_LED_ERROR,
							MSG_FPGA_TRAY_IDENTIFY_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		//verify BOX_ID_LED off
		fpgaTrayIdentifyLedEnable(FALSE);
		if(!midplaneGpioRead(&BOX_ID_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_TRAY_IDENTIFY_LED_ERROR,
							MSG_FPGA_TRAY_IDENTIFY_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
	
		//verify PS0_RTR_LED_L on
		fpgaRtrPs0LedEnable(TRUE);
		if(midplaneGpioRead(&PS0_RTR_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_RTR_PS_0_LED_ERROR,
							MSG_FPGA_RTR_PS_0_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		//verify PS0_RTR_LED_L off
		fpgaRtrPs0LedEnable(FALSE);
		if(!midplaneGpioRead(&PS0_RTR_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_RTR_PS_0_LED_ERROR,
							MSG_FPGA_RTR_PS_0_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		//verify PS1_RTR_LED on
		fpgaRtrPs1LedEnable(TRUE);
		if(midplaneGpioRead(&PS1_RTR_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_RTR_PS_1_LED_ERROR,
							MSG_FPGA_RTR_PS_1_LED_ERROR, 0, 0, 0, 0);
			printk("[FPGA LED] PS1_RTR on \n");
			result = FALSE;
		}
			
		//verify PS1_RTR_LED off
		fpgaRtrPs1LedEnable(FALSE);
		if(!midplaneGpioRead(&PS1_RTR_LED_L))
		{ 
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_RTR_PS_1_LED_ERROR,
                        	MSG_FPGA_RTR_PS_1_LED_ERROR, 0, 0, 0, 0);
			printk("[FPGA LED] PS1_RTR off \n");
            result = FALSE;
		}
		
		//verify OVER_TEMP_LED_L on
		fpgaOverTempLedEnable(TRUE);
		if(midplaneGpioRead(&BOX_OVER_TEMP_LED_L))
		{		
			commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_OVER_TEMP_LED_ERROR,
							MSG_FPGA_OVER_TEMP_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
		
		//verify OVER_TEMP_LED_L off
		fpgaOverTempLedEnable(FALSE);
		if(!midplaneGpioRead(&BOX_OVER_TEMP_LED_L))
		{ 
            commReportError(MEC_FPGA, FPGA_LED_MEC, FPGA_OVER_TEMP_LED_ERROR,
							MSG_FPGA_OVER_TEMP_LED_ERROR, 0, 0, 0, 0);
			result = FALSE;
		}
	}
	else
	{
		//LSI Midplane
		//Do nothing, need operator INT32ervention
	}
    
	return result;
}


/****************************/
/***  LED diag procedure  ***/
/****************************/

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaMidplaneLdDiag
*
* SUMMARY:      diagnostics all the midplane LEDs
*
* SCOPE:        public
*
* DESCRIPTION:  Turn the midplane LED and verify the LED pins 
*				through I2C-GPIO device
*
* RESTRICTIONS:
* 
* RETURNS:      TRUE: 	Pass
*				FALSE:	Fail
*/
INT32 fpgaMidplaneLedDiag(VOID)
{
    if(fpgaSystemLedDiag() & fpgaDriveLedDiag())
	{
        return TRUE;
    }
	else
	{
        return FALSE;
    }
}

/***  LED on the front panel  ***/
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaFrontPanelLEDDiagOn
*
* SUMMARY:      turn on all front panel LEDs
*
* SCOPE:        public
*
* DESCRIPTION:  Turn the front panel LED and verify the LED pins 
*				through I2C-GPIO device
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaFrontPanelLEDDiagOn(VOID)
{
	fpgaLedForProcessorBoot(FALSE);
	fpgaEnableLsCross(FALSE);
	
    //five front panel LEDS
	fpgaControllerFaultLedEnable(TRUE);
    fpgaControllerRtrLedEnable(TRUE);
    fpgaCacheActLedEnable(TRUE);
	fpgaBbuFaultLedEnable(TRUE);
    fpgaBbuChargeLedEnable(TRUE);
	
	//INT32ernal LED
	fpgaInterCacheActLedEnable(TRUE);
	    
    fpgaSevenSegmentDisplaySet(0, 1, '8', '8');
    //mdelay(1000);
    
    fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_1, 0xFF);
    fpgaRegisterWrite(FPGA_SEVEN_SEGMENT_DISPLAY_2, 0xFF);
	
	//need operator intervention or light sensor to verify
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaFrontPanelLEDDiagOff
*
* SUMMARY:      turn off all front panel LEDs
*
* SCOPE:        public
*
* DESCRIPTION:  Turn the front panel LED and verify the LED pins 
*				through I2C-GPIO device
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpgaFrontPanelLEDDiagOff(VOID)
{
	fpgaLedForProcessorBoot(FALSE);

	//five front panel LEDs
    fpgaControllerFaultLedEnable(FALSE);
    fpgaControllerRtrLedEnable(FALSE);
    fpgaCacheActLedEnable(FALSE);
	fpgaBbuFaultLedEnable(FALSE);
    fpgaBbuChargeLedEnable(FALSE);
    
	//INT32ernal LED
	fpgaInterCacheActLedEnable(FALSE);

	//turn off 7-segment display
    fpgaSevenSegmentDisplaySet(0, 1, 'K', 'K');
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         midRmvLedStart
*
* SUMMARY:      turn on remove control LED from midplane
*
* SCOPE:        public
*
* DESCRIPTION:  
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID midRmvLedStart(VOID)
{
	fpgaEnableQuickSwitch(TRUE);
	midplaneGpioWrite(&RDY_RMV_LED_IN, 1);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         midRmvLedStop
*
* SUMMARY:      turn off remove control LED from midplane
*
* SCOPE:        public
*
* DESCRIPTION:  
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID midRmvLedStop(VOID)
{
	fpgaEnableQuickSwitch(TRUE);
	midplaneGpioWrite(&RDY_RMV_LED_IN, 0);
}


#if 0
/*******************************************************************************
* PROCEDURE
*
* NAME:         fpga7SegDisplaySet
*
* SUMMARY:      tools for system stress test
*
* SCOPE:        public
*
* DESCRIPTION:  use 7-segment display to show which stress task is processing
*
* RESTRICTIONS:
* 
* RETURNS:      
*/
VOID fpga7SegDisplaySet
 (
 UINT8 item, 
 UINT8 status
 )
{
	UINT8 	statusTemp;
	
	statusTemp = gSevenSegDisplayStatus;
	
	switch(item)
	{
		case 1:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x01;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xFE;
			}
			break;
			
		case 2:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x02;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xFD;
			}
			break;
			
		case 3:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x04;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xFB;
			}
			break;
			
		case 4:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x08;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xF7;
			}
			break;
			
		case 5:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x10;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xEF;
			}
			break;
			
		case 6:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x20;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xDF;
			}
			break;
			
		case 7:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x40;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0xBF;
			}
			break;
			
		case 8:
			if(status)
			{
				gSevenSegDisplayStatus |= 0x80;
			} 
			else 
			{
				gSevenSegDisplayStatus &= 0x7F;
			}
			break;
		
		default:
			printk("[FPGA LED] Wrong Input. \n");
			break;
	}
	
	//set seven-segment display in diagnostic information mode and show the status
    fpgaRegisterWrite(FPGA_MISCELLANEOUS_CONTROL_3, 
						fpgaRegisterRead(FPGA_MISCELLANEOUS_CONTROL_3) | DEBUG_LED_EN_BIT);
	fpgaRegisterWrite(FPGA_DIAGNOSTIC_STATUS_LED, gSevenSegDisplayStatus);
}
#endif


