/*******************************************************************************

NAME            fpgaRegLib.c
SUMMARY         Zebulon FPGA register access function
VERSION         %version: 1 %
UPDATE DATE     %date_modified: Apr 27 14:40 2009 %
PROGRAMMER      %created_by:    Jim Tu %

        Copyright 2009 Quanta Corporation. All Rights Reserved.

DESCRIPTION:
        This file has code to access Zebulon FPGA register for Pikes Peak RAID
		controller

NOTES:

REFERENCE:

*******************************************************************************/

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/pci.h>
#include "commTypes.h"

#include "fpgaRegLib.h"
#include "fpgaCmdLib.h"

UINT32 sysCtrlRole;
EXPORT_SYMBOL(sysCtrlRole);

EXTERN VOID fpgaEppInit(VOID);
EXTERN VOID fpgaSas2x36PhyEnable(BOOLEAN status);


/***  Encoded Write Sequence Define ***/
STATIC UINT8 	fpgaEncodedWriteSequence[8] = {0x80, 0x01, 0x02, 0x84, 0x89, 0x93, 0x27, 0xCE};

struct pci_dev*        pFpgaDev = NULL;        //global structure for Zebulon FPGA

VOID*			gFpgaBar1VirtualP;
VOID*			gFpgaBar2VirtualP;

EXPORT_SYMBOL(gFpgaBar2VirtualP);
EXPORT_SYMBOL(fpgaRegisterShow);
EXPORT_SYMBOL(fpgaEncodedRegisterWrite);
EXPORT_SYMBOL(fpgaRegisterWrite);
EXPORT_SYMBOL(fpgaRegisterRead);


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaRegisterRead
*
* SUMMARY:      Return the Zebulon FPGA register value
*
* SCOPE:        public
*
* DESCRIPTION:  Low level function to access Zebulon FPGA register
*
* RESTRICTIONS:
*
* RETURNS:      8-bit data
*/
UINT8 fpgaRegisterRead
 (
 UINT32	fpgaOffset	//register offset from BAR1
 )
{
    //printk("[fpgaRegisterRead] offset(0x%x): 0x%x\n", fpgaOffset, *(volatile UINT8*)(gFpgaBar1VirtualP + fpgaOffset));
    return *(volatile UINT8*)(gFpgaBar1VirtualP + fpgaOffset);
}

/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaRegisterWrite
*
* SUMMARY:      Modify the Zebulon FPGA register value
*
* SCOPE:        public
*
* DESCRIPTION:  Low level function to access Zebulon FPGA register
*
* RESTRICTIONS:
*
* RETURNS:      Error:
*               Succcess: TRUE
*/
INT32 fpgaRegisterWrite
 (
 UINT32		fpgaOffset,	//register offset from BAR1
 UINT8		data		//8-bit data needs to be written INT32o register
 )
{
    *(volatile UINT8*)(gFpgaBar1VirtualP + fpgaOffset) = data;
    return TRUE;
}


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaEncodedRegisterWrite
*
* SUMMARY:      Modify the Zebulon FPGA encoded register value to get the
*				access right of protected write register
*
* SCOPE:        public
*
* DESCRIPTION:  Low level function to access Zebulon FPGA encoded register
*
* RESTRICTIONS:
*
* RETURNS:      Error:
*               Succcess: TRUE
*/
VOID fpgaEncodedRegisterWrite(VOID)
{
    INT32 	i;

    for(i=0; i<8; i++)
	{
		//write 8 sequence code INT32 encoded register
        *(volatile UINT8*)(gFpgaBar1VirtualP + (UINT32)FPGA_ENCODED_WRITE) = fpgaEncodedWriteSequence[i];
    }
}


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaProtectedRegisterWrite
*
* SUMMARY:      Modify the Zebulon FPGA protected register value
*
* SCOPE:        public
*
* DESCRIPTION:  Low level function to access Zebulon FPGA register
*
* RESTRICTIONS:
*
* RETURNS:      Error: FALSE
*               Succcess: TRUE
*/
INT32 fpgaProtectedRegisterWrite
 (
 UINT32		fpgaOffset,	//register offset from BAR1
 UINT8		data		//8-bit data needs to be written INT32o register
 )
{
	if (fpgaOffset < 0x31 || fpgaOffset > 0x34)
		return FALSE;

	fpgaEncodedRegisterWrite();
	*(volatile UINT8*)(gFpgaBar1VirtualP + fpgaOffset) = data;
	return TRUE;
}
EXPORT_SYMBOL(fpgaProtectedRegisterWrite);


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaNvsramByteRead
*
* SUMMARY:
*
* SCOPE:        public
*
* DESCRIPTION:
*
* RESTRICTIONS:
*
* RETURNS:      8-bit data
*/
UINT8 fpgaNvsramByteRead
 (
 UINT32	nvsramOffset	//register offset from BAR2
 )
{
    return *(volatile UINT8*)(gFpgaBar2VirtualP + nvsramOffset);
}
EXPORT_SYMBOL(fpgaNvsramByteRead);


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaNvsramByteWrite
*
* SUMMARY:
*
* SCOPE:        public
*
* DESCRIPTION:
*
* RESTRICTIONS:
*
* RETURNS:      Error:
*               Succcess: TRUE
*/
INT32 fpgaNvsramByteWrite
 (
 UINT32		nvsramOffset,	//register offset from BAR1
 UINT8		data		//8-bit data needs to be written INT32o register
 )
{
    *(volatile UINT8*)(gFpgaBar2VirtualP + nvsramOffset) = data;
    return TRUE;
}
EXPORT_SYMBOL(fpgaNvsramByteWrite);


/*******************************************************************************
* PROCEDURE
*
* NAME:         fpgaInit
*
* SUMMARY:      get Zebulon FPGA BAR1 and BAR2
*
* SCOPE:        public
*
* DESCRIPTION:
*
* RESTRICTIONS:
*
* RETURNS:
*/
int fpgaInit( VOID *mmioVirtual, VOID *nvsramVirtual )
{
	gFpgaBar1VirtualP = mmioVirtual;
    gFpgaBar2VirtualP = nvsramVirtual;

	sysCtrlRole = fpgaCtlRoleRead();
#if 0
	//init Bobcat SAS2x36 EPP interface
	fpgaEppInit();
#endif
	return (0);
}


VOID fpgaRegisterShow
 (
 UINT32	fpgaOffset	//register offset from BAR1
 )
{
	printk("data = 0x%02X \n ",*(volatile UINT8*)(gFpgaBar1VirtualP + fpgaOffset));
}


