/*
 * wdt-zebulon.c -- Zebulon FPGA WDT driver
 *
 *
 * (C) Copyright 2015, Jerry Lee <jerry.yj.lee@foxconn.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */
#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/io.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/signal.h>

#include "commTypes.h"
#include "fpgaRegLib.h"
#include "fpgaCmdLib.h"
#include "pci-zebulon.h"
#include "fpga-subfuncs.h"
#include "midplane.h"

static struct fasync_struct *fasync_queue;
static struct resource *wdt_irq;
static volatile void __iomem *mmioVirtual = 0;
static struct platform_device *wdt_device = NULL;

//interrupt processing function only send SIGIO signal
irqreturn_t InterruptHandler(int irq, void *dev_id)
{
    if (!(fpgaRegisterRead(FPGA_INTERRUPT_SOURCE_INT_B) & WDT_INT_BIT))
        return IRQ_NONE;

    if (fasync_queue)
        kill_fasync(&fasync_queue, SIGIO, POLL_IN);

    printk(KERN_DEBUG "FPGA WDT interrupt handled\n");

    return IRQ_HANDLED;
}

//request or free irq  for user mode
int
wdt_request_free_irq(void *arg)
{
    int ret = -EINVAL;
    int irqRequest;
    static int noused;

    irqRequest = FALSE;

    ret = copy_from_user(&irqRequest, arg, sizeof(int));
    if (ret)
    {
        printk("zebulon_wdt_request_free_irq: Copy from user space error!\n");
        return ret;
    }

    if (irqRequest)  //request irq
    {
        ret = request_irq(wdt_irq->start, InterruptHandler, IRQF_SHARED, "Zebulon-WDT", wdt_device);
        if (ret)
        {
            printk("IRQ %d handler could not be set, status: %d\n",wdt_irq->start,ret);
        }
        else
        {
            /* enable FPGA WDT interrupt */
            writeb(
                readb(mmioVirtual + FPGA_INT_MASK_REG) & ~(FPGA_INT_MASK_SUBSYS | FPGA_INT_MASK_WDT),
                      mmioVirtual + FPGA_INT_MASK_REG );
        }
    }
    else    //free irq
    {
        /* Disable FPGA WDT interrupt */
        writeb(
            readb(mmioVirtual + FPGA_INT_MASK_REG ) | (FPGA_INT_MASK_SUBSYS | FPGA_INT_MASK_WDT),
                  mmioVirtual + FPGA_INT_MASK_REG );
        free_irq(wdt_irq->start, wdt_device);
    }

    return ret;
}

/**
 * wdt_ioctl - main ioctl entry point (unlocked)
 * @file - (struct file)
 * @cmd - ioctl opcode
 * @arg -
 */
static long
wdt_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    int ret = 0;

    /*
    ** extract the type and number bitfields, and don't decode
    ** wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok()
    */
    if (_IOC_TYPE(cmd) != ZEBULON_WDT_MODULE_IOC_MAGIC) return -ENOTTY;

    switch (cmd)
    {
        case IOCTL_ZEBULON_WDT_REQUEST_FREE_IRQ:
             ret = wdt_request_free_irq((void *)arg);
             break;

        case IOCTL_ZEBULON_WDT_TEST_INTERRUPT:    //for test
             //ret = processor_test_interrupt((void *)arg);
             break;

        default:
             return -ENOTTY;
    }

    return ret;
}

/**
 * wdt_fasync -
 * @fd -
 * @filep -
 * @mode -
 *
 * Called when application request fasyn callback handler.
 */
static int
wdt_fasync(int fd, struct file *filep, int mode)
{
    return fasync_helper(fd, filep, mode, &fasync_queue);
}

/**
 * wdt_release -
 * @inode -
 * @filep -
 *
 * Called when application releases the fasyn callback handler.
 */
static int
wdt_release(struct inode *inode, struct file *filep)
{
    return fasync_helper(-1, filep, 0, &fasync_queue);
}

static const struct file_operations wdt_fops = {
    .owner          = THIS_MODULE,
    .unlocked_ioctl = wdt_ioctl,
    .release        = wdt_release,
    .fasync         = wdt_fasync,
};

static struct miscdevice zebulon_wdt_dev = {
    .minor  = MISC_DYNAMIC_MINOR,
    .name   = ZEBULON_WDT_DEV_NAME,
    .fops   = &wdt_fops,
};

static int zebulon_wdt_probe(struct platform_device *pdev)
{
    wdt_device = pdev;
    // IO remapping
    mmioVirtual = getFpgaMmioVirtual();
    if( !mmioVirtual )
    {
        dev_err( &pdev->dev, "ioremap failed\n" );
        return -ENOMEM;
    }

    wdt_irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);

    if (!wdt_irq)
    {
        dev_err( &pdev->dev, "no irq resource?\n" );
        return -ENODEV;
    }

    fasync_queue = NULL;

    return 0;
}

static int zebulon_wdt_remove(struct platform_device *pdev)
{
    return 0;
}

static struct platform_driver zebulon_wdt_platform_driver = {
    .probe  = zebulon_wdt_probe,
    .remove = zebulon_wdt_remove,
    .driver = {
        .name   = FPGA_ZEBULON_CWT_NAME,
    },
};

/**
 * zebulon_wdt_init - main entry point for wdt.
 *
 */
void
zebulon_wdt_init(void)
{
    int ret;

    ret = misc_register(&zebulon_wdt_dev);
    if (ret != 0)
        printk(KERN_ERR "%s can't register misc device [minor=%d] (err=%d)\n",
               ZEBULON_WDT_DEV_NAME, MISC_DYNAMIC_MINOR, ret);

    ret = platform_driver_register(&zebulon_wdt_platform_driver);

    if (ret)
    {
        misc_deregister(&zebulon_wdt_dev);
        printk( KERN_WARNING "Cannot register Zebulon-WDT platform driver\n" );
    }
}

/**
 * zebulon_wdt_exit - exit point for wdt
 *
 */
void
zebulon_wdt_exit(void)
{
    platform_driver_unregister(&zebulon_wdt_platform_driver);

    misc_deregister(&zebulon_wdt_dev);
}

