/* Intel Broadwell-DE Memory Controller kernel module
 *
 * This driver supports the memory controllers found on the Intel
 * processor family Broadwell.
 *
 * This file may be distributed under the terms of the
 * GNU General Public License version 2 only.
 *
 * Copyright (c) 2014 by:
 *	 Ken Hu <che-wei.hu@foxconn.com>
 */

#include <asm/delay.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/cdev.h>
#include <linux/spinlock.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <asm/mce.h>
#include <asm/msr.h>
#include "mem_drv.h"

MODULE_LICENSE("GPL");

#define NO_ERROR 0

// For Broadwell-DE IMC PCI device ID
#define PCI_DEVICE_ID_INTEL_BROADWELL_SAD0              0x6ffc  // 15.4
#define PCI_DEVICE_ID_INTEL_BROADWELL_HA0               0x6fa0  // 18.0
#define PCI_DEVICE_ID_INTEL_BROADWELL_HA0_PERFMON       0x6f30  // 18.1
#define PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_TA           0x6fa8  // 19.0
#define PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_RAS          0x6f71  // 19.1
#define PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_TAD0         0x6faa  // 19.2
#define PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_TAD1         0x6fab  // 19.3

// Workaround for BIOS: There is no pci device for 0x6fb2 and 0x6fb3
#define PCI_DEVICE_BASE_ADDRESS_IMC0_ERR0               0x8FFA2000  // 20.2
#define PCI_DEVICE_BASE_ADDRESS_IMC0_ERR1               0x8FFA3000  // 20.3

#define MAX_PCI_DEV_NUM_FOR_INTEL_BROADWELL_IMC         8     

// For Broadwell-DE IMC MCA Bank ID
#define MCA_BANK_ID_INTEL_BROADWELL_IMC_CHANNEL0        9
#define MCA_BANK_ID_INTEL_BROADWELL_IMC_CHANNEL1        10

#define MAX_CHANNEL_NUM                     2
#define MAX_REG_NUM_CORRECTED_ERROR_COUNT   4

#define FIELD_MASK_2BITS                    0x3
#define FIELD_MASK_3BITS                    0x7
#define FIELD_MASK_4BITS                    0xF
#define FIELD_MASK_6BITS                    0x3F                
#define FIELD_MASK_11BITS                   0x7FF
#define FIELD_MASK_13BITS                   0x1FFF
#define FIELD_MASK_14BITS                   0x3FFF
#define FIELD_MASK_20BITS                   0xFFFFF

// SAD registers
#define REG_OFFSET_DRAM_RULE_0              0x60
#define REG_OFFSET_SHIFT_DRAM_RULE          8
#define BIT_OFFSET_RULE_ENABLE              0
#define BIT_OFFSET_INTERLEAVE_MODE          1
#define BIT_OFFSET_ATTR                     2
#define BIT_OFFSET_LIMIT_ADDR               6
#define BIT_OFFSET_A7MODE                   26

#define REG_OFFSET_INTERLEAVE_LIST_0        0x64
#define REG_OFFSET_SHIFT_INTERLEAVE_LIST    8
#define BIT_OFFSET_SHIFT_PACKAGE            4

// HA registers
#define REG_OFFSET_TAD_0                    0x40
#define BIT_OFFSET_LIMIT                    12
#define BIT_OFFSET_CHANNELWAYS              8
#define REG_OFFSET_HASYSDEFEATURE           0x80
#define REG_OFFSET_HASYSDEFEATURE2          0x84
#define BIT_OFFSET_MC_CHAN_SHIFT_UP_ENABLE  22
#define BIT_OFFSET_MC_CHAN_HASH_EN          21
#define BIT_OFFSET_DEMAND_SCRUB             9

// HA Perfmon registers
#define REG_OFFSET_MC5_STATUS_SHADOW_LOW    0x68
#define REG_OFFSET_MC5_STATUS_SHADOW_HIGH   0x6C

// IMC TA registers
#define REG_OFFSET_MCMTR                    0x7C
#define BIT_OFFSET_PASS76                   16
#define BIT_OFFSET_DDR4                     14
#define BIT_OFFSET_ECC_EN                   2
#define BIT_OFFSET_CLOSE_PG                 0

#define REG_OFFSET_TADWAYNESS_0             0x80
#define BIT_OFFSET_TAD_LIMIT                12
#define BIT_OFFSET_TAD_SKT_WAY              10
#define BIT_OFFSET_TAD_CH_WAY               8
#define BIT_OFFSET_TAD_CH_TGT_1             2
#define BIT_OFFSET_TAD_CH_TGT_0             0

// IMC TAD0 registers
#define REG_OFFSET_DIMMMTR_0                0x80
#define BIT_OFFSET_DDR4_3DSNUMRANKS_CS      23
#define BIT_OFFSET_DDR4_MODE                20
#define BIT_OFFSET_RANK_DISABLE             16
#define BIT_OFFSET_DIMM_POP                 14
#define BIT_OFFSET_RA_WIDTH                 2
#define BIT_OFFSET_CA_WIDTH                 0
#define REG_OFFSET_AMAP                     0x8C
#define BIT_OFFSET_HSXPLUS                  0
#define REG_OFFSET_TADCHNILVOFFSET_0        0x90
#define BIT_OFFSET_TAD_OFFSET               6
#define REG_OFFSET_RIRWAYNESSLIMIT_0        0x108
#define BIT_OFFSET_RIR_VAL                  31
#define BIT_OFFSET_RIR_WAY                  28
#define BIT_OFFSET_RIR_LIMIT                1
#define REG_OFFSET_RIRILV_OFFSET_BASE       0x120
#define BIT_OFFSET_RIR_RNK_TGT              20
#define BIT_OFFSET_RIR_OFFSET               2
#define REG_OFFSET_RSP_FUNC_ADDR_MATCH_LO   0x1C0
#define REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI   0x1C4
#define BIT_OFFSET_RSP_FUNC_ADDR_MATCH_EN   11
#define REG_OFFSET_RSP_FUNC_ADDR_MASK_LO    0x1C8
#define REG_OFFSET_RSP_FUNC_ADDR_MASK_HI    0x1CC

// IMC ERR0 registers
#define REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_DEV0_XOR_MSK    0x200
#define REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_DEV1_XOR_MSK    0x204
#define REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_EXTRA           0x208


#define REG_OFFSET_CORRERRCNT               0x104

#define REG_OFFSET_SCRUBCTL                 0xA0
#define BIT_OFFSET_SCRUBEN                  31

#define _SAD_DRAM                           20
#define _TAD                                12
#define _RIR                                5
#define _RANK                               8

#define _SAD_OFFSET                         26
#define _TAD_OFFSET                         _SAD_OFFSET
#define _RIR_OFFSET                         29


struct addr_tranlate_info {
    unsigned long long  core_addr;
    int                 socket_id;
    int                 mc_id;
    int                 node_id;
    int                 channel;
    //int                 vmse_channel;
    //int                 channel_2nd;
    unsigned long long  channel_addr;
    int                 log_rank;
    unsigned long long  rank_addr;
    int                 chip_select;
    int                 chip_id;
    int                 dimm;
    int                 phys_rank;
    int                 bank_group;
    int                 bank;
    int                 row;
    int                 column;
    int                 dramSadDict[_SAD_DRAM];
    int                 dramSadIntDict[_SAD_DRAM];
    int                 haTadDict[_TAD];
    int                 tadOffDict[_TAD];
    int                 mcTadDict[_TAD];
    int                 rirDict[_RIR];
    int                 rirOffDict[_RIR];
    int                 haDict[2];
    int                 mcMtrDict[1];
    int                 dimmMtrDict[2][2];
    int                 amapDict[4];
};

typedef enum {
    ENTRY_ID_SAD0 = 0,
    ENTRY_ID_HA0,
    ENTRY_ID_HA0_PERFMON,
    ENTRY_ID_TA,
    ENTRY_ID_IMC0_TAD0,
    ENTRY_ID_IMC0_TAD1,
    ENTRY_ID_IMC0_RAS,
    ENTRY_ID_IMC0_ERR0,
    ENTRY_ID_IMC0_ERR1,
} PCI_DEV_ENTRY_INDEX;

static struct pci_dev *g_pdev_list[MAX_PCI_DEV_NUM_FOR_INTEL_BROADWELL_IMC];
static int g_major_num;
static struct class *g_pmem_drv_class;
static DEFINE_SPINLOCK(g_mem_drv_lock);
static DEFINE_SPINLOCK(g_error_count_lock);
static bool g_enable_demand_scrub = false;
static bool g_save_demand_scrub = false;
static u32 g_error_count = 0;
static int g_error_counts[MAX_CHANNEL_NUM][MAX_REG_NUM_CORRECTED_ERROR_COUNT][2];
static u32 g_error_entries = 0;
static struct corr_err_info *g_corr_err_head = NULL;
static bool g_ecc_enable = false;
static unsigned char *g_map_addr[MAX_CHANNEL_NUM];
static unsigned int g_save_msr_low = 0;
static unsigned int g_save_msr_high = 0;
static int chXorBitPos[8] = {12, 14, 16, 18, 20, 22, 24, 26};

static unsigned int getbits(unsigned int value, int start, int end)
{
    unsigned int mask = 0, tmp;
    int i;

    for (i = start; i <= end; i++) {
        mask |= (0x1 << i);
    }

    tmp = value & (mask << start);
    tmp >>= start;

    return tmp;
}

static unsigned long long getbitslong(unsigned long long value, int start, int end)
{
    unsigned long long mask = 0, tmp;
    int i;

    for (i = start; i <= end; i++) {
        mask |= (0x1ULL << i);
    }

    tmp = value & (mask << start);
    tmp >>= start;

    return tmp;
}

static void add_corr_err_info_to_list(struct corr_err_info **corr_err_entry)
{
    //unsigned long flags = 0;
    struct corr_err_info *curr_entry;
    
    if (corr_err_entry == NULL)
        return;
    
    // Update total error count
    g_error_count += (*corr_err_entry)->error_count;

    if (g_corr_err_head == NULL) {
        // Error list is empty, this error entry is list head
        g_error_entries++;        
        g_corr_err_head = *corr_err_entry;
        curr_entry = g_corr_err_head;
    }
    else {
        curr_entry = g_corr_err_head;
        while (curr_entry) {

            // If this error entry has existed, update its error count only
            if ((curr_entry->fail_addr == (*corr_err_entry)->fail_addr) &&
                (curr_entry->rank_id == (*corr_err_entry)->rank_id) &&
                (curr_entry->channel_id == (*corr_err_entry)->channel_id) &&
                (curr_entry->cpu_id == (*corr_err_entry)->cpu_id)) {
                curr_entry->error_count += (*corr_err_entry)->error_count;
                kfree(*corr_err_entry);
                (*corr_err_entry) = NULL;
                break;
            }
                
            if (curr_entry->next == NULL) {
                // Append this error entry to list because it is a new one                
                g_error_entries++;
                curr_entry->next = *corr_err_entry;
                break;
            }

            curr_entry = curr_entry->next;
        }
    }

    printk("[mem_drv] Corrected error(ch:%d.rank:%d.count:%d) has been record\n",
            curr_entry->channel_id,
            curr_entry->rank_id,
            curr_entry->error_count);
}


static void update_error_count(int update_channel)
{
    unsigned long flags = 0;
    int channel, end_channel, reg;
    unsigned int value;
    int tmp_error_count;

    if (update_channel == 0) {
        channel = 0;
        end_channel = channel + 1;
    } else if (update_channel == 1) {
        channel = 1;
        end_channel = channel + 1;
    } else {
        channel = 0;
        end_channel = MAX_CHANNEL_NUM;
    }

    spin_lock_irqsave(&g_error_count_lock, flags);

    for (; channel < end_channel; channel++) {
        if (g_map_addr[channel] == NULL)
            continue;

        // Save current error counts
        for (reg = 0; reg < MAX_REG_NUM_CORRECTED_ERROR_COUNT; reg++) {
            value = *(unsigned int *)(g_map_addr[channel] + REG_OFFSET_CORRERRCNT + (reg << 2));

            tmp_error_count = (int)(value & 0x7FFF);

            tmp_error_count -= (int)(g_error_counts[channel][reg][0] & 0x7FFF);

            if ((value & 0x8000) && (tmp_error_count < 0)) {
                g_error_counts[channel][reg][0] += (0x8000 + tmp_error_count);
            }else
                g_error_counts[channel][reg][0] += tmp_error_count;

            value = (value >> 16) & 0xFFFF;

            tmp_error_count = (int)(value & 0x7FFF);

            tmp_error_count -= (int)(g_error_counts[channel][reg][1] & 0x7FFF);

            if ((value & 0x8000) && (tmp_error_count < 0))
                g_error_counts[channel][reg][1] += (0x8000 + tmp_error_count);
            else
                g_error_counts[channel][reg][1] += tmp_error_count;
        }
    }

    spin_unlock_irqrestore(&g_error_count_lock, flags);
}

static int check_memory_error_data(struct notifier_block *nb, unsigned long val,
				                   void *data)
{
    int ret = NO_ERROR;
    struct mce *mce = (struct mce *)data;
    unsigned long lock1_flags = 0;
    unsigned long lock2_flags = 0;
    int channel, end_channel;
    int i;
    struct corr_err_info *perr_info_entry = NULL;
    u16 err_cnt0 = 0, err_cnt1 = 0;
    u8 rank; 
    int tmp_error_counts[MAX_CHANNEL_NUM][MAX_REG_NUM_CORRECTED_ERROR_COUNT][2];

    switch (mce->bank)
    {
        case 7:
            wrmsr_safe(0x41D, 0, 0);
            wrmsr_safe(0x41E, 0, 0);
            channel = 0;
            end_channel = 2;
            break;
        case 9:
            wrmsr_safe(0x425, 0, 0);
            wrmsr_safe(0x426, 0, 0);
            channel = 0;
            end_channel = channel + 1;
            break;
        case 10:
            wrmsr_safe(0x429, 0, 0);
            wrmsr_safe(0x42A, 0, 0);
            channel = 1;
            end_channel = channel + 1;
            break;
        default:
            printk("[mem_drv] ignore mce event come from MCE bank %d\n", mce->bank);
            return NOTIFY_DONE;
    }
    
    printk("[mem_drv] MCE bank %d has mce event\n", mce->bank);
    printk("[mem_drv] MCE status 0x%016llx\n", mce->status);

    spin_lock_irqsave(&g_mem_drv_lock, lock1_flags);

    spin_lock_irqsave(&g_error_count_lock, lock2_flags);

    memcpy(tmp_error_counts, g_error_counts, sizeof(tmp_error_counts));

    spin_unlock_irqrestore(&g_error_count_lock, lock2_flags);

    if (mce->bank == 7)
        update_error_count(2);  // update error counts for both of channels
    else
        update_error_count(channel);

    spin_lock_irqsave(&g_error_count_lock, lock2_flags);

    for (; channel < end_channel; channel++) {
        for (i = 0; i < MAX_REG_NUM_CORRECTED_ERROR_COUNT; i++) {

            // Calculate actual error count for current mce event
            err_cnt0 = g_error_counts[channel][i][0] - tmp_error_counts[channel][i][0];
            err_cnt1 = g_error_counts[channel][i][1] - tmp_error_counts[channel][i][1];

            // Check record of first rank
            if (err_cnt0)
            {    
                rank = i * 2;

                perr_info_entry = kzalloc(sizeof(struct corr_err_info), GFP_KERNEL);

                if (perr_info_entry == NULL)
                {
                    printk("[mem_drv] Corrected error records (ch:%d.rank:%d.count:%d) are lost!",
                           channel, rank, err_cnt0);
                    
                    goto fail;
                }
                
                perr_info_entry->channel_id = channel;
                perr_info_entry->rank_id = rank;
                perr_info_entry->error_count = err_cnt0;
                perr_info_entry->fail_addr = mce->addr;
                perr_info_entry->next = NULL;

                // Update error info list
                add_corr_err_info_to_list(&perr_info_entry);
            }

            // Check record of second rank
            if (err_cnt1)
            {    
                rank = i * 2 + 1;

                perr_info_entry = kzalloc(sizeof(struct corr_err_info), GFP_KERNEL);

                if (perr_info_entry == NULL)
                {                 
                    printk("[mem_drv] Corrected error records (ch:%d.rank:%d.count:%d) are lost!",
                           channel, rank, err_cnt1);

                    goto fail;
                }

                perr_info_entry->channel_id = channel;
                perr_info_entry->rank_id = rank;
                perr_info_entry->error_count = err_cnt1;
                perr_info_entry->fail_addr = mce->addr;
                perr_info_entry->next = NULL;
                
                // Update error info list
                add_corr_err_info_to_list(&perr_info_entry);
            }
        }
    }

    spin_unlock_irqrestore(&g_error_count_lock, lock2_flags);

    spin_unlock_irqrestore(&g_mem_drv_lock, lock1_flags);

fail:
    return NOTIFY_STOP;
}


static struct notifier_block g_mem_mce_dec = {
    .notifier_call      = check_memory_error_data,
};

static int set_demand_scrub(bool enable)
{
    struct pci_dev *pdev;
    u32 value;
    unsigned long flags = 0;

    if (enable == g_enable_demand_scrub)
        return NO_ERROR;
    
    // Get Broadwell's HA0 registers
    pdev = g_pdev_list[ENTRY_ID_HA0];

    if (0 != pci_read_config_dword(pdev, REG_OFFSET_HASYSDEFEATURE, &value)) {
        printk("[mem_drv] Read PCI(%02x:%02x.%x) fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn));
        return READ_PCI_DEV_FAIL;
    }    
    
    if (enable)
        // Enable demand scrub
        value |= (1 << BIT_OFFSET_DEMAND_SCRUB);
    else
        // Disable demand scrub
        value &= ~(1 << BIT_OFFSET_DEMAND_SCRUB);

    spin_lock_irqsave(&g_mem_drv_lock, flags);
    if (0 != pci_write_config_dword(pdev, REG_OFFSET_HASYSDEFEATURE, value)) {
        spin_unlock_irqrestore(&g_mem_drv_lock, flags); 
        printk("[mem_drv] Write PCI(%02x:%02x.%x) fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn));
        return WRITE_PCI_DEV_FAIL;
    }

    g_enable_demand_scrub = enable;

    spin_unlock_irqrestore(&g_mem_drv_lock, flags); 
    
    return NO_ERROR;
}

static void clear_err_list(void)
{
    unsigned long flags = 0;
    struct corr_err_info *curr_entry;
    struct corr_err_info *temp_entry;
    
    if (!g_error_entries)
        return;
    
    spin_lock_irqsave(&g_mem_drv_lock, flags);

    curr_entry = g_corr_err_head;
    while (curr_entry) {
        temp_entry = curr_entry->next;
        kfree(curr_entry);
        curr_entry = temp_entry;
        g_error_entries--;
    }
    g_error_count = 0;
    g_corr_err_head = NULL;

    spin_unlock_irqrestore(&g_mem_drv_lock, flags);
}

static int mem_drv_open(struct inode *inod, struct file *filp)
{
    // Increment module usage count
    if (0 == try_module_get(THIS_MODULE)) {  
        return MODULE_OPEN_FAIL;
    }

    return NO_ERROR;
}

static int mem_drv_close(struct inode *inod, struct file *filp)
{
    // Decrement module usage count
    module_put(THIS_MODULE);

    return NO_ERROR;
}

static int forwardTad(struct addr_tranlate_info *ti)
{
    struct pci_dev *pdev;
    unsigned int regOffset;
    unsigned int haD, haD2, mcmtr, tadReg, imcTadReg;
    unsigned int splitBit, sw, cw, imcCw, swXcw, chInterleave, tmp;
    int match, tad_index, i;
    long long previous_limit;
    unsigned long long limit;

    if (ti->haDict[0] != -1) {
        haD = ti->haDict[0];
    } else {
        pdev = g_pdev_list[ENTRY_ID_HA0];
        regOffset = REG_OFFSET_HASYSDEFEATURE;

        if (0 != pci_read_config_dword(pdev, regOffset, &haD)) {
            printk("[mem_drv] forwardTad: Read PCI(%02x:%02x.%x) Reg 0x%x fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn), regOffset);
            return ADDR_TRANSLATION_FAIL;
        } else {
            ti->haDict[0] = haD;
        }
    }

    if (ti->haDict[1] != -1) {
        haD2 = ti->haDict[1];
    } else {
        regOffset = REG_OFFSET_HASYSDEFEATURE2;
        if (0 != pci_read_config_dword(pdev, regOffset, &haD2)) {
            printk("[mem_drv] forwardTad: Read PCI(%02x:%02x.%x) fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn));
            return ADDR_TRANSLATION_FAIL;
        } else {
            ti->haDict[1] = haD2;
        }
    }

    if (ti->mcMtrDict[0] != -1) {
        mcmtr = ti->mcMtrDict[0];
    } else {
        pdev = g_pdev_list[ENTRY_ID_TA];
        regOffset = REG_OFFSET_MCMTR;
        if (0 != pci_read_config_dword(pdev, regOffset, &mcmtr)) {
            printk("[mem_drv] Read PCI(%02x:%02x.%x) Reg 0x%x fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn), regOffset);
            return -READ_PCI_DEV_FAIL;
        } else {
            ti->mcMtrDict[0] = mcmtr;
        }
    }

    if (((haD2 >> BIT_OFFSET_MC_CHAN_SHIFT_UP_ENABLE) & 0x1) != ((mcmtr >> BIT_OFFSET_PASS76) & FIELD_MASK_2BITS)) {
        printk("Illegal mc setting: haD2: 0x%x mcmtr: 0x%x\n", haD2, mcmtr);
        return ADDR_TRANSLATION_FAIL;
    }

    splitBit = 6 + ((haD2 >> BIT_OFFSET_MC_CHAN_SHIFT_UP_ENABLE) & 0x1);

    previous_limit = -1LL;
    match = 0;
    // Search the DRAM TAD Entries for a match
    for (tad_index = 0; tad_index < _TAD; tad_index++) {
        if (match == 1)
            break;

        if (ti->haTadDict[tad_index] != -1)
            tadReg = ti->haTadDict[tad_index];
        else {
            pdev = g_pdev_list[ENTRY_ID_HA0];
            regOffset = REG_OFFSET_TAD_0 + 4 * tad_index;

            if (0 != pci_read_config_dword(pdev, regOffset, &tadReg)) {
                printk("[mem_drv] forwardTad: Read PCI(%02x:%02x.%x) Reg 0x%x fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn), regOffset);
                return ADDR_TRANSLATION_FAIL;
            } else {
                ti->haTadDict[tad_index] = tadReg;
            }
        }

        limit = ((unsigned long long)((tadReg >> BIT_OFFSET_LIMIT) + 1) << _TAD_OFFSET) - 1;

        // Are we passed the last valid entry?
        if (limit == previous_limit)
            break;

        // Expecting all values to be 46-bit numbers
        if ((previous_limit < (long long)ti->core_addr) && (ti->core_addr <= limit)) {
                match = 1;

                if (ti->mcTadDict[tad_index] != -1)
                    imcTadReg = ti->mcTadDict[tad_index];
                else {
                    pdev = g_pdev_list[ENTRY_ID_TA];
                    regOffset = REG_OFFSET_TADWAYNESS_0 + 4 * tad_index;

                    if (0 != pci_read_config_dword(pdev, regOffset, &imcTadReg)) {
                        printk("[mem_drv] forwardTad: Read PCI(%02x:%02x.%x) Reg 0x%x fail\n", pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn), regOffset);
                        return ADDR_TRANSLATION_FAIL;
                    } else {
                        ti->mcTadDict[tad_index] = imcTadReg;
                    }
                }

            sw = 1 << ((imcTadReg >> BIT_OFFSET_TAD_SKT_WAY) & FIELD_MASK_2BITS);
            cw = ((tadReg >> BIT_OFFSET_CHANNELWAYS) & FIELD_MASK_2BITS) + 1;
            imcCw = ((imcTadReg >> BIT_OFFSET_TAD_CH_WAY) & FIELD_MASK_2BITS) + 1;
            // Determine how many bits are shifted out of address
            swXcw = sw * imcCw;

            // Determine which interleave the system address falls into
            chInterleave = (ti->core_addr >> (splitBit + ((imcTadReg >> BIT_OFFSET_TAD_SKT_WAY) & FIELD_MASK_2BITS))) & 0x3;

            if ((haD2 >> BIT_OFFSET_MC_CHAN_SHIFT_UP_ENABLE) & 0x1) {
                // Hash_ChannelID[0] = ChannelID[0] ^ (A12 ^ A14 ^ A16 ^ A18 ^ A20 ^ A22 ^ A24 ^ A26) AND CSR_hash_enable
                // Hash_ChannelID[1] = ChannelID[1] ^ (A13 ^ A15 ^ A17 ^ A19 ^ A21 ^ A23 ^ A25 ^ A27) AND CSR_hash_enable
                for (i = 0; i < 8; i++) {
                    tmp = (unsigned int)((ti->core_addr >> chXorBitPos[i]) & 0x3);
                    chInterleave ^= tmp;
                }
            }

            chInterleave %= cw;

            // Determine which channel this interleave maps to using Imc Tad
            if (chInterleave == 0)
                ti->channel = (imcTadReg >> BIT_OFFSET_TAD_CH_TGT_0) & FIELD_MASK_2BITS;
            else
                ti->channel = (imcTadReg >> BIT_OFFSET_TAD_CH_TGT_1) & FIELD_MASK_2BITS;
        }

        previous_limit = limit;
    }

    if (match == 0) {
        printk("[mem_drv] forwardTad: No entry in TAD matched to given system address 0x%llx\n", ti->core_addr);
        return ADDR_TRANSLATION_FAIL;
    }

    return NO_ERROR;
}

static int coreAddressToChannel(struct addr_tranlate_info *ti)
{
    return forwardTad(ti);
}

static int ecc_injection_init(void)
{
    int ret = NO_ERROR;
    unsigned long flags = 0;

    // Get MC7 control
    rdmsr_safe(0x287, &g_save_msr_low, &g_save_msr_high);

    // Disable CMCI interrupt
    wrmsr_safe(0x287, g_save_msr_low & 0xBFFFFFFF, g_save_msr_high);

    // Disable memory scrub
    g_save_demand_scrub = false;
    if (g_enable_demand_scrub) {
        g_save_demand_scrub = true;
        ret = set_demand_scrub(false);
    }

    return ret;
}

static int ecc_injection_restore(void)
{
    int ret = NO_ERROR;
    int channel, reg;

    // Enable memory scrub
    ret = set_demand_scrub(g_save_demand_scrub);

    // Restore MC7 control
    wrmsr_safe(0x287, g_save_msr_low, g_save_msr_high);

    // Clear MC5 Shadow
    pci_write_config_dword(g_pdev_list[ENTRY_ID_HA0_PERFMON],
                           REG_OFFSET_MC5_STATUS_SHADOW_LOW,
                           0);
    pci_write_config_dword(g_pdev_list[ENTRY_ID_HA0_PERFMON],
                           REG_OFFSET_MC5_STATUS_SHADOW_HIGH,
                           0);

    // Clear MC7 status
    wrmsr_safe(0x41D, 0, 0);
    wrmsr_safe(0x41E, 0, 0);

    return ret;
}

static int ecc_injection(struct ecc_inj_info inj_info)
{
    int ret = NO_ERROR;
    unsigned long flags = 0;
    unsigned long long curr_addr;
    unsigned long long end_addr;
    u8 *map_addr = NULL;
    u8 *curr_map_addr = NULL;
    u8 *temp_addr;
    struct addr_tranlate_info ti;
    unsigned long long addrmask = 0x3FULL;
    unsigned int orig_error_count = 0;
    unsigned int value;
    unsigned int temp_error_count;
    unsigned long long temp_value;
    int i;

    // We can not test memory address which is lower than 4G
    if (inj_info.address < 0x100000000)
        return ADDR_CAN_NOT_BE_TESTED;

    printk(KERN_DEBUG "[mem_drv] --- ecc_injection start ---\n");

    ecc_injection_init();

    // Start to inject errors
    // Configure for Channel 0
    pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                           REG_OFFSET_RSP_FUNC_ADDR_MASK_HI,
                           (u32)(addrmask >> 35));
    pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                           REG_OFFSET_RSP_FUNC_ADDR_MASK_LO,
                           (u32)(addrmask >> 3));

    *((unsigned int *)(g_map_addr[0] + REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_EXTRA)) = 0x3C000302;
    *((unsigned int *)(g_map_addr[0] + REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_DEV0_XOR_MSK)) = 0xf;
    *((unsigned int *)(g_map_addr[0] + REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_DEV1_XOR_MSK)) = 0x0;

    // Configure for Channel 1
    pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                           REG_OFFSET_RSP_FUNC_ADDR_MASK_HI,
                           (u32)(addrmask >> 35));
    pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                           REG_OFFSET_RSP_FUNC_ADDR_MASK_LO,
                           (u32)(addrmask >> 3));

    *((unsigned int *)(g_map_addr[1] + REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_EXTRA)) = 0x3C000302;
    *((unsigned int *)(g_map_addr[1] + REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_DEV0_XOR_MSK)) = 0xf;
    *((unsigned int *)(g_map_addr[1] + REG_OFFSET_RSP_FUNC_CRC_ERR_INJ_DEV1_XOR_MSK)) = 0x0;

    spin_lock_irqsave(&g_mem_drv_lock, flags);

    // Get error count before error injection
    pci_read_config_dword(g_pdev_list[ENTRY_ID_HA0_PERFMON],
                          REG_OFFSET_MC5_STATUS_SHADOW_HIGH,
                          &value);

    orig_error_count = (value >> 6) & 0x7FFF;

    map_addr = (unsigned char *)ioremap_nocache(inj_info.address, inj_info.size);
    if (map_addr == NULL) {
        ret = MEMORY_MAP_FAIL;
        goto fail;
    }

    inj_info.elem_size >>= 3;
    curr_map_addr = map_addr;
    curr_addr = inj_info.address;
    end_addr = inj_info.address + inj_info.size;

    switch (inj_info.elem_size) {
        case 1:
            inj_info.pattern &= 0xFFULL;
            break;
        case 2:
            inj_info.pattern &= 0xFFFFULL;
            break;
        case 4:
            inj_info.pattern &= 0xFFFFFFFFULL;
            break;
        case 8:
        default:
            break;
    }

    while (curr_addr < end_addr) {

        memset(&ti, -1, sizeof(ti));
        ti.core_addr = curr_addr;

        // Find correct channel mapping to testing address
        if (0 != (ret = coreAddressToChannel(&ti))) {
            printk(KERN_DEBUG "[mem_drv] Address translation is failed\n");
            goto fail;
        } else {
            //printk(KERN_DEBUG "[mem_drv] Attempt inject error to channel %d\n", ti.channel);
        }

        //printk(KERN_DEBUG "[mem_drv] original error count: %d\n", orig_error_count);

        temp_addr = curr_map_addr;

        if (ti.channel == 0) {
            pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                                   REG_OFFSET_RSP_FUNC_ADDR_MATCH_LO,
                                   (u32)(curr_addr >> 3));

            pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                                   REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                   (u32)(curr_addr >> 35));

            // Enable error inject
            pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                                   REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                   ((u32)(curr_addr >> 35) | (0x1 << BIT_OFFSET_RSP_FUNC_ADDR_MATCH_EN)));

            pci_read_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                                  REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                  &value);
        } else {
            pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                                   REG_OFFSET_RSP_FUNC_ADDR_MATCH_LO,
                                   (u32)(curr_addr >> 3));

            pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                                   REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                   (u32)(curr_addr >> 35));
            // Enable error inject
            pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                                   REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                   ((u32)(curr_addr >> 35) | (0x1 << BIT_OFFSET_RSP_FUNC_ADDR_MATCH_EN)));

            pci_read_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                                  REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                  &value);
        }

        if (!(value >> BIT_OFFSET_RSP_FUNC_ADDR_MATCH_EN) & 0x1) {
            printk(KERN_DEBUG "[mem_drv] error inject was not armed!\n");
            ret = ENABLE_ECC_INJECT_FAIL;
            goto fail;
        }

        // Write memory to inject error
        switch (inj_info.elem_size) {
            case 1:
                *curr_map_addr = (u8)inj_info.pattern;
                break;
            case 2:
                *((u16 *)curr_map_addr) = (u16)inj_info.pattern;
                break;
            case 4:
                *((u32 *)curr_map_addr) = (u32)inj_info.pattern;
                break;
            case 8:
            default:
                *((u64 *)curr_map_addr) = inj_info.pattern;
                break;
        }

        // Clear cache
        asm volatile ("wbinvd");

        if (ti.channel == 0) {
            pci_read_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                                  REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                  &value);
        } else {
            pci_read_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                                  REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                  &value);
        }

        if (!((value >> BIT_OFFSET_RSP_FUNC_ADDR_MATCH_EN) & 0x1)) {
            //printk(KERN_DEBUG "[mem_drv] Error is injected successfully\n");

            // Read memory to consume error
            switch (inj_info.elem_size) {
                case 1:
                    temp_value = (unsigned long long)(*curr_map_addr);
                    break;
                case 2:
                    temp_value = (unsigned long long)(*((u16 *)curr_map_addr));
                    break;
                case 4:
                    temp_value = (unsigned long long)(*((u32 *)curr_map_addr));
                    break;
                case 8:
                default:
                    temp_value = (unsigned long long)(*((u64 *)curr_map_addr));
                    break;
            }

            // Clear cache
            asm volatile("clflush %0" : "+m" (*(volatile char __force *)temp_addr));

            if (temp_value != inj_info.pattern) {
                printk(KERN_DEBUG "[mem_drv] Data comparison is failed: expected 0x%016llx current 0x%016llx!\n",
                        inj_info.pattern,
                        temp_value);
                ret = COMPARE_DATA_FAIL;
            }

            if (ret == NO_ERROR) {
                // Get error count before error injection
                pci_read_config_dword(g_pdev_list[ENTRY_ID_HA0_PERFMON],
                                      REG_OFFSET_MC5_STATUS_SHADOW_HIGH,
                                      &value);

                temp_error_count = (value >> 6) & 0x7FFF;

                //printk(KERN_DEBUG "[mem_drv] Error count after injection: %d\n", temp_error_count);

                temp_error_count -= orig_error_count;

                if (temp_error_count != 1) {
                    if (temp_error_count == 0)
                        ret = ECC_INJECT_FAIL;
                    else
                        ret = INCORRECT_ECC_COUNT;
                }
            }

            // Clear dirty data
            switch (inj_info.elem_size) {
                case 1:
                    *curr_map_addr = (u8)0;
                    break;
                case 2:
                    *((u16 *)curr_map_addr) = (u16)0;
                    break;
                case 4:
                    *((u32 *)curr_map_addr) = (u32)0;
                    break;
                case 8:
                default:
                    *((u64 *)curr_map_addr) = 0ULL;
                    break;
            }

            // Clear cache
            asm volatile ("wbinvd");

            //printk(KERN_DEBUG "[mem_drv] Clear dirty data\n");

            if (orig_error_count == (0x7FFF - 1)) {
                pci_write_config_dword(g_pdev_list[ENTRY_ID_HA0_PERFMON],
                                       REG_OFFSET_MC5_STATUS_SHADOW_LOW,
                                       0);
                pci_write_config_dword(g_pdev_list[ENTRY_ID_HA0_PERFMON],
                                       REG_OFFSET_MC5_STATUS_SHADOW_HIGH,
                                       0);

                orig_error_count = 0;

            } else {
                orig_error_count++;
            }

            // Clear MC7 status
            wrmsr_safe(0x41D, 0, 0);

            update_error_count(ti.channel);

            if (ret != NO_ERROR) {
                goto fail;
            }
        } else {
            if (ti.channel == 0) {
                pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0],
                                       REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                       0);
            } else {
                pci_write_config_dword(g_pdev_list[ENTRY_ID_IMC0_TAD0 + 1],
                                       REG_OFFSET_RSP_FUNC_ADDR_MATCH_HI,
                                       0);
            }

            ret = ECC_INJECT_FAIL;
            printk(KERN_DEBUG "[mem_drv] error inject is failed!\n");
            goto fail;
        }

        curr_addr += inj_info.elem_size;
        curr_map_addr += inj_info.elem_size;
    }
    
fail:
    printk(KERN_DEBUG "[mem_drv] --- ecc_injection end ---\n");

    if (map_addr) {
        iounmap(map_addr);
        map_addr = NULL;
    }

    spin_unlock_irqrestore(&g_mem_drv_lock, flags);

    ecc_injection_restore();
    
    return ret;
}

static long mem_drv_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    int ret = NO_ERROR;
    struct get_err_info get_info;
    int i = 0;
    struct corr_err_info *perr_info = NULL;
    struct mem_access_info access_info;
    void *map_addr = NULL;
    struct ecc_inj_info ecc_inj_info;
    unsigned int msr_low, msr_high;
    
    if ((_IOC_TYPE(cmd) & MC_DRV_MAGIC_NUMBER) != MC_DRV_MAGIC_NUMBER) {
        ret = INCORRECT_MAGIC_NUM;
        goto out;
    }


    switch (_IOC_NR(cmd)) {
        case MC_DRV_IOCTL_ENABLE_SCRUB:
            ret = set_demand_scrub(true);     
            break;
        case MC_DRV_IOCTL_DISABLE_SCRUB:
            ret = set_demand_scrub(false); 
            break;
        case MC_DRV_IOCTL_GET_SCRUB_STATUS:
            i = g_enable_demand_scrub? 1 : 0;
            ret = copy_to_user((int *)arg, &i, sizeof(int));
            break;
        case MC_DRV_IOCTL_CLEAR_ERROR:
            clear_err_list();
            break;
        case MC_DRV_IOCTL_GET_ERROR_NUM:
            ret = copy_to_user((u32 *)arg, &g_error_count, sizeof(g_error_count));
            break;
        case MC_DRV_IOCTL_GET_ERROR_ENTRY_NUM:
            ret = copy_to_user((u32 *)arg, &g_error_entries, sizeof(g_error_entries));
            break;
        case MC_DRV_IOCTL_GET_ERROR_INFO:

            ret = copy_from_user(&get_info, (struct get_err_info *)arg, sizeof(get_info));

            if (ret == 0) {
                perr_info = g_corr_err_head;
            
                while (perr_info) {
                    if (i++ == get_info.index)
                        break;
                    perr_info = perr_info->next;
                }

                if (perr_info) {                 
                    get_info.info = *perr_info;
                    ret = copy_to_user((struct get_err_info *)arg, &get_info, sizeof(get_info));
                }
                else
                    ret = ERROR_ENTRY_NOT_FOUND;
            }
            break;
        case MC_DRV_IOCTL_READ_MEMORY:
            ret = copy_from_user(&access_info, (struct mem_access_info *)arg, sizeof(access_info));
            if (ret == 0) {
                map_addr = (void *)ioremap_nocache(access_info.addr, access_info.size);
                if (map_addr == NULL) {
                    ret = MEMORY_MAP_FAIL;
                    return ret;
                }

                access_info.data = 0ULL;

                switch (access_info.size) {
                    case 1:
                        access_info.data = *((unsigned char *)map_addr);
                        break;
                    case 2:
                        access_info.data = *((unsigned short *)map_addr);
                        break;
                    case 4:
                        access_info.data = *((unsigned int *)map_addr);
                        break;
                    case 8:
                        access_info.data = *((unsigned long long *)map_addr);
                        break;
                    default:
                        break;
                }

                iounmap(map_addr);

                ret = copy_to_user((struct mem_access_info *)arg, &access_info, sizeof(access_info));
            }
            break;
        case MC_DRV_IOCTL_CMD_ECC_INJECT:
            g_ecc_enable = false;

            rdmsr_safe(0x790, &msr_low, &msr_high);
            if (!msr_low)
                g_ecc_enable = true;

            if (g_ecc_enable) {                
                ret = copy_from_user(&ecc_inj_info, (struct ecc_inj_info *)arg, sizeof(ecc_inj_info));
        
                if (ret == 0) {
                    ret = ecc_injection(ecc_inj_info);
                    ecc_inj_info.error_code = ret;

                    ret = copy_to_user((struct ecc_inj_info *)arg, &ecc_inj_info, sizeof(ecc_inj_info));
                }
            } else {
                ecc_inj_info.error_code = ECC_INJECT_NOT_SUPPORT;
                ret = copy_to_user((struct ecc_inj_info *)arg, &ecc_inj_info, sizeof(ecc_inj_info));
            }
            break;
        default:
            ret = UNKNOWN_IOCTL_CMD;
            break;
    }
out:
    return ret;
}

static struct file_operations g_mem_dev_file_ops = {
    .owner          = THIS_MODULE,
    .unlocked_ioctl = mem_drv_ioctl,
    .open           = mem_drv_open,
    .release        = mem_drv_close
};

static int __init mem_drv_init(void)
{
    int ret;
    struct pci_dev *pdev = NULL;
    unsigned int dev_id;    
    int i;
    
    for (i = 0; i < MAX_PCI_DEV_NUM_FOR_INTEL_BROADWELL_IMC; i++)
        g_pdev_list[i] = NULL;

    for (i = 0; i < MAX_CHANNEL_NUM; i++)
        g_map_addr[i] = NULL;

    // Get SAD0
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_SAD0;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))         
        goto pci_fail;
    g_pdev_list[ENTRY_ID_SAD0] = pdev;    

    // Get HA0
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_HA0;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))         
        goto pci_fail;
    g_pdev_list[ENTRY_ID_HA0] = pdev;    

    // Get HA0 Perform
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_HA0_PERFMON;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))         
        goto pci_fail;
    g_pdev_list[ENTRY_ID_HA0_PERFMON] = pdev; 

    // Get TA
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_TA;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))
        goto pci_fail;
    g_pdev_list[ENTRY_ID_TA] = pdev;

    // Get RAS
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_RAS;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))         
        goto pci_fail;
    g_pdev_list[ENTRY_ID_IMC0_RAS] = pdev; 

    // Workaround: In some BIOS, Error Register do not exist
    // Get TAD0
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_TAD0;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))         
        goto pci_fail;
    g_pdev_list[ENTRY_ID_IMC0_TAD0] = pdev;    

    // Get TAD1
    dev_id = PCI_DEVICE_ID_INTEL_BROADWELL_IMC0_TAD1;
    if (NULL == (pdev = pci_get_device(PCI_VENDOR_ID_INTEL, dev_id, pdev)))         
        goto pci_fail;
    g_pdev_list[ENTRY_ID_IMC0_TAD1] = pdev;   

    // Get ERR0
    g_map_addr[0] = (unsigned char *)ioremap_nocache(PCI_DEVICE_BASE_ADDRESS_IMC0_ERR0, 0x200);
    if (g_map_addr[0] == NULL) {
        printk("[mem_drv] check_memory_error_data ioremap for 0x%x is failed!\n", PCI_DEVICE_BASE_ADDRESS_IMC0_ERR0);
        return MEMORY_MAP_FAIL;
    }

    // Get ERR1
    g_map_addr[1] = (unsigned char *)ioremap_nocache(PCI_DEVICE_BASE_ADDRESS_IMC0_ERR1, 0x200);
    if (g_map_addr[1] == NULL) {
        iounmap(g_map_addr[0]);
        printk("[mem_drv] check_memory_error_data ioremap for 0x%x is failed\n", PCI_DEVICE_BASE_ADDRESS_IMC0_ERR1);
        return MEMORY_MAP_FAIL;
    }

    // Initialize error count status
    update_error_count(2);

    // Disable demand scrub
    if (NO_ERROR != (ret = set_demand_scrub(true)))
        return ret;
    
    // Create char device
    if ((g_major_num = register_chrdev(0, MC_DEV_NAME, &g_mem_dev_file_ops)) < 0)
    {
        printk("[mem_drv] Register char device fail!\n");
        return REGISTER_CHAR_DEV_FAIL;
    }

    if (NULL == (g_pmem_drv_class = class_create(THIS_MODULE, MC_DEV_NAME)))
        return CREATE_CHAR_DEV_FAIL;
    device_create(g_pmem_drv_class, NULL, MKDEV(g_major_num, 0), NULL, "%s", MC_DEV_NAME);

    // Register MCE notifier
//    mce_register_decode_chain(&g_mem_mce_dec);

    clear_err_list();
    
    printk("[mem_drv] Memory Controller Driver is inserted correctly.\n");

    return NO_ERROR;
pci_fail:
    printk("[mem_drv] Get PCI device (%x:%x) fail!\n", PCI_VENDOR_ID_INTEL, dev_id);
    return GET_PCI_DEV_FAIL;
}

static void __exit mem_drv_exit(void)
{
    int i;
    unsigned long flags = 0;

    spin_lock_irqsave(&g_mem_drv_lock, flags);

//    mce_unregister_decode_chain(&g_mem_mce_dec);

    spin_unlock_irqrestore(&g_mem_drv_lock, flags);

    for (i = 0; i < MAX_CHANNEL_NUM; i++) {
        if (g_map_addr[i]) {
            iounmap(g_map_addr[i]);
            g_map_addr[i] = NULL;
        }
    }
    
    device_destroy(g_pmem_drv_class, MKDEV(g_major_num, 0));

    class_destroy(g_pmem_drv_class);    

    unregister_chrdev(g_major_num, MC_DEV_NAME);
    // Make sure that allocated buffers are removed
    clear_err_list();
       
    printk("[mem_drv] Memory Controller Driver is removed correctly.\n");
}

module_init(mem_drv_init);
module_exit(mem_drv_exit);


