/// @file       NVMeDAPI.cpp
/// @brief      NVMeDAPI class
/// @details    Provide diagnostic and utility APIs for NVMe module
/// @version    1.0.0
/// @date       2014.01.06
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "NVMeDAPI.hpp"
#include "MemoryDDAPI.hpp"
#include "CommonFunctions.hpp"
#include <string.h>
#include <iomanip>

/// @fn         NVMeDAPI::NVMeDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      NVMeDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
NVMeDAPI::NVMeDAPI(STRING className) : BaseClassDAPI(className)
{
    // Initializations.
    if((nvmeDDAPI = new NVMeDDAPI("NVMe_DRVR")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_ERRORCODE | ERROR_DIAGNOSTIC_API_CONSTRUCTOR2_ERROR), "Create NVMe DDAPI fail");
    }

    dataBuffer = (UINT32*)NULL;
    logicalSectorToTest = 0;
    rwBlocks = 0;
    testSize = 0;
    transferSize = 0;
    throttle = 0;
    testPattern = 0;
    testSpeed = 0;
};

/// @fn         NVMeDAPI::~NVMeDAPI(void)
/// @brief      NVMeDAPI destructor
/// @exception  Nothing
/// @return     Nothing
NVMeDAPI::~NVMeDAPI(void)
{
    // delete NVMe Driver API
    if(nvmeDDAPI != NULL)
        delete nvmeDDAPI;
};

/// @fn         activateFirmwareByController(UINT8 controllerID, nvme_fw_active active)
/// @brief      call driver ioctl to activate downloaded firmware
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::activateFirmwareByController(UINT8 controllerID, nvme_fw_active active)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::FIRMWARE_ACTIVATE, &active);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_ERROR_LOG), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         downloadFirmwareByController(UINT8 controllerID, const char *fw)
/// @brief      call driver ioctl to download firmware
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::downloadFirmwareByController(UINT8 controllerID, const char *fw)
{
    ErrorMsg* errorMsg = NULL;
    FILE* file = NULL;
    INT32 fileSize = 0;
    MemoryDDAPI memoryDDAPI("MemoryDDAPI");
    SYSTEM_ADDRESS_T fwBuffer;

    if((errorMsg = memoryDDAPI.getErrorMsg()) != NO_ERROR_MSG)
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, errorMsg);

    if ((file = fopen(fw, "r")) == NULL)
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_FW_DOWNLOAD | ERROR_OPEN_FILE), fw);
    // Allocate a memory buffer for FW
    fseek(file, 0, SEEK_END);
    fileSize = ftell(file);
    fseek(file, 0, SEEK_SET);

    if((errorMsg = memoryDDAPI.getMemoryBuffer(fileSize, &fwBuffer)) != NO_ERROR_MSG)
    {
        fclose(file);
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, errorMsg);
    }

    if (fileSize != (INT32)fread((CHAR *)fwBuffer, 1, fileSize, file))
    {
        fclose(file);
        memoryDDAPI.releaseMemoryBuffer(fwBuffer);
        return returnErrorMsg((ERRORCODE)(ERROR_NVME_DAPI_FW_DOWNLOAD | ERROR_READ_FILE_FAIL), fw);
    }

    fclose(file);

    errorMsg = nvmeDDAPI->downloadFirmware(controllerID, (void *)fwBuffer, fileSize);
    if (errorMsg != NO_ERROR_MSG)
    {
        memoryDDAPI.releaseMemoryBuffer(fwBuffer);
        return returnErrorMsg(ERROR_NVME_DAPI_FW_DOWNLOAD, NULL, errorMsg);
    }

    memoryDDAPI.releaseMemoryBuffer(fwBuffer);

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::doIOCTL
/// @brief      Execute specific IOCLT command
/// @param      Specific NVME controller
/// @param      NVMe ioctl operation code
/// @param      The returned value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::doIOCTL(UINT8 controllerID, UINT32 ioctlOperation, UINT32* returnedValue)
{
    ErrorMsg* errorMsg = NULL;

    // only support no argument input ioctl OP code
    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, (NVMeIOCTL::IOCTL_OP_CODE)ioctlOperation, returnedValue);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_DO_IOCTL), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getBarResourceByController(UINT8 controllerID, nvme_bar **barInfo)
/// @brief      call driver ioctl to get controller's bar resource
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getBarResourceByController(UINT8 controllerID, nvme_bar **barInfo)
{
    int pci_fd;
    char path[512];
    NVMe_DEVICE_INFO nvmeInfo;
    string tempName;
    ErrorMsg* errorMsg = NULL;

    errorMsg = getInformationByController(controllerID, &nvmeInfo);
    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_BAR_RESOURCE), NULL, errorMsg);
    }
    tempName = nvmeInfo.wellKnowName;
    transform(tempName.begin(), tempName.end(), tempName.begin(), ::tolower);
    sprintf(path, "/sys/class/nvme/%s/device/resource0", tempName.c_str());

    pci_fd = open(path, O_RDONLY);
    if (pci_fd < 0)
    {
        sprintf(path, "/sys/class/misc/%s/device/resource0", nvmeInfo.chardeviceNodeName.c_str());
        pci_fd = open(path, O_RDONLY);
    }
    if (pci_fd < 0)
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_BAR_RESOURCE), "can not find pci resource", NULL);

    *barInfo = (nvme_bar *)mmap(0, getpagesize(), PROT_READ, MAP_SHARED, pci_fd, 0);
    if (*barInfo == MAP_FAILED)
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_BAR_RESOURCE | ERROR_NVME_DDAPI_MMAP_FAIL), NULL, NULL);

    return NO_ERROR_MSG;
}

/// @fn         getErrorLogByController(UINT8 controllerID, nvme_error_log_page errLog)
/// @brief      call driver ioctl to get controller's error log
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getErrorLogByController(UINT8 controllerID, nvme_error_log_page* errLog)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::ERROR_LOG, errLog);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_ERROR_LOG), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getFirmwareLogByController(UINT8 controllerID, nvme_firmware_log_page* fwLog)
/// @brief      call driver ioctl to get controller's firmware log
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getFirmwareLogByController(UINT8 controllerID, nvme_firmware_log_page* fwLog)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::FIRMWARE_LOG, fwLog);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_FW_LOG), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::getInformationByController(UINT8 controllerID)
/// @brief      call driver ioctl to get NVMe controller basic informations
/// @param      NVME ID
/// @param      pointer to return NVMe Info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getInformationByController(UINT8 controllerID, NVMeInfo* returnInfo)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->getControllerList(&nvmeList);

    if(nvmeList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_INFORMATION), NULL, errorMsg);
    }
    if(controllerID > nvmeList.size())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_INFORMATION | ERROR_BAD_PARAMETER),
                              "Bad NVME logical ID");
    }
    for(iter = nvmeList.begin(); iter != nvmeList.end(); iter++)
    {
        if (controllerID == (iter - nvmeList.begin()))
        {
            *returnInfo = *iter;
            return NO_ERROR_MSG;
        }
    }

    return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_INFORMATION), NULL, errorMsg);
}

/// @fn         NVMeDAPI::getInformationList(vector<NVMe_DEVICE_INFO>* returnedList)
/// @brief      Get NVMe information list
/// @param      pointer to NVMe information list on the host
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getInformationList(vector<NVMe_DEVICE_INFO>* returnList)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->getControllerList(returnList);

    if(returnList->empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_INFORMATION), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getIdentityCTRLByController(UINT8 controllerID, nvme_id_ctrl* idCtrl)
/// @brief      call driver ioctl to get controller's information
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getIdentityCTRLByController(UINT8 controllerID, nvme_id_ctrl* idCtrl)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::CONTROLLER_INFORMATION, idCtrl);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_IDENTITY_CONTROLLER), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getIdentityNSByController(UINT8 controllerID, nvme_id_ns* idNS)
/// @brief      call driver ioctl to get controller's namespace information
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getIdentityNSByController(UINT8 controllerID, nvme_id_ns* idNS)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::NAMESPACE_INFORMATION, idNS);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_IDENTITY_NAMESPACE), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::getLogicalIDByName(string wellKnowName, UINT8* returnID)
/// @brief      Get NVMe controller logical ID
/// @param      Controller well known name
/// @param      pointer to NVMe controller logical ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getLogicalIDByName(string wellKnowName, UINT8* returnID)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->getControllerList(&nvmeList);

    if(nvmeList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_LOGICALID_BY_NAME), NULL, errorMsg);
    }

    ConvertStringToUpperCase((char*)wellKnowName.c_str());

    for(iter = nvmeList.begin(); iter != nvmeList.end(); iter++)
    {
        if (wellKnowName == iter->wellKnowName)
        {
            *returnID = (iter - nvmeList.begin());
            return NO_ERROR_MSG;
        }
    }

    return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_LOGICALID_BY_NAME | ERROR_BAD_PARAMETER_FIVE),
                                    NULL, NULL);
}

/// @fn         getNameSpaceIDByController(UINT8 controllerID, int* returnValue)
/// @brief      call driver ioctl to get namespace ID
/// @param      NVME ID
/// @param      pointer to return ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getNameSpaceIDByController(UINT8 controllerID, int* returnValue)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::NAMESPACE_ID, (void *)returnValue);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_NAMESPACE_ID), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getSmartLogByController(UINT8 controllerID, nvme_smart_log* smartLog)
/// @brief      call driver ioctl to get smart log
/// @param      NVME ID
/// @param      pointer to return smart log structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::getSmartLogByController(UINT8 controllerID, nvme_smart_log* smartLog)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::SMART_LOG, smartLog);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_GET_SMART_LOG), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         NVMeDAPI::resetIOC(UINT8 controllerID)
/// @brief      reset specific NVME controller
/// @param      NVME ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::resetNVMe(UINT8 controllerID)
{
    ErrorMsg* errorMsg = NULL;

    errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::EXECUTE_NVME_RESET, NULL);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_RESET), NULL, errorMsg);
    }
    
    return NO_ERROR_MSG;
}

ErrorMsg*
NVMeDAPI::readFromDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

ErrorMsg*
NVMeDAPI::writeToDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

/// @fn         NVMeDAPI::writeZeroByController(UINT8 controllerID, UINT64 startBlock, UINT16 blockCount);
/// @brief      Write zero to specified NVMe block
/// @param      Specific NVME controller
/// @param      start block to write
/// @param      block count to write zero
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVMeDAPI::writeZeroByController(UINT8 controllerID, UINT64 startBlock, UINT16 blockCount)
{
    ErrorMsg* errorMsg = NULL;
    nvme_write_zero_parms parms;

    parms.startBlock = startBlock;
    parms.blockCount = blockCount;

    if((errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::NAMESPACE_ID, (void *)&parms.nsid)) == NO_ERROR_MSG)
    {
        errorMsg = nvmeDDAPI->doIOCTLCommand(controllerID, NVMeIOCTL::WRITE_ZERO, &parms);
    }

    if(errorMsg != NO_ERROR_MSG)
        return returnErrorMsg(ERRORCODE(ERROR_NVME_DAPI_DO_IOCTL), NULL, errorMsg);

    return NO_ERROR_MSG;
}
