/// @file       SATADAPI.cpp
/// @brief      SATADAPI className
/// @details    Provide diagnostic and utility APIs for SATA module
/// @version    1.0.0
/// @date       2015.12.01
/// @author     Ben Yang
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.
#include "SATADAPI.hpp"
using namespace::std;
/// @fn         SATADAPI::SATADAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      SATADAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
SATADAPI::SATADAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
{
     ErrorMsg *pErrorMsg = NULL;
     m_pSATADDAPI = NULL;
     m_strOutputResultString = "";

     m_pSATADDAPI = new SATADDAPI("SATADDAPI");
     if(m_pSATADDAPI == NULL)
     {
	  pErrorMsg = m_pSATADDAPI->getErrorMsg();
     }

     if(pErrorMsg != NULL)
     {
	  addErrorMsg(ERROR_SATA_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
     }
}

/// @fn         SATADAPI::~SATADAPI(void)
/// @brief      SATADAPI destructor
/// @exception  Nothing
/// @return     Nothing
SATADAPI::~SATADAPI()
{
     m_strOutputResultString = "";

     if(m_pSATADDAPI != NULL)
     {
	  delete m_pSATADDAPI;
     }
}

/// @fn         ErrorMsg *SATADAPI::IOCTL(UINT32 in_u32LogicalID, SATAIOCTLOperation in_eOperation)
/// @brief      Function for abstraction of DAPI IOCTL operation
/// @param      in_u32LogicalID is a logicalID for SATA device node name (eg. 1)
/// @param      in_eOperation desired driver opertaion
/// @param      inout_pXMLData is a pointer to XML data for SATA information
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADAPI::IOCTL(UINT32 in_u32LogicalID, SATAIOCTLOperation in_eOperation)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if(in_eOperation != eGenerateSATAList)
     {
	  if((pErrorMsg = generateSATAList()) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
     }
     switch(in_eOperation)
     {
     case eGenerateSATAList:
	  if((pErrorMsg = m_pSATADDAPI->IOCTL(0, eGenerateSATAList)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;
     case eListController:
	  if((pErrorMsg = m_pSATADDAPI->IOCTL(0, eListController)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;
     case eListPort:
	  if((pErrorMsg = m_pSATADDAPI->IOCTL(0, eListPort)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;
     case eListDrive:
	  if((pErrorMsg = m_pSATADDAPI->IOCTL(0, eListDrive)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;

     case eFlushBufferCache:
	  if((pErrorMsg = m_pSATADDAPI->IOCTL(in_u32LogicalID, eFlushBufferCache)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;

     case eSATIdentifyCommandTest:
	  if((pErrorMsg = m_pSATADDAPI->IOCTL(in_u32LogicalID, eSATIdentifyCommandTest)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;

     default:
	  errorCode = ERROR_SATA_DAPI_NO_SUCH_IOCTL_OPERATION;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_IOCTL | errorCode), NULL, NULL);
	  break;
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *SATADAPI::generateSATAList()
/// @brief      generate SATA devie list
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADAPI::generateSATAList()
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NULL;
     if((pErrorMsg = m_pSATADDAPI->generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_GENERATE_SATA_LIST, NULL, pErrorMsg);
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::readBySCSI(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32SectorCount)
{
     ERRORCODE errorCode = NO_ERROR;
//     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::writeBySCSI(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32SectorCount)
{
     ERRORCODE errorCode = NO_ERROR;
//     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::getLogicalIdByName(string in_strWellKnownName, UINT32 *out_u32LogicalID)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
     }

     if((pErrorMsg = m_pSATADDAPI->getLogicalIdByName(in_strWellKnownName, out_u32LogicalID)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_GET_LOGICALID_BY_NAME, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::getDataByID(UINT32 in_u32LogicalID, SATADataElement in_eSATADataElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = m_pSATADDAPI->getDataByID(in_u32LogicalID, in_eSATADataElement, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_GET_DATA_BY_ID, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::getErrorRegisterByID(UINT32 in_u32LogicalID, SATAErrorElement in_eSATAErrorElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = m_pSATADDAPI->getErrorRegisterByID(in_u32LogicalID, in_eSATAErrorElement, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_GET_ERROR_REGISTER_BY_ID, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::getSMARTInfoByID(UINT32 in_u32LogicalID, UINT32 in_u32StartIDNumber, UINT32 in_u32EndIDNumber, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = m_pSATADDAPI->getSMARTInfoByID(in_u32LogicalID, in_u32StartIDNumber, in_u32EndIDNumber, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_GET_SMART_INFO_BY_ID, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::getDataByName(string in_strWellKnownName, SATADataElement in_eSATADataElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
     }

     if((pErrorMsg = m_pSATADDAPI->getDataByName(in_strWellKnownName, in_eSATADataElement, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_GET_DATA_BY_NAME, NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADAPI::identifyCommandTest(UINT32 in_u32LogicalID)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     if((pErrorMsg = m_pSATADDAPI->generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_IOCTL, NULL, pErrorMsg);
     }
     if((pErrorMsg = m_pSATADDAPI->identifyCommandTest(in_u32LogicalID)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_IDENTIFY_COMMAND_TEST , NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *SATADAPI::readPerformanceTest(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32TestSize)
/// @brief      Read specific size(MB) of data from starting logical sector on hard disk through logicalID to get read performance (MB/s)
/// @param      in_u32LogicalID is a logicalID for SATA device node name (eg. 1)
/// @param      in_u32StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADAPI::readPerformanceTest(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32TestSizeMB)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     m_pSATADDAPI = new SATADDAPI("SATA");

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pSATADDAPI->generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     if((pErrorMsg = m_pSATADDAPI->getDataByID(in_u32LogicalID, eDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     UINT32 u32CapacityMB = 0;
     if((pErrorMsg = m_pSATADDAPI->readCapacity16(in_u32LogicalID, &u32CapacityMB)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     UINT32 u32StartSector = in_u32StartSector;
     UINT32 u32LastSector = u32CapacityMB*1048576/SATA_SECTOR_SIZE;

     struct itimerval startTime, endTime, setting;

     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(SATA_SECTOR_SIZE);

     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  errorCode = ERROR_SATA_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }
     //getitimer() is used rather than gettimeofday() because it is much more consistent (on my machine, at least)
     setting.it_interval.tv_sec = 1000;
     setting.it_interval.tv_usec = 0;
     setting.it_value.tv_sec = 1000;
     setting.it_value.tv_usec = 0;
     // setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
     setitimer(ITIMER_REAL, &setting, NULL);

     // Now do the timings for real
     initProgress(in_u32TestSizeMB);
     do
     {
	  getitimer(ITIMER_REAL, &startTime);
	  pErrorMsg = m_pSATADDAPI->PREAD(s32FileDescriptor, u32pBuffer, SATA_SECTOR_SIZE, u32StartSector);
	  getitimer(ITIMER_REAL, &endTime);
	  dElapsed = dElapsed + (startTime.it_value.tv_sec - endTime.it_value.tv_sec) + ((startTime.it_value.tv_usec - endTime.it_value.tv_usec) / 1000000.0);

	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  u32StartSector += SATA_SECTOR_SIZE/512;
	  if(u32StartSector >= u32LastSector)
	  {
	       u32StartSector = in_u32StartSector;
	  }
	  u64TotalByte += SATA_SECTOR_SIZE;
	  if((u64TotalByte % 1048576) == 0)
	  {
	       updateProgress(1);//every 1 MB update progress
	  }
     }while(u64TotalByte < ((UINT64)(in_u32TestSizeMB))*1048576);

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
     {
	  ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	  ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1048576/dElapsed << " MB/s" << endl;
	  m_strOutputResultString = ssTemp.str();
     }
     else
     {
	  ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	  ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " MB/s" << endl;
	  m_strOutputResultString = ssTemp.str();
     }

     // release resource
     free(u32pBuffer);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_SATA_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *SATADAPI::writePerformanceTest(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32TestSize)
/// @brief      Write specific size(MB) of data from starting logical sector on hard disk through logicalID to get read performance (MB/s)
/// @param      in_u32LogicalID is a logicalID for SATA device node name (eg. 1)
/// @param      in_u32StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADAPI::writePerformanceTest(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32TestSizeMB)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     m_pSATADDAPI = new SATADDAPI("SATA");

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pSATADDAPI->generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     if((pErrorMsg = m_pSATADDAPI->getDataByID(in_u32LogicalID, eDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_WRITE_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     UINT32 u32CapacityMB = 0;
     if((pErrorMsg = m_pSATADDAPI->readCapacity16(in_u32LogicalID, &u32CapacityMB)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }
     UINT32 u32StartSector = in_u32StartSector;
     UINT32 u32LastSector = u32CapacityMB*1048576/SATA_SECTOR_SIZE;

     struct itimerval startTime, endTime, setting;
     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(SATA_SECTOR_SIZE);
     for(UINT32 u32CNT=0; u32CNT<SATA_SECTOR_SIZE/4; u32CNT++)
     {
	  u32pBuffer[u32CNT] = 0xAAAA5555;// test pattern
     }

     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  errorCode = ERROR_SATA_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }

     setting.it_interval.tv_sec = 1000;
     setting.it_interval.tv_usec = 0;
     setting.it_value.tv_sec = 1000;
     setting.it_value.tv_usec = 0;
     // setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
     setitimer(ITIMER_REAL, &setting, NULL);
     // Now do the timings for real
     initProgress(in_u32TestSizeMB);
     do
     {
	  getitimer(ITIMER_REAL, &startTime);
	  pErrorMsg = m_pSATADDAPI->PWRITE(s32FileDescriptor, u32pBuffer, SATA_SECTOR_SIZE, u32StartSector);
	  getitimer(ITIMER_REAL, &endTime);
	  dElapsed = dElapsed + (startTime.it_value.tv_sec - endTime.it_value.tv_sec) + ((startTime.it_value.tv_usec - endTime.it_value.tv_usec) / 1000000.0);
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
	  }

	  u32StartSector += SATA_SECTOR_SIZE/512;
	  if(u32StartSector >= u32LastSector)
	  {
	       u32StartSector = in_u32StartSector;
	  }
	  u64TotalByte += SATA_SECTOR_SIZE;
	  if((u64TotalByte % 1048576) == 0)
	  {
	       updateProgress(1);//every 1 MB update progress
	  }
     }while(u64TotalByte < ((UINT64)(in_u32TestSizeMB))*1048576);

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
     {
	  ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	  ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1048576/dElapsed << " MB/s" << endl;
	  m_strOutputResultString = ssTemp.str();
     }
     else
     {
	  ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	  ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " MB/s" << endl;
	  m_strOutputResultString = ssTemp.str();
     }

     // release resource
     free(u32pBuffer);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_SATA_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST | errorCode), NULL, NULL);
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *SATADAPI::readwriteTest(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32TestSizeMB)
/// @brief      Write specific size(MB) of data from starting logical sector on hard disk through logicalID and read it back to compare data
/// @param      in_u32LogicalID is a logicalID for SATA device node name (eg. 1)
/// @param      in_u32StartSector is starting sector to test
/// @param      in_u32TestSizeMB is test size by MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADAPI::readwriteTest(UINT32 in_u32LogicalID, UINT32 in_u32StartSector, UINT32 in_u32TestSizeMB)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     m_pSATADDAPI = new SATADDAPI("SATA");

     string strSGDeviceNodeName = "";
     if((pErrorMsg = m_pSATADDAPI->generateSATAList()) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
     }
     if((pErrorMsg = m_pSATADDAPI->getDataByID(in_u32LogicalID, eDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
     }

     UINT32 u32CapacityMB = 0;
     if((pErrorMsg = m_pSATADDAPI->readCapacity16(in_u32LogicalID, &u32CapacityMB)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg(ERROR_SATA_DAPI_READ_PERFORMANCE_TEST, NULL, pErrorMsg);
     }

     UINT32 u32StartSector = in_u32StartSector;
     UINT32 u32LastSector = u32CapacityMB*1048576/SATA_SECTOR_SIZE;

     struct itimerval startTime, endTime, setting;
     UINT64 u64TotalByte = 0;
     double dElapsed = 0;
     UINT32 *u32pBuffer = (UINT32 *)malloc(SATA_SECTOR_SIZE);
     for(UINT32 u32CNT=0; u32CNT<SATA_SECTOR_SIZE/4;u32CNT++)
     {
	  u32pBuffer[u32CNT] = 0x5555AAAA;//test pattern
     }

     UINT32 *u32pBufferRead = (UINT32 *)malloc(SATA_SECTOR_SIZE);
     INT32 s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK);
     if(s32FileDescriptor < 0)
     {
	  errorCode = ERROR_SATA_DAPI_OPEN_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_WRITE_TEST | errorCode), NULL, NULL);
     }
     //getitimer() is used rather than gettimeofday() because it is much more consistent (on my machine, at least)

     setting.it_interval.tv_sec = 1000;
     setting.it_interval.tv_usec = 0;
     setting.it_value.tv_sec = 1000;
     setting.it_value.tv_usec = 0;
     // setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
     setitimer(ITIMER_REAL, &setting, NULL);
     // Now do the timings for real
     initProgress(in_u32TestSizeMB);
     do
     {
	  getitimer(ITIMER_REAL, &startTime);
	  pErrorMsg = m_pSATADDAPI->PWRITE(s32FileDescriptor, u32pBuffer, SATA_SECTOR_SIZE, in_u32StartSector);

	  if(pErrorMsg != NULL)
	  {
	       free(u32pBuffer);
	       free(u32pBufferRead);
	       return returnErrorMsg(ERROR_SATA_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
	  }

	  pErrorMsg = m_pSATADDAPI->PREAD(s32FileDescriptor, u32pBufferRead, SATA_SECTOR_SIZE, in_u32StartSector);
	  getitimer(ITIMER_REAL, &endTime);
	  dElapsed = dElapsed + (startTime.it_value.tv_sec - endTime.it_value.tv_sec) + ((startTime.it_value.tv_usec - endTime.it_value.tv_usec) / 1000000.0);

	  if(pErrorMsg != NULL)
	  {
	       free(u32pBuffer);
	       free(u32pBufferRead);
	       return returnErrorMsg(ERROR_SATA_DAPI_READ_WRITE_TEST, NULL, pErrorMsg);
	  }

	  if(memcmp(u32pBuffer, u32pBufferRead, SATA_SECTOR_SIZE) != 0)
	  {
	       free(u32pBuffer);
	       free(u32pBufferRead);
	       errorCode = ERROR_SATA_DAPI_COMPARE_DATA_ERROR;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_WRITE_TEST | errorCode), NULL, NULL);
	  }

	  u32StartSector += SATA_SECTOR_SIZE/512;
	  if(u32StartSector >= u32LastSector)
	  {
	       u32StartSector = in_u32StartSector;
	  }
	  u64TotalByte += SATA_SECTOR_SIZE;
	  if((u64TotalByte % 1048576) == 0)
	  {
	       updateProgress(1);//every 1 MB update progress
	  }
     }while(u64TotalByte < (UINT64)(in_u32TestSizeMB)*1048576);

     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     if((u64TotalByte / dElapsed) > 1.0)//more than 1MB/s
     {
	  ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	  ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1048576/dElapsed << " MB/s" << endl;
	  m_strOutputResultString = ssTemp.str();
     }
     else
     {
	  ssTemp << "Spend time : " << fixed << setprecision(2) << dElapsed << " s" << endl;
	  ssTemp << "Speed      : " << fixed << setprecision(2) << u64TotalByte/1024/dElapsed << " MB/s" << endl;
	  m_strOutputResultString = ssTemp.str();
     }

     // release resource
     free(u32pBuffer);
     free(u32pBufferRead);
     if(close(s32FileDescriptor) < 0)
     {
	  errorCode = ERROR_SATA_DAPI_CLOSE_DEVICE_FAIL;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DAPI_READ_WRITE_TEST | errorCode), NULL, NULL);
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}
