/// @file       NetAppPSU.cpp
/// @brief      NetAppPSU class
/// @details    Provide low layer functional operations for NetApp PSU
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include "NetAppPSU.hpp"
#include "PCIDDAPI.hpp"

/// @fn         NetAppPSU::NetAppPSU(string in_strConfig) : PMBusPSU()
/// @brief      NetAppPSU constructor
/// @param      in_strConfig is a string for PSU configuration
/// @exception  Nothing
/// @return     Nothing
NetAppPSU::NetAppPSU(string in_strConfig) : PMBusPSU()
{
    CHAR    *szTemp;
    string  strTemp;

    // Expected config format: "name i2c_bus slave_addr"
    szTemp = strtok((CHAR *)in_strConfig.c_str(), " ");
    if (szTemp == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_CONSTRUCTOR | ERROR_CONFIGURATION_INCORRECT_FORMAT), NULL);
        return;
    }
    setName(szTemp);
}

/// @fn         NetAppPSU::~NetAppPSU(void)
/// @brief      BaseClassNetAppPSU destructor
/// @exception  Nothing
/// @return     Nothing
NetAppPSU::~NetAppPSU(void)
{

}

/// @fn         NetAppPSU::getFWVersion(string   *out_strpValue)
///
/// @brief      Get FW version
/// @param      out_strpValue is a pointer to a string for the FW version
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getFWVersion(string *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    CHAR        chData[5];
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(), PMBUS_CMD_MFR_SPECIFIC_FW_REV, PMBUS_CMD_MFR_SPECIFIC_FW_REV, (UINT8 *)chData, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_FW_VERSION, NULL, pErrorMsg);
    }

    *out_strpValue = chData;

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::getLEDList(list<string> *out_pList)
///
/// @brief      Get list of LED names
/// @param      out_pList is a pointer to a list of strings for LED names
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getLEDList(list<string> *out_pList)
{
    if (PS_FAULT_LED_NAME)
    {
        out_pList->push_back(PS_FAULT_LED_NAME);
    }
    if (PS_SAA_LED_NAME)
    {
        out_pList->push_back(PS_SAA_LED_NAME);
    }
    if (PS_STDBY_LED_NAME)
    {
        out_pList->push_back(PS_STDBY_LED_NAME);
    }
    if (FAN_CRU1_SAR_LED_NAME)
    {
        out_pList->push_back(FAN_CRU1_SAR_LED_NAME);
    }
    if (FAN_CRU1_SAA_LED_NAME)
    {
        out_pList->push_back(FAN_CRU1_SAA_LED_NAME);
    }
    if (FAN_CRU2_SAR_LED_NAME)
    {
        out_pList->push_back(FAN_CRU2_SAR_LED_NAME);
    }
    if (FAN_CRU2_SAA_LED_NAME)
    {
        out_pList->push_back(FAN_CRU2_SAA_LED_NAME);
    }

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::getLEDStatus(string in_strDeviceName,
///                                    string *out_strpStatus)
///
/// @brief      Get status of the given LED
/// @param      in_strDeviceName is a string for LED
/// @param      out_strpStatus is a pointer to a string for LED status
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getLEDStatus(string in_strDeviceName,
                        string *out_strpStatus)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT16      u16BitMask = 0, u16Data = 0;
    INT32       s32Count = 0;

    if (PS_FAULT_LED_NAME && ((CHAR *)PS_FAULT_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_PS_FAULT_LED;
    }
    else if (PS_SAA_LED_NAME && ((CHAR *)PS_SAA_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_PS_SAA_LED;
    }
    else if (PS_STDBY_LED_NAME && ((CHAR *)PS_STDBY_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_PS_STDBY_LED;
    }
    else if (FAN_CRU1_SAR_LED_NAME && ((CHAR *)FAN_CRU1_SAR_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU1_SAR_LED;
    }
    else if (FAN_CRU1_SAA_LED_NAME && ((CHAR *)FAN_CRU1_SAA_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU1_SAA_LED;
    }
    else if (FAN_CRU2_SAR_LED_NAME && ((CHAR *)FAN_CRU2_SAR_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU2_SAR_LED;
    }
    else if (FAN_CRU2_SAA_LED_NAME && ((CHAR *)FAN_CRU2_SAA_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU2_SAA_LED;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_GET_LED_STATUS | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                                 PMBUS_CMD_MFR_SPECIFIC_LED_CONTROL,
                                 PMBUS_CMD_MFR_SPECIFIC_LED_CONTROL,
                                 &u16Data,
                                 &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_LED_STATUS, NULL, pErrorMsg);
    }

    if (u16Data & u16BitMask)
    {
        *out_strpStatus = "off";
    }
    else
    {
        *out_strpStatus = "on";
    }

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::getPartNumber(string *out_strpValue)
///
/// @brief      Get part number
/// @param      out_strpValue is a pointer to a string for the part number
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getPartNumber(string *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    CHAR        chData[16];
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                                 PMBUS_CMD_MFR_SPECIFIC_PSU_PART_NUMBER,
                                 PMBUS_CMD_MFR_SPECIFIC_PSU_PART_NUMBER,
                                 (UINT8 *)chData,
                                 &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_PART_NUMBER, NULL, pErrorMsg);
    }

    out_strpValue->clear();
    out_strpValue->append((chData + 3), 13);

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::getSerialNumber(string *out_strpValue)
///
/// @brief      Get serial number
/// @param      out_strpValue is a pointer to a string for the serial number
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getSerialNumber(string *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    CHAR        chData[16];
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(), PMBUS_CMD_MFR_SERIAL, PMBUS_CMD_MFR_SERIAL, (UINT8 *)chData, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_SERIAL_NUMBER, NULL, pErrorMsg);
    }

    out_strpValue->clear();
    out_strpValue->append((chData + 3), 13);

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::getSupportFanSpeed(string  in_strDeviceName,
///                                           list<string>   *out_pList)
///
/// @brief      Get supported speeds for the given FAN
/// @param      in_strDeviceName is a string for FAN name
/// @param      out_pList is a pointer to a list of strings for supported speeds
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getSupportFanSpeed(string         in_strDeviceName,
                              list<string>   *out_pList)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT8       u8Data[8];
    INT32       s32Count = 0;
    INT32       s32ByteOffset = 0;
    INT32       s32MaxRPM = 0;

    if (FAN1_NAME && ((CHAR *)FAN1_NAME == in_strDeviceName))
    {
        s32ByteOffset += 0;
    }
    else if (FAN2_NAME && ((CHAR *)FAN2_NAME == in_strDeviceName))
    {
        s32ByteOffset += 2;
    }
    else if (FAN3_NAME && ((CHAR *)FAN3_NAME == in_strDeviceName))
    {
        s32ByteOffset += 4;
    }
    else if (FAN4_NAME && ((CHAR *)FAN4_NAME == in_strDeviceName))
    {
        s32ByteOffset += 6;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_GET_SUPPORT_FAN_SPEED | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                                 PMBUS_CMD_MFR_SPECIFIC_FAN_SPEED,
                                 PMBUS_CMD_MFR_SPECIFIC_FAN_SPEED,
                                 u8Data, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_SUPPORT_FAN_SPEED, NULL, pErrorMsg);
    }

    s32MaxRPM = (INT32)transformLinearToHuman(BYTES_TO_LINEAR_DATA(u8Data + s32ByteOffset));

    out_pList->push_back("0");
    out_pList->push_back(std::to_string(s32MaxRPM));

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::getVendor(string *out_strpValue)
///
/// @brief      Get vendor name
/// @param      out_strpValue is a pointer to a string for the vendor name
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::getVendor(string *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    CHAR        chData[12];
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(), PMBUS_CMD_MFR_ID, PMBUS_CMD_MFR_ID, (UINT8 *)chData, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_VENDOR_NAME, NULL, pErrorMsg);
    }

    out_strpValue->clear();
    out_strpValue->append((chData + 3), 9);

    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::program(CHAR    *in_chpFwBuffer,
///                                INT32   in_s32FwSize)
/// @brief      Check if PSU is present
/// @param      out_bpPresent is a pointer to a BOOL for the status
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::program(UINT8   in_u8Sector,
                   CHAR    *in_chpFwBuffer,
                   INT32   in_s32FwSize)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    INT32       s32Count;
    UINT8       u8DownloadInfo[32];
    struct ProgrammingStatus progStatus;
    INT32       s32RetryCount = 3;
    INT32       s32Offset = 0, s32DataSize = 28;
    UINT8       u8FwData[32];

    if (in_u8Sector != 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_PROGRAM | ERROR_BAD_PARAMETER_ONE),
                              "sector should be 0 only");
    }

    if ((pErrorMsg = m_pI2CDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    // Skip the 32 bytes file header
    in_chpFwBuffer += 32;
    in_s32FwSize -= 32;

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                             PMBUS_CMD_MFR_DOWNLOAD_INFO,
                             PMBUS_CMD_MFR_DOWNLOAD_INFO,
                             u8DownloadInfo,
                             &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    memset(&progStatus, 0, sizeof(progStatus));

    // Enable programming mode
    progStatus.PROG_MODE = 1;
    pErrorMsg = m_pI2CDAPI->Write(getName(),
                              PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                              (UINT8 *)&progStatus,
                              1);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    // Wait to change mode
    usleep(transformLinearToHuman(*((UINT16 *)u8DownloadInfo)) * 1000);

    cout << "erase flash" << endl;
    // Erase flash memory
    progStatus.FLASH_ERASE = 1;
    pErrorMsg = m_pI2CDAPI->Write(getName(),
                              PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                              (UINT8 *)&progStatus,
                              1);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    // Wait for flash erase
    s32RetryCount = 10;
    do
    {
        usleep(1000000);
        pErrorMsg = m_pI2CDAPI->Read(getName(),
                                 PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                                 PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                                 (UINT8 *)&progStatus,
                                 &s32Count);
        if (pErrorMsg != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
        }
    } while (progStatus.FLASH_ERASE && --s32RetryCount);

    if (!s32RetryCount)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_PROGRAM | ERROR_TIME_OUT),
                              "erase timeout");
    }

    cout << "Write FW to PSU" << endl;
    while (s32Offset < in_s32FwSize)
    {
        int progress = (s32Offset * 100) / in_s32FwSize;

        printf("\r%d%%", progress);
        fflush(stdout);

        memset(u8FwData, 0, sizeof(u8FwData));
        *((UINT32 *)u8FwData) = s32Offset;

        if ((in_s32FwSize - s32Offset + 1) < s32DataSize)
        {
            s32DataSize = in_s32FwSize - s32Offset + 1;
        }
        memcpy(u8FwData + 4, in_chpFwBuffer, s32DataSize);

        s32RetryCount = 3;
        do
        {
            if (pErrorMsg != NO_ERROR_MSG)
            {
                deleteErrorMsg(pErrorMsg);
            }

            pErrorMsg = m_pI2CDAPI->Write(getName(),
                                      PMBUS_CMD_MFR_FW_DOWNLOAD_DATA_ADDRESS_MODE,
                                      u8FwData,
                                      s32DataSize + 4);
        } while (pErrorMsg && --s32RetryCount);

        if (pErrorMsg != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
        }

        usleep(1000000);

        s32RetryCount = 3;
        do
        {
            if (pErrorMsg != NO_ERROR_MSG)
            {
                deleteErrorMsg(pErrorMsg);
            }

            pErrorMsg = m_pI2CDAPI->Read(getName(),
                                     PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                                     PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                                     (UINT8 *)&progStatus,
                                     &s32Count);
        } while (pErrorMsg && --s32RetryCount);

        if (pErrorMsg != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
        }

        if (progStatus.PROG_ERROR)
        {
            string strError = "data[" + std::to_string(s32Offset+s32DataSize);
            strError += "-" + std::to_string(s32Offset);
            strError += "]";
            return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_PROGRAM | ERROR_PSU_DDAPI_PROGRAM_DATA_FAIL),
                    strError.c_str());
        }

        s32Offset += s32DataSize;
        in_chpFwBuffer += s32DataSize;

        usleep(1000000);
    }

    printf("\r");
    cout << "Write FW complete" << endl;

    // Exit programming mode
    memset(&progStatus, 0, sizeof(progStatus));
    pErrorMsg = m_pI2CDAPI->Write(getName(),
                              PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                              (UINT8 *)&progStatus,
                              1);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    usleep(30000000);

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                             PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                             PMBUS_CMD_MFR_PROGRAMMING_STATUS,
                             (UINT8 *)&progStatus,
                             &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    if (progStatus.PROG_ERROR)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_PROGRAM | ERROR_PSU_DDAPI_PROGRAM_DATA_FAIL),
                              "whole fw upgrade");
    }

    cout << "FW upgrade is successful" << endl;

    usleep(250000);


    return NO_ERROR_MSG;
}

/// @fn         NetAppPSU::setLEDStatus(string   in_strDeviceName,
///                                    string   in_strStatus)
///
/// @brief      Set LED to be the given status
/// @param      in_strDeviceName is a string for LED
/// @param      in_strStatus is a string for LED status
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NetAppPSU::setLEDStatus(string    in_strDeviceName,
                        string    in_strStatus)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT16      u16BitMask = 0, u16Data = 0;
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                                 PMBUS_CMD_MFR_SPECIFIC_LED_CONTROL,
                                 PMBUS_CMD_MFR_SPECIFIC_LED_CONTROL,
                                 &u16Data,
                                 &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_SET_LED_STATUS, NULL, pErrorMsg);
    }

    if (PS_FAULT_LED_NAME && ((CHAR *)PS_FAULT_LED_NAME == in_strDeviceName))
    {
        u16Data &= ~BIT_MASK_FAULT_OVERRIDE;
        u16BitMask = BIT_MASK_PS_FAULT_LED;
    }
    else if (PS_SAA_LED_NAME && ((CHAR *)PS_SAA_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_PS_SAA_LED;
    }
    else if (PS_STDBY_LED_NAME && ((CHAR *)PS_STDBY_LED_NAME == in_strDeviceName))
    {
        u16Data &= ~BIT_MASK_STDBY_OVERRIDE;
        u16BitMask = BIT_MASK_PS_STDBY_LED;
    }
    else if (FAN_CRU1_SAR_LED_NAME && ((CHAR *)FAN_CRU1_SAR_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU1_SAR_LED;
    }
    else if (FAN_CRU1_SAA_LED_NAME && ((CHAR *)FAN_CRU1_SAA_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU1_SAA_LED;
    }
    else if (FAN_CRU2_SAR_LED_NAME && ((CHAR *)FAN_CRU2_SAR_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU2_SAR_LED;
    }
    else if (FAN_CRU2_SAA_LED_NAME && ((CHAR *)FAN_CRU2_SAA_LED_NAME == in_strDeviceName))
    {
        u16BitMask = BIT_MASK_FAN_CRU2_SAA_LED;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_GET_LED_STATUS | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    if (in_strStatus == "on")
    {
        u16Data &= ~u16BitMask;
    }
    else if (in_strStatus == "off")
    {
        u16Data |= u16BitMask;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_SET_LED_STATUS | ERROR_UNEXPECTED_VALUE),
                              in_strStatus.c_str());
    }

    pErrorMsg = m_pI2CDAPI->Write(getName(), PMBUS_CMD_MFR_SPECIFIC_LED_CONTROL, &u16Data, 1);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_SET_LED_STATUS, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}
