/// @file       SATADDAPI.cpp
/// @brief      SATADDAPI class
/// @details    Provide low layer diagnostic and utility APIs for SATA module
/// @version    1.0.0
/// @date       2015.12.01
/// @author     Ben Yang
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.
#include "SATADDAPI.hpp"

/// @fn         SATADDAPI::SATADDAPI(STRING in_strClassName):BaseClassDDAPI(in_szClassName)
/// @brief      SATADDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
SATADDAPI::SATADDAPI(STRING in_szClassName) : BaseClassDDAPI(in_szClassName)
{
     m_pPCIDDAPI = NULL;
     m_pPCIDDAPI = new PCIDDAPI("PCI");
     if(m_pPCIDDAPI == NULL)
     {
	  addErrorMsg(ERROR_SATA_DDAPI_CONSTRUCTOR, NULL);
     }
     m_SATAControllerList.clear();
}

/// @fn         SATADDAPI::~SATADDAPI(void)
/// @brief      SATADDAPI destructor
/// @exception  Nothing
/// @return     Nothing
SATADDAPI::~SATADDAPI(void)
{
     if(m_pPCIDDAPI != NULL)
     {
	  delete m_pPCIDDAPI;
     }
}

/// @fn         ErrorMsg *SATADDAPI::flushBufferCache(int in_s32FileDescriptor)
/// @brief      Flush buffer cache on hard disk
/// @param      in_s32FileDescriptor is a opened file discription for hard disk device node
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::flushBufferCache(int in_s32FileDescriptor)
{
     ERRORCODE errorCode = NO_ERROR;
     sync();
     fsync(in_s32FileDescriptor);          //flush buffers
     fdatasync(in_s32FileDescriptor);          //flush buffers sync();
     if(ioctl(in_s32FileDescriptor, BLKFLSBUF, NULL))          //do it again, big time
     {
	  errorCode = ERROR_SATA_DDAPI_FLUSH_BUFFER_CACHE;
	  return returnErrorMsg(errorCode, NULL, NULL);
     }
     else
     {
	  do_drive_cmd(in_s32FileDescriptor, NULL, 0);          //IDE: await completion
     }
     sync();
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *SATADDAPI::IOCTL(char *in_szDeviceNodeName, int in_eOperation, SATAInformationXMLData *inout_pXMLData)
/// @brief      Function for abstraction of DDAPI IOCTL operation
/// @param      in_u32LogicalID is a logicalID for SATA device node name (eg. 1)
/// @param      in_eOperation desired driver opertaion
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::IOCTL(UINT32 in_u32LogicalID, SATAIOCTLOperation in_eOperation)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NULL;
     string strDeviceNodeName = "";
     int s32FileDescriptor = 0;
     switch(in_eOperation)
     {
     case eGenerateSATAList:
          pErrorMsg = generateSATAList();
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DDAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;
     case eListController:
	  pErrorMsg = listController();
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DDAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;
     case eListPort:
	  pErrorMsg = listPort();
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DDAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;
     case eListDrive:
	  pErrorMsg = listDrive();
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg(ERROR_SATA_DDAPI_IOCTL, NULL, pErrorMsg);
	  }
	  break;

     case eFlushBufferCache:
	  if((pErrorMsg = getDataByID(in_u32LogicalID, eDriveDeviceNodeName, &strDeviceNodeName)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, pErrorMsg);	       
	  }
	  if((errorCode = checkDeviceNodeName(strDeviceNodeName)) != NO_ERROR)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, NULL);
	  }
	  if((s32FileDescriptor = open(strDeviceNodeName.c_str(), O_RDWR | O_SYNC | O_NONBLOCK)) < 0)
	  {
	       errorCode = ERROR_SATA_DDAPI_OPEN_DEVICE_FAIL;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, NULL);
	  }
	  pErrorMsg = flushBufferCache(s32FileDescriptor);
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, pErrorMsg);
	  }
	  if(close(s32FileDescriptor) < 0)
	  {
	       errorCode = ERROR_SATA_DDAPI_CLOSE_DEVICE_FAIL;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, NULL);
	  }
	  break;

     case eSATIdentifyCommandTest:
	  if((pErrorMsg = identifyCommandTest(in_u32LogicalID)) != NO_ERROR_MSG)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, pErrorMsg);
	  }
	  break;

     default:
	  errorCode = ERROR_SATA_DDAPI_NO_SUCH_IOCTL_OPERATION;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IOCTL | errorCode), NULL, NULL);
	  break;
     }

     return pErrorMsg;
}

/// @fn         ErrorMsg *SATADDAPI::generateSATAList()
/// @brief      list SATA controller's ports
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::generateSATAList()
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     list <SATAControllerStruct> *inout_pSATAControllerList = &m_SATAControllerList;;
     if((pErrorMsg = initControllerInfo(inout_pSATAControllerList)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GENERATE_SATA_LIST | errorCode), NULL, pErrorMsg);
     }

     for(list <SATAControllerStruct>::iterator listIterator = inout_pSATAControllerList->begin(); listIterator != inout_pSATAControllerList->end() ; listIterator++)
     {
	  if((errorCode = initPortInfo(inout_pSATAControllerList)) != NO_ERROR)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GENERATE_SATA_LIST | errorCode), NULL, NULL);
	  }
     }

     for(list <SATAControllerStruct>::iterator listIterator = inout_pSATAControllerList->begin(); listIterator != inout_pSATAControllerList->end() ; listIterator++)
     {
	  if((errorCode = getPortDeviceNode(inout_pSATAControllerList)) != NO_ERROR)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GENERATE_SATA_LIST | errorCode), NULL, NULL);
	  }
     }

#if 0// for print out data
     for(list <SATAControllerStruct>::iterator controllerListIterator = inout_pSATAControllerList->begin(); controllerListIterator != inout_pSATAControllerList->end() ; controllerListIterator++)
     {
	  cout << controllerListIterator->u32PortAmount << endl;
	  cout << controllerListIterator->u32ControllerNumber << endl;
	  cout << controllerListIterator->strControllerWellKnownName << endl;
	  cout << controllerListIterator->strControllerAddress << endl;
	  cout << controllerListIterator->u32Bus << endl;
	  cout << controllerListIterator->u32Dev << endl;
	  cout << controllerListIterator->u32Func << endl;
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       cout << "logical ID = ";
	       cout << portListIterator->u32LogicalID << endl;
	       cout << "port # = ";
	       cout << portListIterator->u32PortNumber << endl;
	       cout << "port well name = ";
	       cout << portListIterator->strPortWellKnownName << endl;
	       cout << "port device node = ";
	       cout << portListIterator->strDeviceNodeName << endl;
	       cout << "port sg device node = ";
	       cout << portListIterator->strSGDeviceNodeName << endl;
	  }
     }
#endif
     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADDAPI::listController()
{
     ERRORCODE errorCode = NO_ERROR;
     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  cout << controllerListIterator->strControllerWellKnownName << " (" << setfill('0') << setw(4) << controllerListIterator->u32Domain << ":" << controllerListIterator->strControllerAddress << ")" << endl;
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADDAPI::listPort()
{
     ERRORCODE errorCode = NO_ERROR;
     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       cout << portListIterator->strPortWellKnownName << endl;
	  }
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

ErrorMsg *SATADDAPI::listDrive()
{
     ERRORCODE errorCode = NO_ERROR;
     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       if(portListIterator->strDeviceNodeName != "")
	       {
		    cout << portListIterator->strDriveWellKnownName << endl;
	       }
	  }
     }
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ErrorMsg *SATADDAPI::identifyCommandTest(UINT32 in_u32LogicalID)
/// @brief      Send ATA identify command to SATA device to test identify command
/// @param      in_u32LogicalID is a logicalID for SATA device node name (eg. 1)
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::identifyCommandTest(UINT32 in_u32LogicalID)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     string strFirmwareVersion = "";
     string strModelName = "";
     string strSerialNumber = "";

     if((pErrorMsg = SATIdentify16(in_u32LogicalID, &strFirmwareVersion, &strModelName, &strSerialNumber)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IDENTIFY_COMMAND_TEST | errorCode), NULL, pErrorMsg);
     }

     if(strFirmwareVersion=="" || strModelName=="" || strSerialNumber=="")
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_IDENTIFY_COMMAND_TEST | errorCode), NULL, NULL);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         SATADDAPI::checkDeviceNodeName(char *in_szDeviceNodeName, int in_eOperation)
/// @brief      check format of device node name
/// @param      in_strDeviceNodeName is a char string for SATA device node name (eg. /dev/sdb)
/// @exception  Nothing
/// @return     A 32bits error code
ERRORCODE SATADDAPI::checkDeviceNodeName(string in_strDeviceNodeName)
{
     ERRORCODE errorCode = NO_ERROR;
     if(in_strDeviceNodeName.find("/dev/s") == string::npos)
     {
	  errorCode = ERROR_SATA_DDAPI_WRONG_DEVICE_NAME_FORMAT;
	  return errorCode;
     }
     return errorCode;
}

/// @fn         ERRORCODE SATADDAPI::initControllerInfo(list <SATAControllerStruct> *inout_pSATAControllerList)
/// @brief      get controllers in the system
/// @param      *inout_pSATAControllerList is a pointer for SATA list
/// @exception  Nothing
/// @return     A 32bits error code
ErrorMsg *SATADDAPI::initControllerInfo(list <SATAControllerStruct> *inout_pSATAControllerList)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     UINT32 u32ControllerAmount = 0;
#if 0
     CHAR szInput[128] = "";
     CHAR *chpTemp = NULL;
     FILE *pFilePtr;

     pFilePtr = popen("lspci | grep 8086:8c02 | wc -l", "r");// grep sata controller
     if(pFilePtr == NULL)
     {
	  errorCode = ERROR_SATA_DDAPI_OPEN_PIPE_FAIL;
	  return errorCode;
     }
     fgets(szInput, sizeof(szInput), pFilePtr);
     strtok(szInput, "\n");
     fclose(pFilePtr);

     UINT32 u32ControllerAmount = (UINT32) (atoi(szInput));
     if(u32ControllerAmount == 0)
     {
	  errorCode = ERROR_SATA_DDAPI_NO_ANY_SATA_CONTROLLER;
	  return errorCode;
     }
     pFilePtr = popen("lspci | grep 8086:8c02 | awk '{print $1}'", "r");
     if(pFilePtr == NULL)
     {
	  errorCode = ERROR_SATA_DDAPI_OPEN_PIPE_FAIL;
	  return errorCode;
     }
#endif
     PCIClassCode ePCIClassCode = PCI_SATA;
     vector <PCIInformantion> pciInfoList;
     if((pErrorMsg = m_pPCIDDAPI->getPciInfoByClass(ePCIClassCode, &pciInfoList)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_INIT_CONTROLLER_INFO | errorCode), NULL, pErrorMsg);
     }

     vector <PCIInformantion>::iterator PCIInformantionIterator;
     for(PCIInformantionIterator = pciInfoList.begin(); PCIInformantionIterator != pciInfoList.end(); PCIInformantionIterator++)
     {
	  SATAControllerStruct tempSATAControllerStruct;
//     for(UINT32 u32CNT=0; u32CNT<u32ControllerAmount; u32CNT++)
	  tempSATAControllerStruct.u32ControllerNumber = u32ControllerAmount;

	  stringstream ssTemp;
	  ssTemp.str("");
	  ssTemp.clear();
//	  ssTemp << u32CNT;
	  ssTemp << u32ControllerAmount;
	  tempSATAControllerStruct.strControllerWellKnownName = "SATA" + ssTemp.str();
#if 0
	  fgets(szInput, sizeof(szInput), pFilePtr);
	  strtok(szInput, "\n");

	  tempSATAControllerStruct.strControllerAddress = szInput;

	  chpTemp = strtok(szInput, ":.\n");
	  tempSATAControllerStruct.u32Bus = (UINT32) strtol(chpTemp, NULL, 16);

	  chpTemp = strtok(NULL, ":.\n");
	  tempSATAControllerStruct.u32Dev = (UINT32) strtol(chpTemp, NULL, 16);

	  chpTemp = strtok(NULL, ":.\n");
	  tempSATAControllerStruct.u32Func = (UINT32) strtol(chpTemp, NULL, 16);
#endif
	  tempSATAControllerStruct.u32Domain = (UINT32)PCIInformantionIterator->u16Domain;
	  tempSATAControllerStruct.u32Bus = (UINT32)PCIInformantionIterator->u16Bus;
	  tempSATAControllerStruct.u32Dev = (UINT32)PCIInformantionIterator->u16Device;
	  tempSATAControllerStruct.u32Func = (UINT32)PCIInformantionIterator->u8Function;
	  ssTemp.str("");
	  ssTemp.clear();
	  ssTemp << hex << setfill('0') << setw (2) << tempSATAControllerStruct.u32Bus << ":";
	  ssTemp << hex << setfill('0') << setw (2) << tempSATAControllerStruct.u32Dev << ".";
	  ssTemp << hex << tempSATAControllerStruct.u32Func;
	  tempSATAControllerStruct.strControllerAddress = ssTemp.str();

	  UINT32 u32BarAddress = 0;
	  pErrorMsg = m_pPCIDDAPI->getBarByBDF(PCIInformantionIterator->u16Domain, PCIInformantionIterator->u16Bus, PCIInformantionIterator->u16Device, PCIInformantionIterator->u8Function, 5/*choose Bar5*/, &u32BarAddress);
	  if(pErrorMsg != NO_ERROR_MSG)
	  {
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_INIT_CONTROLLER_INFO | errorCode), NULL, pErrorMsg);
	  }

	  INT32 s32FileDescriptor = 0;
	  if((s32FileDescriptor = open("/dev/mem", O_RDONLY | O_SYNC)) < 0)
	  {
	       errorCode = ERROR_SATA_DDAPI_OPEN_MMAP_FAIL;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_INIT_CONTROLLER_INFO | errorCode), NULL, NULL);
	  }
	  HBAMEM *AHCIMEM = (HBAMEM *)mmap(NULL, sizeof(HBAMEM), PROT_READ, MAP_SHARED, s32FileDescriptor, u32BarAddress/*0xfbf02000*/);
	  if(AHCIMEM == NULL)
	  {
	       errorCode = ERROR_SATA_DDAPI_MMAP_FAIL;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_INIT_CONTROLLER_INFO | errorCode), NULL, NULL);
	  }
	  tempSATAControllerStruct.u32PortAmount = (UINT32)((AHCIMEM->cap) & 0xf) + 1;
	  if(munmap((void *)AHCIMEM, sizeof(HBAMEM)) < 0)
	  {
	       errorCode = ERROR_SATA_DDAPI_MUNMAP_FAIL;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_INIT_CONTROLLER_INFO | errorCode), NULL, NULL);
	  }
	  inout_pSATAControllerList->push_back(tempSATAControllerStruct);
	  u32ControllerAmount++;
     }

     if(u32ControllerAmount == 0)
     {
	  errorCode = ERROR_SATA_DDAPI_NO_ANY_SATA_CONTROLLER;
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_INIT_CONTROLLER_INFO | errorCode), NULL, NULL);
     }
#if 0
     fclose(pFilePtr);
#endif
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ERRORCODE SATADDAPI::initPortInfo(list <SATAControllerStruct> *inout_pSATAControllerList)
/// @brief      get controllers in the system
/// @param      *inout_pSATAControllerList is a pointer for SATA list
/// @exception  Nothing
/// @return     A 32bits error code
ERRORCODE SATADDAPI::initPortInfo(list <SATAControllerStruct> *inout_pSATAControllerList)
{
     ERRORCODE errorCode = NO_ERROR;

     INT32 s32FileDescriptor = 0;
     if((s32FileDescriptor = open("/dev/mem", O_RDONLY | O_SYNC)) < 0)
     {
	  errorCode = ERROR_SATA_DDAPI_OPEN_MMAP_FAIL;
	  return errorCode;
     }

     HBAMEM *AHCIMEM = (HBAMEM *)mmap(NULL, sizeof(HBAMEM), PROT_READ, MAP_SHARED, s32FileDescriptor, 0xfbf02000);
     if(AHCIMEM == NULL)
     {
	  errorCode = ERROR_SATA_DDAPI_MMAP_FAIL;
	  return errorCode;
     }

     UINT32 u32LogicalID = 0;
     for(list <SATAControllerStruct>::iterator controllerListIterator = inout_pSATAControllerList->begin(); controllerListIterator != inout_pSATAControllerList->end(); controllerListIterator++)
     {
	  controllerListIterator->u32LogicalID = u32LogicalID;
	  u32LogicalID++;
	  for(UINT32 u32CNT=0; u32CNT<controllerListIterator->u32PortAmount; u32CNT++)
	  {
	       SATAPortStruct tempSATAPortStruct;
	       tempSATAPortStruct.u32LogicalID = u32LogicalID;
	       u32LogicalID++;
	       tempSATAPortStruct.u32PortNumber = u32CNT;
	       stringstream ssTemp;
	       ssTemp.str("");
	       ssTemp.clear();
	       ssTemp << u32CNT;
	       tempSATAPortStruct.strPortWellKnownName = controllerListIterator->strControllerWellKnownName + "-Port" + ssTemp.str();
	       tempSATAPortStruct.strDriveWellKnownName = controllerListIterator->strControllerWellKnownName + "-Drive" + ssTemp.str();
	       tempSATAPortStruct.u32PortErrorRegisters = (AHCIMEM->ports[u32CNT]).serr;
	       tempSATAPortStruct.u32PortStatusRegisters = (AHCIMEM->ports[u32CNT]).ssts;
	       controllerListIterator->SATAPortList.push_back(tempSATAPortStruct);
	  }
     }
     if(munmap((void *)AHCIMEM, sizeof(HBAMEM)) < 0)
     {
	  errorCode = ERROR_SATA_DDAPI_MUNMAP_FAIL;
	  return errorCode;
     }
     return errorCode;
}

/// @fn         ERRORCODE SATADDAPI::getPortDeviceNode(SATAInformationXMLData *inout_pXMLData)
/// @brief      Map port number and device node name
/// @param      *inout_pSATAControllerList is a pointer for SATA list
/// @exception  Nothing
/// @return     A 32bits error code
ERRORCODE SATADDAPI::getPortDeviceNode(list <SATAControllerStruct> *inout_pSATAControllerList)
{
     ERRORCODE errorCode = NO_ERROR;

     CHAR *chpDirToRead = (CHAR *) "/sys/block/";
     DIR *pDir = opendir(chpDirToRead);
     if(pDir == NULL)
     {
	  errorCode = ERROR_SATA_DDAPI_OPEN_DIR_FAIL;
	  return errorCode;
     }
     struct dirent *pDirent;
     struct stat fileAttribute;
     while((pDirent = readdir(pDir)) != NULL)
     {
	  CHAR szFileName[256] = "";
	  snprintf(szFileName, sizeof(szFileName), "%s%s", chpDirToRead, pDirent->d_name);
	  lstat(szFileName, &fileAttribute);

	  if(S_ISDIR(fileAttribute.st_mode))
	  {
	       // This directory pDirent is another directory
	  }
	  else if(S_ISLNK(fileAttribute.st_mode))
	  {
	       CHAR szBuffer[256];
	       INT32 s32Length = readlink(szFileName, szBuffer, sizeof(szBuffer) - 1);
	       if(s32Length != -1)
	       {
		    szBuffer[s32Length] = '\0';
	       }
	       else
	       {
		    closedir(pDir);
		    errorCode = ERROR_SATA_DDAPI_READ_LINK_ERROR;
		    return errorCode;
	       }

	       UINT32 u32PortNumber = 100;//set port number as infinite amount
	       for(list <SATAControllerStruct>::iterator controllerListIterator = inout_pSATAControllerList->begin(); controllerListIterator != inout_pSATAControllerList->end() ; controllerListIterator++)
	       {
		    UINT32 u32WellKnownNameNumber = 0;

		    CHAR szTemp[256] = "";
		    snprintf(szTemp, sizeof(szTemp), "%s", controllerListIterator->strControllerAddress.c_str());

		    CHAR *chpAddress = strstr(szBuffer, szTemp);
		    if(chpAddress == NULL)
		    {
			 continue;
		    }
		    strncpy(szTemp, "host", sizeof(szTemp));
		    chpAddress = strstr(chpAddress, szTemp);
		    if(chpAddress != NULL)
		    {
			 u32PortNumber = (UINT32) (atoi(&chpAddress[4]));
			 for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
			 {
			      if(u32PortNumber == portListIterator->u32PortNumber)
			      {
				   portListIterator->strDeviceNodeName = (string)"/dev/" + pDirent->d_name;
				   // start to get SG device node
				   CHAR szSGDirToRead[128] = "";
				   snprintf(szSGDirToRead, sizeof(szSGDirToRead), "/sys/block/%s/device/scsi_generic/", pDirent->d_name);
				   DIR *pSGDir = opendir(szSGDirToRead);
				   if(pSGDir == NULL)
				   {
					errorCode = ERROR_SATA_DDAPI_OPEN_DIR_FAIL;
					closedir(pSGDir);
					closedir(pDir);
					return errorCode;
				   }
				   struct dirent *pSGEntry;
				   struct stat SGFileAttribute;
				   while((pSGEntry = readdir(pSGDir)) != NULL)
				   {
					CHAR szSGFileName[256];

					snprintf(szSGFileName, sizeof(szSGFileName), "%s", szSGDirToRead);
					lstat(szSGFileName, &SGFileAttribute);

					if(S_ISDIR(SGFileAttribute.st_mode))
					{
					     if(!strcmp(pSGEntry->d_name, ".") || !strcmp(pSGEntry->d_name, ".."))
					     {
						  continue;
					     }
					     portListIterator->strSGDeviceNodeName = (string)"/dev/" + pSGEntry->d_name;
					}
				   }
				   closedir(pSGDir);
				   // end to get SG device node
			      }
			      else // remove unmatch element
			      {
				   portListIterator = controllerListIterator->SATAPortList.erase(portListIterator);
				   portListIterator--;
			      }
			 }
			 // modify u32WellKnownName
			 for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
			 {
				   stringstream ssTemp;
				   ssTemp.str("");
				   ssTemp.clear();
				   ssTemp<<u32WellKnownNameNumber;
				   portListIterator->strDriveWellKnownName = controllerListIterator->strControllerWellKnownName + "-Drive" + ssTemp.str();
				   u32WellKnownNameNumber++;
			 }
		    }
	       }
	  }
	  else if(S_ISREG(fileAttribute.st_mode))
	  {
	       // This pDirent is a regular filename
	  }
     }
     closedir(pDir);
     return errorCode;
}

/// @fn         ErrorMsg *SATADDAPI::getLogicalIdByName(string in_strWellKnownName, UINT32 *out_u32pLogicalID)
/// @brief      Get logicalID by well known name
/// @param      in_strWellKnownName is well known name
/// @param      *out_u32pLogicalID is a pointer for returned logicalID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::getLogicalIdByName(string in_strWellKnownName, UINT32 *out_u32pLogicalID)
{
     ERRORCODE errorCode = ERROR_SATA_DDAPI_LOGICALID_NOT_FOUND;

     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  if(convertStringToUpperCase(controllerListIterator->strControllerWellKnownName) == convertStringToUpperCase(in_strWellKnownName))
	  {
	       errorCode = NO_ERROR;
	       *out_u32pLogicalID = controllerListIterator->u32LogicalID;
	       return returnErrorMsg(errorCode, NULL, NULL);
	  }
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       if(convertStringToUpperCase(portListIterator->strPortWellKnownName) == convertStringToUpperCase(in_strWellKnownName))
	       {
		    errorCode = NO_ERROR;
		    *out_u32pLogicalID = portListIterator->u32LogicalID;
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	       else if(convertStringToUpperCase(portListIterator->strDriveWellKnownName)==convertStringToUpperCase(in_strWellKnownName) && portListIterator->strDeviceNodeName!="")
	       {
		    errorCode = NO_ERROR;
		    *out_u32pLogicalID = portListIterator->u32LogicalID;
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
     }
     return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_LOGICALID_BY_NAME | errorCode), NULL, NULL);
}

/// @fn         ErrorMsg *SATADDAPI::getDataByID(UINT32 in_u32LogicalID, SATADataElement in_eSATADataElement, string *out_strpReturnData)
/// @brief      Get data by logicalID
/// @param      in_u32LogicalID is logicalID for SATA device
/// @param      in_eSATADataElement is which data should be returned
/// @param      *out_strpReturnData is a string pointer for returned data
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::getDataByID(UINT32 in_u32LogicalID, SATADataElement in_eSATADataElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = ERROR_SATA_DDAPI_LOGICALID_NOT_FOUND;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  if(controllerListIterator->u32LogicalID == in_u32LogicalID)
	  {
	       errorCode = NO_ERROR;
	       stringstream ssTemp;
	       ssTemp.str("");
	       ssTemp.clear();
	       switch (in_eSATADataElement)
	       {
	       case eControllerDataAll:
		    getDataByID(in_u32LogicalID, eControllerID, out_strpReturnData);
		    ssTemp << "ID          : " << *out_strpReturnData << endl;
		    *out_strpReturnData = "";
		    getDataByID(in_u32LogicalID, eControllerBDF, out_strpReturnData);
		    ssTemp << "BDF         : " << *out_strpReturnData << endl;
		    *out_strpReturnData = "";
		    getDataByID(in_u32LogicalID, eControllerPortAmount, out_strpReturnData);
		    ssTemp << "Port Amount : " << *out_strpReturnData;
		    *out_strpReturnData = "";
		    *out_strpReturnData = ssTemp.str();
		    break;
	       case eControllerID:
		    ssTemp << controllerListIterator->u32ControllerNumber;
		    *out_strpReturnData = ssTemp.str();
		    break;
	       case eControllerBDF:
		    *out_strpReturnData = controllerListIterator->strControllerAddress;
		    break;
	       case eControllerPortAmount:
		    ssTemp << controllerListIterator->u32PortAmount;
		    *out_strpReturnData = ssTemp.str();
		    break;
	       default:
		    errorCode = ERROR_SATA_DDAPI_NO_SUCH_DATA_ELEMENT;
		    return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, NULL);
	       }
	       return returnErrorMsg(errorCode, NULL, NULL);
	  }
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       if(portListIterator->u32LogicalID==in_u32LogicalID)
	       {
		    errorCode = NO_ERROR;
		    UINT32 u32CapacityMB = 0;
		    stringstream ssTemp;
		    ssTemp.str("");
		    ssTemp.clear();


		    switch (in_eSATADataElement)
		    {
		    case eDriveDataAll:
			 getDataByID(in_u32LogicalID, eDrivePortNumber, out_strpReturnData);
			 ssTemp << "Port Number         : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveDeviceNodeName, out_strpReturnData);
			 ssTemp << "Device Node Name    : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveSGDeviceNodeName, out_strpReturnData);
			 ssTemp << "SG Device Node Name : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveCapacity, out_strpReturnData);
			 ssTemp << "Capacity            : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveSerialNumber, out_strpReturnData);
			 ssTemp << "Serial Number       : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveModelName, out_strpReturnData);
			 ssTemp << "Model Name          : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveFirmwareVersion, out_strpReturnData);
			 ssTemp << "Firmware Version    : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
//			 getDataByID(in_u32LogicalID, eDriveLinkStatus, out_strpReturnData);
//			 ssTemp << "Link Status         : " << *out_strpReturnData << endl;
//			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveLinkSpeed, out_strpReturnData);
			 ssTemp << "Link Speed          : " << *out_strpReturnData << endl;
			 *out_strpReturnData = "";
			 getDataByID(in_u32LogicalID, eDriveTemperature, out_strpReturnData);
			 ssTemp << "Temperature         : " << *out_strpReturnData;
			 *out_strpReturnData = "";
			 *out_strpReturnData = ssTemp.str();
			 break;
		    case eDriveDeviceNodeName:
			 *out_strpReturnData = portListIterator->strDeviceNodeName;
			 break;
		    case eDriveSGDeviceNodeName:
			 *out_strpReturnData = portListIterator->strSGDeviceNodeName;
			 break;
		    case eDriveSerialNumber:
			 pErrorMsg = SATIdentify16(in_u32LogicalID, &(portListIterator->strFirmwareVersion), &(portListIterator->strModelName), &(portListIterator->strSerialNumber));
			 if(pErrorMsg != NO_ERROR_MSG)
			 {
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, pErrorMsg);
			 }
			 *out_strpReturnData = portListIterator->strSerialNumber;
			 break;
		    case eDriveModelName:
			 pErrorMsg = SATIdentify16(in_u32LogicalID, &(portListIterator->strFirmwareVersion), &(portListIterator->strModelName), &(portListIterator->strSerialNumber));
			 if(pErrorMsg != NO_ERROR_MSG)
			 {
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, pErrorMsg);
			 }
			 *out_strpReturnData = portListIterator->strModelName;
			 break;
		    case eDriveFirmwareVersion:
			 pErrorMsg = SATIdentify16(in_u32LogicalID, &(portListIterator->strFirmwareVersion), &(portListIterator->strModelName), &(portListIterator->strSerialNumber));
			 if(pErrorMsg != NO_ERROR_MSG)
			 {
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, pErrorMsg);
			 }
			 *out_strpReturnData = portListIterator->strFirmwareVersion;
			 break;
		    case eDriveCapacity:
			 if((pErrorMsg = readCapacity16(in_u32LogicalID, &u32CapacityMB)) != NO_ERROR_MSG)
			 {
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, pErrorMsg);
			 }
			 ssTemp.str("");
			 ssTemp.clear();
			 ssTemp << u32CapacityMB;
			 portListIterator->strCapacityMB = ssTemp.str() + " MB";
			 *out_strpReturnData = ssTemp.str() + " MB";
			 break;
		    case eDrivePortNumber:
			 ssTemp.str("");
			 ssTemp.clear();
			 ssTemp << portListIterator->u32PortNumber;
			 *out_strpReturnData = ssTemp.str();
			 break;
		    case eDriveLinkStatus:
			 switch((portListIterator->u32PortStatusRegisters)&0x000f)
			 {
			 case 1:
			 case 3:
			      portListIterator->strLinkDetection = "linked";
			      break;
			 case 0:
			 case 4:
			      portListIterator->strLinkDetection = "unlinked";
			      break;
			 default:
			      portListIterator->strLinkDetection = "";
			      errorCode = ERROR_SATA_DDAPI_LINK_STATUS_UNKNOWN;
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, NULL);
			      break;
			 }
			 *out_strpReturnData = portListIterator->strLinkDetection;
			 break;
		    case eDriveLinkSpeed:
			 switch(((portListIterator->u32PortStatusRegisters)&0x00f0)>>4)
			 {
			 case 0:
			      portListIterator->strLinkSpeed = "0 Gb/s";
			      break;
			 case 1:
			      portListIterator->strLinkSpeed = "1.5 Gb/s";
			      break;
			 case 2:
			      portListIterator->strLinkSpeed = "3 Gb/s";
			      break;
			 case 3:
			      portListIterator->strLinkSpeed = "6 Gb/s";
			      break;
			 default:
			      portListIterator->strLinkDetection = "";
			      errorCode = ERROR_SATA_DDAPI_LINK_SPEED_UNKNOWN;
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, NULL);
			      break;
			 }
			 *out_strpReturnData = portListIterator->strLinkSpeed;
			 break;
		    case eDriveTemperature:
			 if((pErrorMsg = getSMARTInfoByID(in_u32LogicalID, 194, 194, out_strpReturnData)) != NO_ERROR_MSG)
			 {
			      return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, pErrorMsg);
			 }
			 for(list <SMARTAttributeStruct>::iterator SMARTAttributeListIterator = portListIterator->SMARTAttributeList.begin(); SMARTAttributeListIterator != portListIterator->SMARTAttributeList.end(); SMARTAttributeListIterator++)
			 {
			      if(SMARTAttributeListIterator->u32IDNumber == 194)
			      {
				   *out_strpReturnData = SMARTAttributeListIterator->strReadableValue;
			      }
			 }
			 break;
		    default:
			 errorCode = ERROR_SATA_DDAPI_NO_SUCH_DATA_ELEMENT;
			 return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, NULL);
			 break;
		    }
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
     }
     return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_ID | errorCode), NULL, NULL);
}

/// @fn         ErrorMsg *SATADDAPI::getErrorRegisterByID(UINT32 in_u32LogicalID, SATAErrorElement in_eSATAErrorElement, string *out_strpReturnData)
/// @brief      Get error register data by logicalID
/// @param      in_u32LogicalID is logicalID for SATA device
/// @param      in_eSATAErrorElement is which error register data should be returned
/// @param      *out_strpReturnData is a string pointer for returned data
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::getErrorRegisterByID(UINT32 in_u32LogicalID, SATAErrorElement in_eSATAErrorElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = ERROR_SATA_DDAPI_LOGICALID_NOT_FOUND;
     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  if(controllerListIterator->u32LogicalID == in_u32LogicalID)
	  {
	       errorCode = NO_ERROR;
	       stringstream ssTemp;
	       ssTemp.str("");
	       ssTemp.clear();
	       switch (in_eSATAErrorElement)
	       {
	       case eClearErr:
		    cout << "clear controller errors" << endl;
		    break;
	       default:
		    errorCode = ERROR_SATA_DDAPI_NO_SUCH_ERROR_ELEMENT;
		    return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_ERROR_REGISTER_BY_ID | errorCode), NULL, NULL);
		    break;
	       }
	       return returnErrorMsg(errorCode, NULL, NULL);
	  }
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       if(portListIterator->u32LogicalID==in_u32LogicalID)
	       {
		    errorCode = NO_ERROR;
		    stringstream ssTemp;
		    ssTemp.str("");
		    ssTemp.clear();
		    switch (in_eSATAErrorElement)
		    {
		    case eClearErr:
			 cout << "clear port errors" << endl;
			 break;
		    case eAllErr:
			 getErrorRegisterByID(in_u32LogicalID, eTransportStateTransitionErr, out_strpReturnData);
			 ssTemp << "Transport State Transition Error                 : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eLinkSequenceErr, out_strpReturnData);
			 ssTemp << "Link Sequence Error                              : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eHandshakeErr, out_strpReturnData);
			 ssTemp << "Handshake Error                                  : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eCRCErr, out_strpReturnData);
			 ssTemp << "CRC Error                                        : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eDisparityErr, out_strpReturnData);
			 ssTemp << "Disparity Error                                  : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, e10bTo8bDecodeErr, out_strpReturnData);
			 ssTemp << "10B To 8B Decode Error                           : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, ePHYInternalErr, out_strpReturnData);
			 ssTemp << "PHY Internal Error                               : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eInternalErr, out_strpReturnData);
			 ssTemp << "Internal Error                                   : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eProtocolErr, out_strpReturnData);
			 ssTemp << "Protocol Error                                   : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, ePersistentCommunicationOrDataIntegrityErr, out_strpReturnData);
			 ssTemp << "Persistent Communication or Data Integrity Error : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eTransientDataIntegrityErr, out_strpReturnData);
			 ssTemp << "Transient Data Integrity Error                   : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eRecoverCommunicationErr, out_strpReturnData);
			 ssTemp << "Recover Communications Error                     : " << *out_strpReturnData << endl;
			 getErrorRegisterByID(in_u32LogicalID, eRecoverDataIntegrityErr, out_strpReturnData);
			 ssTemp << "Recover Data Intregrity Error                    : " << *out_strpReturnData;
			 *out_strpReturnData = ssTemp.str();
			 break;
		    case eTransportStateTransitionErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>24)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eLinkSequenceErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>23)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eHandshakeErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>22)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eCRCErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>21)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eDisparityErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>20)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case e10bTo8bDecodeErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>19)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case ePHYInternalErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>17)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eInternalErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>11)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eProtocolErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>10)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case ePersistentCommunicationOrDataIntegrityErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>9)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eTransientDataIntegrityErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>8)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eRecoverCommunicationErr:
			 ssTemp << (((portListIterator->u32PortErrorRegisters)>>1)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    case eRecoverDataIntegrityErr:
			 ssTemp << ((portListIterator->u32PortErrorRegisters)&0x1);
			 *out_strpReturnData = "0x" + ssTemp.str();
			 break;
		    default:
			 errorCode = ERROR_SATA_DDAPI_NO_SUCH_ERROR_ELEMENT;
			 return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_ERROR_REGISTER_BY_ID | errorCode), NULL, NULL);

			 break;
		    }
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
     }
     return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_ERROR_REGISTER_BY_ID | errorCode), NULL, NULL);
}

/// @fn         ErrorMsg *SATADDAPI::getSMARTInfoByID(UINT32 in_u32LogicalID, UINT32 in_u32StartIDNumber, UINT32 in_u32EndIDNumber, string *out_strpReturnData)
/// @brief      Get SMART information by logicalID
/// @param      in_u32LogicalID is logicalID for SATA device
/// @param      in_u32StartIDNumber is starting data ID for SMART information
/// @param      in_u32EndIDNumber is endding data ID for SMART information
/// @param      *out_strpReturnData is a string pointer for returned data
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::getSMARTInfoByID(UINT32 in_u32LogicalID, UINT32 in_u32StartIDNumber, UINT32 in_u32EndIDNumber, string *out_strpReturnData)
{
     ERRORCODE errorCode = ERROR_SATA_DDAPI_LOGICALID_NOT_FOUND;

     for(list <SATAControllerStruct>::iterator controllerListIterator = m_SATAControllerList.begin(); controllerListIterator != m_SATAControllerList.end() ; controllerListIterator++)
     {
	  if(controllerListIterator->u32LogicalID == in_u32LogicalID)
	  {
	       errorCode = ERROR_SATA_DDAPI_WRONG_LOGICAL_ID;
	       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_SMART_INFO_BY_ID | errorCode), NULL, NULL);
	  }
	  for(list <SATAPortStruct>::iterator portListIterator = controllerListIterator->SATAPortList.begin(); portListIterator != controllerListIterator->SATAPortList.end() ; portListIterator++)
	  {
	       if(portListIterator->u32LogicalID==in_u32LogicalID)
	       {
		    errorCode = NO_ERROR;
		    stringstream ssTemp;
		    ssTemp.str("");
		    ssTemp.clear();

		    SkDisk *pDeviceStruct;

		    if(sk_disk_open(portListIterator->strDeviceNodeName.c_str(), &pDeviceStruct) < 0) 
		    {
			 errorCode = ERROR_SATA_DDAPI_OPEN_DEVICE_FAIL;
			 return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_SMART_INFO_BY_ID | errorCode), NULL, NULL);
		    }

		    if(sk_disk_dump(pDeviceStruct, &portListIterator) < 0)
		    {
			 errorCode = ERROR_SATA_DDAPI_DUMP_SMART_INFO_ERROR;
			 return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_SMART_INFO_BY_ID | errorCode), NULL, NULL);
		    }
		    if(pDeviceStruct)
		    {
			 sk_disk_free(pDeviceStruct);
		    }
		    ssTemp << left << setw(5) << "ID";
		    ssTemp << left << setw(25) << "Name";
		    ssTemp << left << setw(8) << "Current";
		    ssTemp << left << setw(8) << "Worst";
		    ssTemp << left << setw(10) << "Threshold";
		    ssTemp << left << setw(20) << "Readable";
		    ssTemp << left << setw(18) << "RawData";
		    ssTemp << left << setw(10) << "Type";
		    ssTemp << left << setw(10) << "Update";
		    ssTemp << left << setw(15) << "HealthStatus";
		    ssTemp << left << setw(20) << "PastHealthStatus" << endl;
		    ssTemp << "-------------------------------------------------------------------------------------------------------------------------------------------------" << endl;

		    for(list <SMARTAttributeStruct>::iterator SMARTAttributeListIterator = portListIterator->SMARTAttributeList.begin(); SMARTAttributeListIterator != portListIterator->SMARTAttributeList.end(); SMARTAttributeListIterator++)
		    {
			 if(in_u32StartIDNumber<=SMARTAttributeListIterator->u32IDNumber && in_u32EndIDNumber>=SMARTAttributeListIterator->u32IDNumber)
			 {
			      ssTemp << left << setw(5) << SMARTAttributeListIterator->u32IDNumber;
			      ssTemp << left << setw(25) << SMARTAttributeListIterator->strName;
			      ssTemp << left << setw(8) <<  SMARTAttributeListIterator->strCurrentValue;
			      ssTemp << left << setw(8) << SMARTAttributeListIterator->strWorstValue;
			      ssTemp << left << setw(10) << SMARTAttributeListIterator->strThresholdValue;
			      ssTemp << left << setw(20) << SMARTAttributeListIterator->strReadableValue;
			      ssTemp << left << setw(18) << SMARTAttributeListIterator->strRawData;
			      ssTemp << left << setw(10) << SMARTAttributeListIterator->strType;
			      ssTemp << left << setw(10) << SMARTAttributeListIterator->strUpdate;
			      ssTemp << left << setw(15) << SMARTAttributeListIterator->strHealthStatus;
			      ssTemp << left << setw(20) << SMARTAttributeListIterator->strPastHealthStatus << endl;
			 }
		    }
		    *out_strpReturnData = ssTemp.str();
		    return returnErrorMsg(errorCode, NULL, NULL);
	       }
	  }
     }
     return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_SMART_INFO_BY_ID | errorCode), NULL, NULL);
}

/// @fn         ErrorMsg *SATADDAPI::getDataByName(string in_strWellKnownName, SATADataElement in_eSATADataElement, string *out_strpReturnData)
/// @brief      Get data by well known name
/// @param      in_strWellKnownName is well known name for SATA device
/// @param      in_eSATADataElement is which data should be returned
/// @param      *out_strpReturnData is a string pointer for returned data
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::getDataByName(string in_strWellKnownName, SATADataElement in_eSATADataElement, string *out_strpReturnData)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     UINT32 u32LogicalID = 0;
     if((pErrorMsg = getLogicalIdByName(in_strWellKnownName, &u32LogicalID)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_NAME | errorCode), NULL, pErrorMsg);
     }

     if((pErrorMsg = getDataByID(u32LogicalID, in_eSATADataElement, out_strpReturnData)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_GET_DATA_BY_NAME | errorCode), NULL, pErrorMsg);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ERRORCODE SATADDAPI::SCSICDBGenerate(UINT32 in_u32BlockSize, UINT32 in_u32LogicalSector, UINT32 u32OperationSwitch)
/// @brief      Generate SCSI CDB by operation switch
/// @param      in_u32BlockSize is block size for SATA device
/// @param      in_u32LogicalSector is starting logical sector for SATA device
/// @param      in_u32OperationSwitch is operation switch to generate CDB
/// @exception  Nothing
/// @return     A 32bits error code
ERRORCODE SATADDAPI::SCSICDBGenerate(UINT32 in_u32BlockSize, UINT32 in_u32LogicalSector, UINT32 in_u32OperationSwitch)
{
     ERRORCODE errorCode = NO_ERROR;

     UCHAR uchaCDBBlockTemp[64];
     memset(uchaCDBBlockTemp, 0, sizeof(uchaCDBBlockTemp));
     UINT64 u64LogicalSectorTemp = in_u32BlockSize*in_u32LogicalSector/512;
     UINT32 accessCount = in_u32BlockSize/512;

     if(in_u32BlockSize%512 != 0)
     {
	  errorCode = ERROR_SATA_DDAPI_BLOCKSIZE_IS_NOT_MULTIPLE_OF_512;
	  return errorCode;
     }

     switch(in_u32OperationSwitch)
     {
     case 0://read command
	  uchaCDBBlockTemp[0] = 0x88;
	  uchaCDBBlockTemp[1] = 0;

	  uchaCDBBlockTemp[2] = (u64LogicalSectorTemp>>56) & 0xff;
	  uchaCDBBlockTemp[3] = (u64LogicalSectorTemp>>48) & 0xff;
	  uchaCDBBlockTemp[4] = (u64LogicalSectorTemp>>40) & 0xff;
	  uchaCDBBlockTemp[5] = (u64LogicalSectorTemp>>32) & 0xff;
	  uchaCDBBlockTemp[6] = (u64LogicalSectorTemp>>24) & 0xff;
	  uchaCDBBlockTemp[7] = (u64LogicalSectorTemp>>16) & 0xff;
	  uchaCDBBlockTemp[8] = (u64LogicalSectorTemp>>8) & 0xff;
	  uchaCDBBlockTemp[9] = u64LogicalSectorTemp & 0xff;

	  uchaCDBBlockTemp[10] = (accessCount>>24) & 0xff;
	  uchaCDBBlockTemp[11] = (accessCount>>16) & 0xff;
	  uchaCDBBlockTemp[12] = (accessCount>>8) & 0xff;
	  uchaCDBBlockTemp[13] = accessCount & 0xff;

	  uchaCDBBlockTemp[14] = 0;
	  uchaCDBBlockTemp[15] = 0;
	  break;
     case 1://write command
	  uchaCDBBlockTemp[0] = 0x8a;
	  uchaCDBBlockTemp[1] = 0;

	  uchaCDBBlockTemp[2] = (u64LogicalSectorTemp>>56) & 0xff;
	  uchaCDBBlockTemp[3] = (u64LogicalSectorTemp>>48) & 0xff;
	  uchaCDBBlockTemp[4] = (u64LogicalSectorTemp>>40) & 0xff;
	  uchaCDBBlockTemp[5] = (u64LogicalSectorTemp>>32) & 0xff;
	  uchaCDBBlockTemp[6] = (u64LogicalSectorTemp>>24) & 0xff;
	  uchaCDBBlockTemp[7] = (u64LogicalSectorTemp>>16) & 0xff;
	  uchaCDBBlockTemp[8] = (u64LogicalSectorTemp>>8) & 0xff;
	  uchaCDBBlockTemp[9] = u64LogicalSectorTemp & 0xff;

	  uchaCDBBlockTemp[10] = (accessCount>>24) & 0xff;
	  uchaCDBBlockTemp[11] = (accessCount>>16) & 0xff;
	  uchaCDBBlockTemp[12] = (accessCount>>8) & 0xff;
	  uchaCDBBlockTemp[13] = accessCount & 0xff;

	  uchaCDBBlockTemp[14] = 0;
	  uchaCDBBlockTemp[15] = 0;
	  break;
     case 2://read capacity16
	  uchaCDBBlockTemp[0] = 0x9e;
	  uchaCDBBlockTemp[1] = 0x10;

	  uchaCDBBlockTemp[2] = 0;
	  uchaCDBBlockTemp[3] = 0;
	  uchaCDBBlockTemp[4] = 0;
	  uchaCDBBlockTemp[5] = 0;
	  uchaCDBBlockTemp[6] = 0;
	  uchaCDBBlockTemp[7] = 0;
	  uchaCDBBlockTemp[8] = 0;
	  uchaCDBBlockTemp[9] = 0;

	  uchaCDBBlockTemp[10] = 0;
	  uchaCDBBlockTemp[11] = 0;
	  uchaCDBBlockTemp[12] = 0;
	  uchaCDBBlockTemp[13] = 1;

	  uchaCDBBlockTemp[14] = 0;
	  uchaCDBBlockTemp[15] = 0;
	  break;
     case 3://sat identify16
	  uchaCDBBlockTemp[0] = 0x85;
	  uchaCDBBlockTemp[1] = 0x8;

	  uchaCDBBlockTemp[2] = 0xe;
	  uchaCDBBlockTemp[3] = 0;
	  uchaCDBBlockTemp[4] = 0;
	  uchaCDBBlockTemp[5] = 0;
	  uchaCDBBlockTemp[6] = 0x01;
	  uchaCDBBlockTemp[7] = 0;
	  uchaCDBBlockTemp[8] = 0;
	  uchaCDBBlockTemp[9] = 0;

	  uchaCDBBlockTemp[10] = 0;
	  uchaCDBBlockTemp[11] = 0;
	  uchaCDBBlockTemp[12] = 0;
	  uchaCDBBlockTemp[13] = 0;

	  uchaCDBBlockTemp[14] = 0xec;
	  uchaCDBBlockTemp[15] = 0;
	  break;
     }
     memcpy(m_uchaCDBBlock, uchaCDBBlockTemp, m_u32CDBLength);

     return errorCode;
}

/// @fn         ERRORCODE SATADDAPI::SCSICDBSend(INT32 in_s32FileDescriptor, UINT32 in_u32BlockSize, UINT32* inout_u32pBuffer, BOOL in_blsZeroReadOneWrite)
/// @brief      Send SCSI command by generated CDB to SATA device
/// @param      in_u32BlockSize is block size for SATA device
/// @param      in_u32LogicalSector is starting logical sector for SATA device
/// @param      in_u32OperationSwitch is operation switch to generate CDB
/// @exception  Nothing
/// @return     A 32bits error code
ERRORCODE SATADDAPI::SCSICDBSend(INT32 in_s32FileDescriptor, UINT32 in_u32BlockSize, UINT32* inout_u32pBuffer, BOOL in_blsZeroReadOneWrite)
{
     ERRORCODE errorCode = NO_ERROR;

     sg_io_hdr_t io_hdr;
     UCHAR uchaSneseBuffer[32] = "";

     /* Prepare READ_16 command */
     memset(&io_hdr, 0, sizeof(sg_io_hdr_t));
     io_hdr.interface_id = 'S';
     io_hdr.cmd_len = m_u32CDBLength;
     /* io_hdr.iovec_count = 0; */  /* memset takes care of this */
     io_hdr.mx_sb_len = sizeof(uchaSneseBuffer);

     switch(in_blsZeroReadOneWrite)
     {
     case 1:
	  io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
	  break;
     default:
	  io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
	  break;
     }

     io_hdr.dxfer_len = in_u32BlockSize;
     io_hdr.dxferp = (UCHAR*)inout_u32pBuffer;
     io_hdr.cmdp = m_uchaCDBBlock;
     io_hdr.sbp = uchaSneseBuffer;
     io_hdr.timeout = 10000;     /* 10000 millisecs == 10 seconds */
     /* io_hdr.flags = 0; */     /* take defaults: indirect IO, etc */
     /* io_hdr.pack_id = 0; */
     /* io_hdr.usr_ptr = NULL; */

     if(ioctl(in_s32FileDescriptor, SG_IO, &io_hdr) < 0)
     {
	  perror("cdb_16_byte_command: SG_IO ioctl error");
	  errorCode = ERROR_SATA_DDAPI_SCSICDBSEND_IOCTL_FAIL;
	  return errorCode;
     }

     /* now for the error processing */
     string straSenseKeyContent[16] = {
	  "No Sense",
	  "Recovered Error",
	  "Not Ready",
	  "Medium Error",
	  "Hardware Error",
	  "Illegal Request",
	  "Unit Attention",
	  "Data Protect",
	  "Blank Check",
	  "Vendor Specific",
	  "Copy Aborted",
	  "Aborted Command",
	  "Equal",
	  "Volume Overflow",
	  "Miscompare",
	  "Completed"
     };

     UINT32 u32SenseKeyValue = 0;
     UINT32 u32AdditionalSenseCode = 0;
     UINT32 u32AdditionalSenseCodeQualifier = 0;
     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     for(UINT32 u32CNT=0; u32CNT<sizeof(uchaSneseBuffer); u32CNT++)
     {
          if(uchaSneseBuffer[u32CNT] != 0)
          {
	       // get sense key
	       ssTemp << (uchaSneseBuffer[2] & 0xf);
	       ssTemp >> u32SenseKeyValue;
	       
	       // get attitional sense code
	       ssTemp.str("");
	       ssTemp.clear();
	       ssTemp << (uchaSneseBuffer[12] & 0xff);
	       ssTemp >> u32AdditionalSenseCode;

	       // get attitional sense code qualifier
	       ssTemp.str("");
	       ssTemp.clear();
	       ssTemp << (uchaSneseBuffer[13] & 0xff);
	       ssTemp >> u32AdditionalSenseCodeQualifier;

	       break;
          }
     }

     if(u32SenseKeyValue != 0)
     {
	  printf("Status: CC\n");
	  cout << "SK:     " << straSenseKeyContent[u32SenseKeyValue] << endl;
	  printf("SKV:    %02X\n", u32SenseKeyValue);
	  printf("ASC:    %02X\n", u32AdditionalSenseCode);
	  printf("ASCQ:   %02X\n", u32AdditionalSenseCodeQualifier);
	  errorCode = ERROR_SATA_DDAPI_SCSICDBSEND_SENSE_ERROR;
	  return errorCode;
     }

     return errorCode;
}

/// @fn         ERRORCODE SATADDAPI::readCapacity16(UINT32 in_u32LogicalID, UINT32 *out_u32pCapacityMB)
/// @brief      Map port number and device node name
/// @param      in_u32LogicalID is logicalID for SATA device
/// @param      *out_u32pCapacityMB is pointer for returned drive size in MB
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::readCapacity16(UINT32 in_u32LogicalID, UINT32 *out_u32pCapacityMB)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     string strSGDeviceNodeName = "";
     if((pErrorMsg = getDataByID(in_u32LogicalID, eDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_READ_CAPACITY_16 | errorCode), NULL, pErrorMsg);
     }

     INT32 s32FileDescriptor;
     UINT32 u32aBuffer[512/4];//allocate 512 bytes u32aBuffer
     if((s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDONLY | O_SYNC)) < 0)
     {
	  errorCode = ERROR_SATA_DDAPI_OPEN_DEVICE_FAIL;
	  close(s32FileDescriptor);
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_READ_CAPACITY_16 | errorCode), NULL, NULL);
     }
     m_u32CDBLength = 16;
     memset(m_uchaCDBBlock, 0, sizeof(m_uchaCDBBlock));

     if((errorCode = SCSICDBGenerate(SATA_SECTOR_SIZE, 0, 2)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_READ_CAPACITY_16 | errorCode), NULL, NULL);
     }

     if((errorCode = SCSICDBSend(s32FileDescriptor, SATA_SECTOR_SIZE, u32aBuffer, 0)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_READ_CAPACITY_16 | errorCode), NULL, NULL);
     }

     for(UINT32 u32CNT=0; u32CNT<3; u32CNT++)
     {
	  u32aBuffer[u32CNT] = SWAB4(u32aBuffer[u32CNT]);
     }
     UINT64 u64CapacityBytes = (((((UINT64)u32aBuffer[0])<<32) | (UINT64)u32aBuffer[1])+1) * (UINT64)u32aBuffer[2];
     *out_u32pCapacityMB = u64CapacityBytes/1048576;

     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ERRORCODE SATADDAPI::SATIdentify16(UINT32 in_u32LogicalID, string* out_strpFirmwareVersion, string* out_strpModelName, string* out_strpSerialNumber)
/// @brief      send identify16 to SATA device by SAT
/// @param      in_u32LogicalID is logicalID for SATA device
/// @param      *out_strpFirmwareVersion is pointer to return SATA drive firmware version
/// @param      *out_strpModelName is pointer to return SATA drive model name
/// @param      *out_strpSerialNumber is pointer to return SATA drive serial number
/// @exception  Nothing
/// @return     A 32bits error code
ErrorMsg *SATADDAPI::SATIdentify16(UINT32 in_u32LogicalID, string* out_strpFirmwareVersion, string* out_strpModelName, string* out_strpSerialNumber)
{
     ERRORCODE errorCode = NO_ERROR;
     ErrorMsg *pErrorMsg = NO_ERROR_MSG;

     string strSGDeviceNodeName = "";
     if((pErrorMsg = getDataByID(in_u32LogicalID, eDriveSGDeviceNodeName, &strSGDeviceNodeName)) != NO_ERROR_MSG)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, pErrorMsg);
     }
     UINT32 u32Buffer[512/4];//allocate 512 bytes u32Buffer
     INT32 s32FileDescriptor;
     if((s32FileDescriptor = open(strSGDeviceNodeName.c_str(), O_RDONLY | O_SYNC)) < 0)
     {
	  errorCode = ERROR_SATA_DDAPI_OPEN_DEVICE_FAIL;
	  close(s32FileDescriptor);
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, NULL);
     }
     m_u32CDBLength = 16;
     memset(m_uchaCDBBlock, 0, sizeof(m_uchaCDBBlock));

     if((errorCode = SCSICDBGenerate(SATA_SECTOR_SIZE, 0, 3)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, NULL);
     }
     if((errorCode = SCSICDBSend(s32FileDescriptor, SATA_SECTOR_SIZE, u32Buffer, 0)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, NULL);
     }
     string strData = "";
     // get serial number
     if((errorCode = dumpHexDataToString(out_strpSerialNumber, &(u32Buffer[5]), 20, "ASCII", 0)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, NULL);
     }
     for(UINT32 u32CNT = 0; u32CNT<(*out_strpSerialNumber).size()-1; u32CNT=u32CNT+2)     // swap characters
     {
	  strData = strData + (*out_strpSerialNumber).at(u32CNT+1);
	  strData = strData + (*out_strpSerialNumber).at(u32CNT);
     }

     string strTemp = "";
     for(UINT32 u32CNT=0; u32CNT<strData.size(); u32CNT++)     // delete empty character
     {
//	  strData.erase(std::remove(strData.begin(), strData.end(), ' '), strData.end());
	  if(strData[u32CNT] != ' ')
	  {
	       strTemp = strTemp + strData[u32CNT];
	  }
     }
     strData = strTemp;
     *out_strpSerialNumber = strData;

     // get part number
     strData = "";
     if((errorCode = dumpHexDataToString(out_strpModelName, &(u32Buffer[13]), 42, "ASCII", 0)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, NULL);
     }
     for(UINT32 u32CNT=0; u32CNT<(*out_strpModelName).size()-1; u32CNT=u32CNT+2)     // swap characters
     {
	  strData = strData + (*out_strpModelName).at(u32CNT+1);
	  strData = strData + (*out_strpModelName).at(u32CNT);
     }
     strData.erase(0, 2);     // erase first 2 extra-characters

     strTemp = "";
     for(UINT32 u32CNT=0; u32CNT<strData.size(); u32CNT++)     // delete empty character
     {
//	  strData.erase(std::remove(strData.begin(), strData.end(), ' '), strData.end());
	  if(strData[u32CNT] != ' ')
	  {
	       strTemp = strTemp + strData[u32CNT];
	  }
     }
     strData = strTemp;
     *out_strpModelName = strData;

     // get firmware version
     strData = "";
     if((errorCode = dumpHexDataToString(out_strpFirmwareVersion, &(u32Buffer[11]), 10, "ASCII", 0)) != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_SAT_IDENTIFY_16 | errorCode), NULL, NULL);
     }
     for(UINT32 u32CNT=0; u32CNT<(*out_strpFirmwareVersion).size()-1; u32CNT=u32CNT+2)     // swap characters
     {
	  strData = strData + (*out_strpFirmwareVersion).at(u32CNT+1);
	  strData = strData + (*out_strpFirmwareVersion).at(u32CNT);
     }
     strData.erase(0, 2);     // erase first 2 extra-characters

     strTemp = "";
     for(UINT32 u32CNT=0; u32CNT<strData.size(); u32CNT++)     // delete empty character
     {
//	  strData.erase(std::remove(strData.begin(), strData.end(), ' '), strData.end());
	  if(strData[u32CNT] != ' ')
	  {
	       strTemp = strTemp + strData[u32CNT];
	  }
     }
     strData = strTemp;
     *out_strpFirmwareVersion = strData;
     return returnErrorMsg(errorCode, NULL, NULL);
}

/// @fn         ERRORCODE SATADDAPI::SATIdentify16(UINT32 in_u32LogicalID, string* out_strpFirmwareVersion, string* out_strpModelName, string* out_strpSerialNumber)
/// @brief      read pattern from SATA device
/// @param      in_s32FileDescriptor is file descriptor to read by Linux
/// @param      *out_u32pBuffer is a pointer for buffering data
/// @param      in_u32BlockSize is block size for SATA device
/// @param      in_u32LogicalSector is starting logical sector for SATA device
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::PREAD(INT32 in_s32FileDescriptor, UINT32* out_u32pBuffer, UINT32 in_u32BlockSize, UINT32 in_u32LogicalSector)
{
     ERRORCODE errorCode = NO_ERROR;

     m_u32CDBLength = 16;
     memset(m_uchaCDBBlock, 0, sizeof(m_uchaCDBBlock));

     errorCode = SCSICDBGenerate(in_u32BlockSize, in_u32LogicalSector, 0);
     if(errorCode != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PREAD | errorCode), NULL, NULL);
     }

     errorCode = SCSICDBSend(in_s32FileDescriptor, in_u32BlockSize, out_u32pBuffer, 0);
     if(errorCode != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PREAD | errorCode), NULL, NULL);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
/*
  int retCode = lseek(in_s32FileDescriptor, in_u32LogicalSector, SEEK_SET);
  if(retCode == -1)
  {
       errorCode = ERROR_SATA_DDAPI_LSEEK_POSITION;
       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PREAD | errorCode), NULL, NULL);
  }
  retCode = read(in_s32FileDescriptor, out_u32pBuffer, in_u32BlockSize);

  if(retCode == -1)
  {
       errorCode = ERROR_SATA_DDAPI_READ_DATA;
       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PREAD | errorCode), NULL, NULL);
  }
*/
}

/// @fn         ErrorMsg *SATADDAPI::PWRITE(INT32 in_s32FileDescriptor, UINT32* in_u32pBuffer, UINT32  in_u32BlockSize, UINT32 in_u32LogicalSector)
/// @brief      write pattern to SATA device
/// @param      in_s32FileDescriptor is file descriptor to read by Linux
/// @param      *out_u32pBuffer is a pointer for buffering data
/// @param      in_u32BlockSize is block size for SATA device
/// @param      in_u32LogicalSector is starting logical sector for SATA device
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg *SATADDAPI::PWRITE(INT32 in_s32FileDescriptor, UINT32* in_u32pBuffer, UINT32  in_u32BlockSize, UINT32 in_u32LogicalSector)
{
     ERRORCODE errorCode = NO_ERROR;

     m_u32CDBLength = 16;
     memset(m_uchaCDBBlock, 0, sizeof(m_uchaCDBBlock));

     errorCode = SCSICDBGenerate(in_u32BlockSize, in_u32LogicalSector, 1);
     if(errorCode != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PWRITE | errorCode), NULL, NULL);
     }

     errorCode = SCSICDBSend(in_s32FileDescriptor, in_u32BlockSize, in_u32pBuffer, 1);
     if(errorCode != NO_ERROR)
     {
	  return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PWRITE | errorCode), NULL, NULL);
     }

     return returnErrorMsg(errorCode, NULL, NULL);
/*
  int retCode = 0;
  errorCode = NO_ERROR;

  retCode = lseek(in_s32FileDescriptor, in_u32LogicalSector, SEEK_SET);
  if(retCode == -1)
  {
       errorCode = ERROR_SATA_DDAPI_LSEEK_POSITION;
       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PWRITE | errorCode), NULL, NULL);
  }

  retCode = write(in_s32FileDescriptor, in_u32pBuffer, in_u32BlockSize);
  if(retCode == -1)
  {
       errorCode = ERROR_SATA_DDAPI_WRITE_DATA;
       return returnErrorMsg((ERRORCODE)(ERROR_SATA_DDAPI_PWRITE | errorCode), NULL, NULL);
  }
*/
}

/// @fn         ERRORCODE SATADDAPI::dumpHexDataToString(string *in_strpDataBuffer, UINT32 *in_u32pDataBuffer, UINT32 in_u32DataLengthInBytes, string in_strDumpStyle, BOOL in_blsDumpAddressHead)
/// @brief      chage hexdump UINT32 data to formatted output string
/// @param      *out_strpDataBuffer is a pointer for output string
/// @param      *in_u32pDataBuffer is a pointer for original input UINT32 dataa
/// @param      in_u32DataLengthInBytes is input data length(bytes)
/// @param      in_strDumpStyle is switch for dumping style(ASCII, HEX, or "" is for ALL)
/// @param      in_blsDumpAddressHead is switch for printing address header or not
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ERRORCODE SATADDAPI::dumpHexDataToString(string *out_strpDataBuffer, UINT32 *in_u32pDataBuffer, UINT32 in_u32DataLengthInBytes, string in_strDumpStyle, BOOL in_blsDumpAddressHead)
{
     ERRORCODE errorCode = NO_ERROR;
     
     if((in_strDumpStyle != "") && (in_strDumpStyle != "HEX") && (in_strDumpStyle != "ASCII"))
     {
	  errorCode = ERROR_SATA_DDAPI_NO_SUCH_DUMP_HEX_FORMAT;
	  return errorCode;
     }

     UINT32 u32I, u32J;
     stringstream ssTemp;
     ssTemp.str("");
     ssTemp.clear();
     for(u32I=0; u32I<in_u32DataLengthInBytes + ((in_u32DataLengthInBytes % HEXDUMP_COLS) ? (HEXDUMP_COLS - in_u32DataLengthInBytes % HEXDUMP_COLS) : 0); u32I++)
     {
	  // print offset
	  if(in_blsDumpAddressHead == 1)
	  {
	       if(u32I % HEXDUMP_COLS == 0)
	       {
		    if(u32I != 0)
		    {
			 //printf("\n");
			 ssTemp << "\n";
		    }
		    //printf("%08X: ", i);
		    ssTemp << hex << uppercase << setw(8) << setfill('0') << u32I;
		    ssTemp << ": ";
	       }
	  }
	  if(in_strDumpStyle=="" || in_strDumpStyle=="HEX")
	  {
	       // print hex data
	       if(u32I < in_u32DataLengthInBytes)
	       {
		    //printf("%02X", 0xFF & ((char*)in_u32pDataBuffer)[i]);
		    ssTemp << hex << uppercase << setfill('0') << setw(2) << (0xFF & ((char*)in_u32pDataBuffer)[u32I]);
	       }
	       else // end of block, just aligning for ASCII dump
	       {
		    //printf("  ");
		    ssTemp << "  ";
	       }
	       // print space
	       if(u32I%4 == 3)
	       {
		    //printf(" ");
		    ssTemp << " ";
	       }
	  }
	  if(in_strDumpStyle=="" || in_strDumpStyle=="ASCII")
	  {
	       // print ASCII dump
	       if((u32I % HEXDUMP_COLS == (HEXDUMP_COLS - 1)))
	       {
		    for(u32J = u32I - (HEXDUMP_COLS - 1); u32J<=u32I; u32J++)
		    {
			 if(u32J >= in_u32DataLengthInBytes) // end of block, not really printing
			 {
			      //putchar(' ');
			      ssTemp << " ";
			 }
			 else if(isprint(((char*)in_u32pDataBuffer)[u32J])) // printable char
			 {
			      //putchar(0xFF & ((char*)in_u32pDataBuffer)[j]);        
			      ssTemp << uppercase << (char)(0xFF & ((char*)in_u32pDataBuffer)[u32J]);
			 }
			 else // other char
			 {
			      //putchar('.');
			      ssTemp << ".";
			 }
		    }
	       }
	  }
     }
     *out_strpDataBuffer = ssTemp.str();
     return errorCode;
}

/// @fn         SATADDAPI::convertStringToUpperCase(string in_strInputString, string *out_strpOutputString)
/// @brief      convert string to upper case
/// @param      in_strInputString is input string
/// @exception  Nothing
/// @return     returned output string
string SATADDAPI::convertStringToUpperCase(string in_strInputString)
{
     string strOutputString = "";
     for(string::iterator stringIterator = in_strInputString.begin(); stringIterator != in_strInputString.end(); stringIterator++)
     {
	  strOutputString += toupper(*stringIterator);
     }
     return strOutputString;
}


/// @fn         UINT32 SATADDAPI::SWAB4(UINT32 in_s32Value)
/// @brief      little endian, big endian exchange
/// @param      in_s32Value is input value
/// @exception  Nothing
/// @return     returned UINT32 value
UINT32 SATADDAPI::SWAB4(UINT32 in_s32Value)
{
     UINT32 u32T;
     u32T = in_s32Value ^ ((in_s32Value <<16) | (in_s32Value >> 16));
     u32T &= ~0x00ff0000;
     in_s32Value = (in_s32Value << 24) | (in_s32Value >> 8);
     return in_s32Value ^ (u32T >> 8);
}
