#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "smp_lib.h"

#define SMP_FN_DRIVE_SLOT_DATA              0xC6
#define SMP_FN_DRIVE_SLOT_DATA_RESP_LEN     32

/* This is a Serial Attached SCSI (SAS) Serial Management Protocol (SMP)
 * utility.
 *
 * This utility issues a DRIVE SLOT DATA function and outputs its
 * response.
 */

static const char * version_str = "0.01 20160503";



static struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {"hex", 0, 0, 'H'},
    {"interface", 1, 0, 'I'},
    {"phy", 1, 0, 'p'},
    {"raw", 0, 0, 'r'},
    {"sa", 1, 0, 's'},
    {"verbose", 0, 0, 'v'},
    {"version", 0, 0, 'V'},
    {0, 0, 0, 0},
};


static void
usage(void)
{
    fprintf(stderr, "Usage: "
          "smp_drive_slot_data [--help] [--hex] [--interface=PARAMS]\n"
          "                           [--phy=ID] [--raw] [--sa=SAS_ADDR]\n"
          "                           [--verbose] [--version] SMP_DEVICE[,N]\n"
          "  where:\n"
          "    --help|-h                       print out usage message\n"
          "    --hex|-H                        print response in hexadecimal\n"
          "    --interface=PARAMS|-I PARAMS    specify or override "
          "interface\n"
          "    --phy=ID|-p ID                  phy identifier (def: 0)\n"
          "    --raw|-r                        output response in binary\n"
          "    --sa=SAS_ADDR|-s SAS_ADDR       SAS address of SMP "
          "target (use leading\n"
          "                                    '0x' or trailing 'h'). "
          "Depending on\n"
          "                                    the interface, may not be "
          "needed\n"
          "    --verbose|-v                    increase verbosity\n"
          "    --version|-V                    print version string and exit\n"
          "Performs a SMP DRIVE SLOT DATA function\n"
          );
}

static void
dStrRaw(const char* str, int len)
{
    int k;

    for (k = 0 ; k < len; ++k)
        printf("%c", str[k]);
}


int
main(int argc, char * argv[])
{
    int res, c, k, len, act_resplen;
    int do_hex = 0;
    int phy_id = 0;
    int phy_id_given = 0;
    int do_raw = 0;
    int verbose = 0;
    long long sa_ll;
    unsigned long long sa = 0;
    char i_params[256];
    char device_name[512];
    char b[256];
    unsigned char smp_req[] = {SMP_FRAME_TYPE_REQ, SMP_FN_DRIVE_SLOT_DATA,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    unsigned char smp_resp[SMP_FN_DRIVE_SLOT_DATA_RESP_LEN];
    struct smp_req_resp smp_rr;
    struct smp_target_obj tobj;
    int subvalue = 0;
    char * cp;
    int ret = 0;

    memset(device_name, 0, sizeof device_name);
    while (1) {
        int option_index = 0;

        c = getopt_long(argc, argv, "hHI:p:rs:vV", long_options,
                        &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 'h':
        case '?':
            usage();
            return 0;
        case 'H':
            ++do_hex;
            break;
        case 'I':
            strncpy(i_params, optarg, sizeof(i_params));
            i_params[sizeof(i_params) - 1] = '\0';
            break;
        case 'p':
           phy_id = smp_get_num(optarg);
           if ((phy_id < 0) || (phy_id > 254)) {
                fprintf(stderr, "bad argument to '--phy', expect "
                        "value from 0 to 254\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            ++phy_id_given;
            break;
        case 'r':
            ++do_raw;
            break;
        case 's':
           sa_ll = smp_get_llnum(optarg);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad argument to '--sa'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            sa = (unsigned long long)sa_ll;
            break;
        case 'v':
            ++verbose;
            break;
        case 'V':
            fprintf(stderr, "version: %s\n", version_str);
            return 0;
        default:
            fprintf(stderr, "unrecognised switch code 0x%x ??\n", c);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (optind < argc) {
        if ('\0' == device_name[0]) {
            strncpy(device_name, argv[optind], sizeof(device_name) - 1);
            device_name[sizeof(device_name) - 1] = '\0';
            ++optind;
        }
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == device_name[0]) {
        cp = getenv("SMP_UTILS_DEVICE");
        if (cp)
            strncpy(device_name, cp, sizeof(device_name) - 1);
        else {
            fprintf(stderr, "missing device name on command line\n    [Could "
                    "use environment variable SMP_UTILS_DEVICE instead]\n");
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if ((cp = strchr(device_name, SMP_SUBVALUE_SEPARATOR))) {
        *cp = '\0';
        if (1 != sscanf(cp + 1, "%d", &subvalue)) {
            fprintf(stderr, "expected number after separator in SMP_DEVICE "
                    "name\n");
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == sa) {
        cp = getenv("SMP_UTILS_SAS_ADDR");
        if (cp) {
           sa_ll = smp_get_llnum(cp);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad value in environment variable "
                        "SMP_UTILS_SAS_ADDR\n");
                fprintf(stderr, "    use 0\n");
                sa_ll = 0;
            }
            sa = (unsigned long long)sa_ll;
        }
    }
    if (sa > 0) {
        if (! smp_is_naa5(sa)) {
            fprintf(stderr, "SAS (target) address not in naa-5 format "
                    "(may need leading '0x')\n");
            if ('\0' == i_params[0]) {
                fprintf(stderr, "    use '--interface=' to override\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
        }
    }

    res = smp_initiator_open(device_name, subvalue, i_params, sa,
                             &tobj, verbose);
    if (res < 0)
        return SMP_LIB_FILE_ERROR;

    smp_req[9] = phy_id;
    if (verbose) {
        fprintf(stderr, "    Drive slot data request: ");
        for (k = 0; k < (int)sizeof(smp_req); ++k)
            fprintf(stderr, "%02x ", smp_req[k]);
        fprintf(stderr, "\n");
    }
    memset(&smp_rr, 0, sizeof(smp_rr));
    smp_rr.request_len = sizeof(smp_req);
    smp_rr.request = smp_req;
    smp_rr.max_response_len = sizeof(smp_resp);
    smp_rr.response = smp_resp;
    res = smp_send_req(&tobj, &smp_rr, verbose);

    if (res) {
        fprintf(stderr, "smp_send_req failed, res=%d\n", res);
        if (0 == verbose)
            fprintf(stderr, "    try adding '-v' option for more debug\n");
        ret = -1;
        goto err_out;
    }
    if (smp_rr.transport_err) {
        fprintf(stderr, "smp_send_req transport_error=%d\n",
                smp_rr.transport_err);
        ret = -1;
        goto err_out;
    }
    act_resplen = smp_rr.act_response_len;
    if ((act_resplen >= 0) && (act_resplen < 4)) {
        fprintf(stderr, "response too short, len=%d\n", act_resplen);
        ret = SMP_LIB_CAT_MALFORMED;
        goto err_out;
    }
    //len = smp_resp[3];
    len = 0x6;
    if ((0 == len) && (0 == smp_resp[2])) {
        len = smp_get_func_def_resp_len(smp_resp[1]);
        if (len < 0) {
            len = 0;
            if (verbose > 0)
                fprintf(stderr, "unable to determine response length\n");
        }
    }
    len = 4 + (len * 4);        /* length in bytes, excluding 4 byte CRC */
    if ((act_resplen >= 0) && (len > act_resplen)) {
        if (verbose)
            fprintf(stderr, "actual response length [%d] less than deduced "
                    "length [%d]\n", act_resplen, len);
        len = act_resplen;
    }
    if (do_hex || do_raw) {
        if (do_hex)
            dStrHex((const char *)smp_resp, len, 1);
        else
            dStrRaw((const char *)smp_resp, len);
        if (SMP_FRAME_TYPE_RESP != smp_resp[0])
            ret = SMP_LIB_CAT_MALFORMED;
        else if (smp_resp[1] != smp_req[1])
            ret = SMP_LIB_CAT_MALFORMED;
        else if (smp_resp[2]) {
            if (verbose)
                fprintf(stderr, "Drive slot data result: %s\n",
                        smp_get_func_res_str(smp_resp[2], sizeof(b), b));
            ret = smp_resp[2];
        }
        goto err_out;
    }
    if (SMP_FRAME_TYPE_RESP != smp_resp[0]) {
        fprintf(stderr, "expected SMP frame response type, got=0x%x\n",
                smp_resp[0]);
        ret = SMP_LIB_CAT_MALFORMED;
        goto err_out;
    }
    if (smp_resp[1] != smp_req[1]) {
        fprintf(stderr, "Expected function code=0x%x, got=0x%x\n",
                smp_req[1], smp_resp[1]);
        ret = SMP_LIB_CAT_MALFORMED;
        goto err_out;
    }
    if (smp_resp[2]) {
        cp = smp_get_func_res_str(smp_resp[2], sizeof(b), b);
        fprintf(stderr, "Drive slot data result%s: %s\n",
                (phy_id_given ? "" : " (for phy_id=0)"), cp);
        ret = smp_resp[2];
        goto err_out;
    }
    printf("Drive slot data response:\n");
    printf("  drive power status: 0x%X (", smp_resp[12]);
    if (smp_resp[12] == 0x23)
        printf("drive powered off)\n");
    else if (smp_resp[12] == 0x2C)
        printf("drive powered on)\n");
    else
        printf("Unknown)\n");
        
    printf("  drive power cycle count: %d\n", smp_resp[13]);
    printf("  drive power on count: %d\n", smp_resp[14]);
    printf("  drive power off count: %d\n", smp_resp[15]);
    printf("  power control in progress: ");
    if (smp_resp[20] & 0x1)
        printf("Yes\n");
    else
        printf("No\n");
    printf("  local PHY state: 0x%02X (", smp_resp[21]);
    switch (smp_resp[21])
    {
        case 0x00:
            printf("PHY is enabled and operational)\n");
            break;
        case 0x11:
            printf("PHY enabled – Unknown physical link rate)\n");
            break;
        case 0x12:
            printf("PHY enabled – No signal detected (Host In PHY only; cable present))\n");
            break;
        case 0x20:
            printf("PHY disabled – Manual from SES command)\n");
            break;
        case 0x21:
            printf("PHY disabled – Drive not physically present; SAS Cable not physically present)\n");
            break;
        case 0x22:
            printf("PHY disabled – Burst Loss of DWORD Synchronization Threshold exceeded)\n");
            break;
        case 0x23:
            printf("PHY disabled – Loss of DWORD Synchronization Threshold exceeded)\n");
            break;
        case 0x24:
            printf("PHY disabled – Burst Running Disparity Error Threshold exceeded)\n");
            break;
        case 0x25:
            printf("PHY disabled – Running Disparity Error Threshold exceeded)\n");
            break;
        case 0x26:
            printf("PHY disabled – Burst Invalid DWORD Threshold exceeded)\n");
            break;
        case 0x27:
            printf("PHY disabled – Invalid DWORD Threshold exceeded)\n");
            break;
        case 0x28:
            printf("PHY disabled – Burst PHY Reset Problem Threshold exceeded)\n");
            break;
        case 0x29:
            printf("PHY disabled – PHY Reset Problem Threshold exceeded)\n");
            break;
        case 0x2A:
            printf("PHY disabled – Burst PHY CHANGE Threshold Exceeded)\n");
            break;
        case 0x2B:
            printf("PHY disabled – PHY CHANGE Threshold Exceeded)\n");
            break;
        case 0x2C:
            printf("Unused)\n");
            break;
        case 0x2D:
            printf("PHY disabled – Mirrored disable from partner module)\n");
            break;
        case 0x2E:
            printf("PHY disabled – Burst CRC error Threshold exceeded)\n");
            break;
        case 0x2F:
            printf("PHY disabled – CRC error Threshold exceeded)\n");
            break;
        case 0x40:
            printf("PHY disabled – SMP PHY_CONTROL(DISABLE))\n");
            break;
        case 0x50:
            printf("PHY disabled – Expander Reference Clock Error)\n");
            break;
        case 0x60:
            printf("PHY disabled – Phy not used)\n");
            break;
        case 0xFF:
            printf("PHY state unknown -- Initializing)\n");
            break;
        default:
            printf("Unknown)\n");
            break;
    }

err_out:
    res = smp_initiator_close(&tobj);
    if (res < 0) {
        fprintf(stderr, "close error: %s\n", safe_strerror(errno));
        if (0 == ret)
            return SMP_LIB_FILE_ERROR;
    }
    if (ret < 0)
        ret = SMP_LIB_CAT_OTHER;
    if (verbose && ret)
        fprintf(stderr, "Exit status %d indicates error detected\n", ret);
    return ret;
}
