#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "smp_lib.h"

/* This is a Serial Attached SCSI (SAS) Serial Management Protocol (SMP)
 * utility.
 *
 * This utility issues a REPORT PHY ERROR LOG LIST function and outputs its
 * response.
 */

static const char * version_str = "0.01 20160516";

#define SMP_FN_DRIVE_SLOT_DATA_LIST              0xCB

static struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {"hex", 0, 0, 'H'},
    {"interface", 1, 0, 'I'},
    {"num", 1, 0, 'n'},
    {"one", 0, 0, 'o'},
    {"phy", 1, 0, 'p'},
    {"raw", 0, 0, 'r'},
    {"sa", 1, 0, 's'},
    {"summary", 0, 0, 'S'},
    {"verbose", 0, 0, 'v'},
    {"version", 0, 0, 'V'},
    {0, 0, 0, 0},
};

struct opts_t {
    int do_hex;
    int do_num;
    int num_given;
    int do_1line;
    int phy_id;
    int phy_id_given;
    int do_raw;
    int do_summary;
    int verbose;
    int sa_given;
    unsigned long long sa;
};

static void
usage(void)
{
    fprintf(stderr, "Usage: "
          "smp_drive_slot_data_list [--help] [--hex]\n"
          "                                [--interface=PARAMS]\n"
          "                                [--num=NUM] [--one] [--phy=ID]\n"
          "                                [--raw] [--sa=SAS_ADDR]\n"
          "                                [--summary] [--verbose]\n"
          "                                [--version] SMP_DEVICE[,N]\n"
          "  where:\n"
          "    --help|-h                       print out usage message\n"
          "    --hex|-H                        print response in hexadecimal\n"
          "    --interface=PARAMS|-I PARAMS    specify or override "
          "interface\n"
          "    --num=NUM|-n NUM                maximum number of descriptors to fetch "
          "(def: 1)\n"
          "    --one|-o                        one line output per response "
          "descriptor (phy)\n"
          "    --phy=ID|-p ID                  phy identifier [or starting phy id]\n"
          "    --raw|-r                        output response in binary\n"
          "    --sa=SAS_ADDR|-s SAS_ADDR       SAS address of SMP "
          "target (use leading\n"
          "                                    '0x' or trailing 'h'). "
          "Depending on\n"
          "                                    the interface, may not be "
          "needed\n"
          "    --summary|-S                    output 1 line per active phy; "
          "typically\n"
          "                                    equivalent to: '-o -d 1 -n 254 -b' .\n"
          "                                    This option is assumed if '--phy=ID' "
          "not given\n"
          "    --verbose|-v                    increase verbosity\n"
          "    --version|-V                    print version string and exit\n"
          "Performs a SMP DRIVE SLOT DATA LIST function\n"
          "If '--phy=ID' not given then '--summary' is assumed.\n"
          );
}

static void
dStrRaw(const char* str, int len)
{
    int k;

    for (k = 0 ; k < len; ++k)
        printf("%c", str[k]);
}


/* Returns 0 when successful, -1 for low level errors and > 0
   for other error categories. */
static int
do_drive_slot_data_list(struct smp_target_obj * top, int sphy_id,
                        unsigned char * resp, int max_resp_len, int *resp_len,
                        struct opts_t * op)
{
    unsigned char smp_req[] = {SMP_FRAME_TYPE_REQ, SMP_FN_DRIVE_SLOT_DATA_LIST, 0, 3,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    struct smp_req_resp smp_rr;
    char b[256];
    char * cp;
    int len, res, k, dword_resp_len, act_resplen;

    dword_resp_len = (max_resp_len - 8) / 4;
    smp_req[2] = (dword_resp_len < 0x100) ? dword_resp_len : 0xff;
    smp_req[4] = sphy_id;
    smp_req[5] = op->do_num;
    
    if (op->verbose) {
        fprintf(stderr, "    Drive Slot Data List request: ");
        for (k = 0; k < (int)sizeof(smp_req); ++k)
            fprintf(stderr, "%02x ", smp_req[k]);
        fprintf(stderr, "\n");
    }
    memset(&smp_rr, 0, sizeof(smp_rr));
    smp_rr.request_len = sizeof(smp_req);
    smp_rr.request = smp_req;
    smp_rr.max_response_len = max_resp_len;
    smp_rr.response = resp;

    res = smp_send_req(top, &smp_rr, op->verbose);
    if (res) {
        fprintf(stderr, "smp_send_req failed, res=%d\n", res);
        if (0 == op->verbose)
            fprintf(stderr, "    try adding '-v' option for more debug\n");
        return -1;
    }
    if (smp_rr.transport_err) {
        fprintf(stderr, "smp_send_req transport_error=%d\n",
                smp_rr.transport_err);
        return -1;
    }
    act_resplen = smp_rr.act_response_len;
    if ((act_resplen >= 0) && (act_resplen < 4)) {
        fprintf(stderr, "response too short, len=%d\n", act_resplen);
        return SMP_LIB_CAT_MALFORMED;
    }
    /*
    len = resp[3];
    if ((0 == len) && (0 == resp[2])) {
        len = smp_get_func_def_resp_len(resp[1]);
        if (len < 0) {
            len = 0;
            if (op->verbose > 0)
                fprintf(stderr, "unable to determine response length\n");
        }
    }
    len = 4 + (len * 4);        // length in bytes, excluding 4 byte CRC
    if ((act_resplen >= 0) && (len > act_resplen)) {
        if (op->verbose)
            fprintf(stderr, "actual response length [%d] less than "
                    "deduced length [%d]\n", act_resplen, len);
        len = act_resplen;
    }
    */
    len = act_resplen + 4;
    *resp_len = act_resplen;
    if (op->do_hex || op->do_raw) {
        if (op->do_hex)
            dStrHex((const char *)resp, len, 1);
        else
            dStrRaw((const char *)resp, len);
        if (SMP_FRAME_TYPE_RESP != resp[0])
            return SMP_LIB_CAT_MALFORMED;
        if (resp[1] != smp_req[1])
            return SMP_LIB_CAT_MALFORMED;
        if (resp[2]) {
            if (op->verbose)
                fprintf(stderr, "Drive Slot Data LIST result: %s\n",
                        smp_get_func_res_str(resp[2], sizeof(b), b));
            return resp[2];
        }
        return 0;
    }
    if (SMP_FRAME_TYPE_RESP != resp[0]) {
        fprintf(stderr, "expected SMP frame response type, got=0x%x\n",
                resp[0]);
        return SMP_LIB_CAT_MALFORMED;
    }
    if (resp[1] != smp_req[1]) {
        fprintf(stderr, "Expected function code=0x%x, got=0x%x\n",
                smp_req[1], resp[1]);
        return SMP_LIB_CAT_MALFORMED;
    }
    if (resp[2]) {
        if (SMP_FRES_NO_PHY != resp[2]) {
            cp = smp_get_func_res_str(resp[2], sizeof(b), b);
            fprintf(stderr, "Drive Slot Data LIST result: %s\n", cp);
        }
        return resp[2];
    }
    return 0;
}

int
main(int argc, char * argv[])
{
    int res, c, len, num_desc;
    int desc_len, k, j, no_more, off;
    long long sa_ll;
    char i_params[256];
    char device_name[512];
    unsigned char resp[1020 + 8];
    struct smp_target_obj tobj;
    int subvalue = 0;
    char * cp;
    int ret = 0;
    struct opts_t opts;

    memset(&opts, 0, sizeof(opts));
    memset(device_name, 0, sizeof device_name);
    while (1) {
        int option_index = 0;

        c = getopt_long(argc, argv, "hHI:n:op:rs:SvV",
                        long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 'h':
        case '?':
            usage();
            return 0;
        case 'H':
            ++opts.do_hex;
            break;
        case 'I':
            strncpy(i_params, optarg, sizeof(i_params));
            i_params[sizeof(i_params) - 1] = '\0';
            break;
        case 'n':
           opts.do_num = smp_get_num(optarg);
           if ((opts.do_num < 0) || (opts.do_num > 254)) {
                fprintf(stderr, "bad argument to '--num', expect value "
                        "from 0 to 254\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            ++opts.num_given;
            break;
        case 'o':
            ++opts.do_1line;
            break;
        case 'p':
           opts.phy_id = smp_get_num(optarg);
           if ((opts.phy_id < 0) || (opts.phy_id > 254)) {
                fprintf(stderr, "bad argument to '--phy', expect "
                        "value from 0 to 254\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            ++opts.phy_id_given;
            break;
        case 'r':
            ++opts.do_raw;
            break;
        case 's':
           sa_ll = smp_get_llnum(optarg);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad argument to '--sa'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            opts.sa = (unsigned long long)sa_ll;
            if (opts.sa > 0)
                ++opts.sa_given;
            break;
        case 'S':
            ++opts.do_summary;
            break;
        case 'v':
            ++opts.verbose;
            break;
        case 'V':
            fprintf(stderr, "version: %s\n", version_str);
            return 0;
        default:
            fprintf(stderr, "unrecognised switch code %c [0x%x] ??\n", c, c);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (optind < argc) {
        if ('\0' == device_name[0]) {
            strncpy(device_name, argv[optind], sizeof(device_name) - 1);
            device_name[sizeof(device_name) - 1] = '\0';
            ++optind;
        }
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == device_name[0]) {
        cp = getenv("SMP_UTILS_DEVICE");
        if (cp)
            strncpy(device_name, cp, sizeof(device_name) - 1);
        else {
            fprintf(stderr, "missing device name on command line\n    [Could "
                    "use environment variable SMP_UTILS_DEVICE instead]\n");
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if ((cp = strchr(device_name, SMP_SUBVALUE_SEPARATOR))) {
        *cp = '\0';
        if (1 != sscanf(cp + 1, "%d", &subvalue)) {
            fprintf(stderr, "expected number after separator in SMP_DEVICE "
                    "name\n");
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == opts.sa) {
        cp = getenv("SMP_UTILS_SAS_ADDR");
        if (cp) {
           sa_ll = smp_get_llnum(cp);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad value in environment variable "
                        "SMP_UTILS_SAS_ADDR\n");
                fprintf(stderr, "    use 0\n");
                sa_ll = 0;
            }
            opts.sa = (unsigned long long)sa_ll;
        }
    }
    if (opts.sa > 0) {
        if (! smp_is_naa5(opts.sa)) {
            fprintf(stderr, "SAS (target) address not in naa-5 format "
                    "(may need leading '0x')\n");
            if ('\0' == i_params[0]) {
                fprintf(stderr, "    use '--interface=' to override\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
        }
    }
    if ((0 == opts.do_summary) && (0 == opts.do_1line) &&
        (! opts.num_given) && (0 == opts.phy_id_given))
        ++opts.do_summary;
        
    if (opts.do_summary) {
        opts.do_1line = 1;
        opts.do_num = 254;
    } else if (! opts.num_given)
        opts.do_num = 1;
        

    res = smp_initiator_open(device_name, subvalue, i_params, opts.sa,
                             &tobj, opts.verbose);
    if (res < 0)
        return SMP_LIB_FILE_ERROR;

    no_more = 0;
    for (j = 0; (j < opts.do_num) && (! no_more); j += num_desc) {
        memset(resp, 0, sizeof(resp));
        if ((opts.phy_id + j) > 254) {
            ret = 0;    /* off the end so not error */
            break;
        }
        //ret = do_drive_slot_data_list(&tobj, opts.phy_id + j, resp, sizeof(resp),
        //                              &opts);
        ret = do_drive_slot_data_list(&tobj, opts.phy_id + j, resp, sizeof(resp), &len,
                                      &opts);

        if (ret) {
            if (SMP_FRES_NO_PHY == ret)
                ret = 0;    /* off the end so not error */
            break;
        }
        num_desc = resp[5];
        if (num_desc == 0)
            no_more = 1;
        if (opts.do_hex || opts.do_raw)
            continue;
        //len = (resp[3] * 4) + 4;    /* length in bytes excluding CRC field */
        if ((0 == j) && (! opts.do_1line))
        {
            if (! opts.do_1line) {
                printf("Drive Slot Data response header:\n");
                printf("  starting phy id: %d\n", resp[4]);
                printf("  number of discover list descriptors: %d\n", resp[5]);
                printf("  descriptor length: %d bytes\n", resp[6]);
            }
        }

        desc_len = resp[6];
        if (len != (12 + (num_desc * desc_len))) {
            fprintf(stderr, ">>> Response length of %d bytes doesn't match "
                    "%d descriptors, each\n  of %d bytes plus a 12 byte "
                    "header and 4 byte CRC\n", len + 4, num_desc, desc_len);
            if (len < (12 + (num_desc * desc_len))) {
                ret = SMP_LIB_CAT_MALFORMED;
                break;
            }
        }
        for (k = 0; k < num_desc; ++k) {
            off = 12 + (k * desc_len);
            if (opts.do_1line) {

                if ((j == 0) && (k == 0))
                {
                    printf("PHY\tDrivePwr\tDrivePwr\tDrivePwr\tDrivePwr\tPwrCtrl\t\tLocal\n");
                    printf("ID\tStatus\t\tCycleCnt\tOnCnt\t\tOffCnt\t\tinProgress\tPHYState\n");
                    printf("======================================================================================================\n");
                }
                printf("%3d\t", opts.phy_id + j + k);
                
                switch (resp[off])
                {
                    case 0x02:
                        printf("MISS\t");
                        break;
                    case 0x05:
                        printf("BUSY\t");
                        break;
                    case 0x23:
                        printf("OFF\t");
                        break;
                    case 0x2C:
                        printf("ON\t");
                        break;
                    default:
                        printf("UNKNOWN\t");
                        break;
                }
                printf("\t%d\t\t%d\t\t%d\t\t",
                       resp[off + 1], resp[off + 2], resp[off + 3]);
                if (resp[off + 8] & 0x1)
                    printf("YES\t\t");
                else
                    printf("ON\t\t");
                    
                printf("0x%02X\n", resp[off + 9]);
                
            } else {
                printf("descriptor %d:\n", j + k);
                printf("  phy identifier: %d\n", opts.phy_id + j + k);
                printf("  drive power status: 0x%X ", resp[off]);
                switch (resp[off])
                {
                    case 0x02:
                        printf("(MISS)\n");
                        break;
                    case 0x05:
                        printf("(BUSY)\n");
                        break;
                    case 0x23:
                        printf("(OFF)\n");
                        break;
                    case 0x2C:
                        printf("(ON)\n");
                        break;
                    default:
                        printf("(UNKNOWN)\n");
                        break;
                }
                printf("  drive power cycle count: %d\n", resp[off + 1]);
                printf("  drive power on count: %d\n", resp[off + 2]);
                printf("  drive power off count: %d\n", resp[off + 3]);
                printf("  power control in progress : ");
                if (resp[off + 8] & 0x1)
                    printf("YES\n");
                else
                    printf("ON\n");
                printf("  local phy state : 0x%02X ", resp[off + 9]);
                switch (resp[off + 9])
                {
                    case 0x00:
                        printf("(PHY is enabled and operational)\n");
                        break;
                    case 0x11:
                        printf("(PHY enabled – Unknown physical link rate)\n");
                        break;
                    case 0x12:
                        printf("(PHY enabled – No signal detected (Host In PHY only; cable present))\n");
                        break;
                    case 0x20:
                        printf("(PHY disabled – Manual from SES command)\n");
                        break;
                    case 0x21:
                        printf("(PHY disabled – Drive not physically present; SAS Cable not physically present)\n");
                        break;
                    case 0x22:
                        printf("(PHY disabled – Burst Loss of DWORD Synchronization Threshold exceeded)\n");
                        break;
                    case 0x23:
                        printf("(PHY disabled – Loss of DWORD Synchronization Threshold exceeded)\n");
                        break;
                    case 0x24:
                        printf("(PHY disabled – Burst Running Disparity Error Threshold exceeded)\n");
                        break;
                    case 0x25:
                        printf("(PHY disabled – Running Disparity Error Threshold exceeded)\n");
                        break;
                    case 0x26:
                        printf("(PHY disabled – Burst Invalid DWORD Threshold exceeded)\n");
                        break;
                    case 0x27:
                        printf("(PHY disabled – Invalid DWORD Threshold exceeded)\n");
                        break;
                    case 0x28:
                        printf("(PHY disabled – Burst PHY Reset Problem Threshold exceeded)\n");
                        break;
                    case 0x29:
                        printf("(PHY disabled – PHY Reset Problem Threshold exceeded)\n");
                        break;
                    case 0x2a:
                        printf("(PHY disabled – Burst PHY CHANGE Threshold Exceeded)\n");
                        break;
                    case 0x2b:
                        printf("(PHY disabled – PHY CHANGE Threshold Exceeded)\n");
                        break;
                    case 0x2c:
                        printf("(Unused)\n");
                        break;
                    case 0x2d:
                        printf("(PHY disabled – Mirrored disable from partner module)\n");
                        break;
                    case 0x2e:
                        printf("(PHY disabled – Burst CRC error Threshold exceeded)\n");
                        break;
                    case 0x2f:
                        printf("(PHY disabled – CRC error Threshold exceeded)\n");
                        break;
                    case 0x40:
                        printf("(PHY disabled – SMP PHY_CONTROL(DISABLE))\n");
                        break;
                    case 0x50:
                        printf("(PHY disabled – Expander Reference Clock Error)\n");
                        break;
                    case 0x60:
                        printf("(PHY disabled – Phy not used)\n");
                        break;
                    case 0xff:
                        printf("(PHY state unknown -- Initializing)\n");
                        break;
                    default:
                        printf("(unknown state)\n");
                        break;
                }
            }
        }
    }

err_out:
    res = smp_initiator_close(&tobj);
    if (res < 0) {
        if (0 == ret)
            return SMP_LIB_FILE_ERROR;
    }
    if (ret < 0)
        ret = SMP_LIB_CAT_OTHER;
    if (opts.verbose && ret)
        fprintf(stderr, "Exit status %d indicates error detected\n", ret);
    return ret;
}
