#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "smp_lib.h"

/* This is a Serial Attached SCSI (SAS) Serial Management Protocol (SMP)
 * utility.
 *
 * This utility issues a REPORT PHY ERROR LOG LIST function and outputs its
 * response.
 */

static const char * version_str = "0.01 20160505";

#define SMP_FN_REPORT_PHY_ERR_LOG_LIST              0xCA

static struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {"hex", 0, 0, 'H'},
    {"interface", 1, 0, 'I'},
    {"num", 1, 0, 'n'},
    {"one", 0, 0, 'o'},
    {"phy", 1, 0, 'p'},
    {"raw", 0, 0, 'r'},
    {"sa", 1, 0, 's'},
    {"summary", 0, 0, 'S'},
    {"verbose", 0, 0, 'v'},
    {"version", 0, 0, 'V'},
    {0, 0, 0, 0},
};

struct opts_t {
    int do_hex;
    int do_num;
    int num_given;
    int do_1line;
    int phy_id;
    int phy_id_given;
    int do_raw;
    int do_summary;
    int verbose;
    int sa_given;
    unsigned long long sa;
};

static void
usage(void)
{
    fprintf(stderr, "Usage: "
          "smp_rep_phy_err_log_list [--help] [--hex]\n"
          "                                [--interface=PARAMS]\n"
          "                                [--num=NUM] [--one] [--phy=ID]\n"
          "                                [--raw] [--sa=SAS_ADDR]\n"
          "                                [--summary] [--verbose]\n"
          "                                [--version] SMP_DEVICE[,N]\n"
          "  where:\n"
          "    --help|-h                       print out usage message\n"
          "    --hex|-H                        print response in hexadecimal\n"
          "    --interface=PARAMS|-I PARAMS    specify or override "
          "interface\n"
          "    --num=NUM|-n NUM                maximum number of descriptors to fetch "
          "(def: 1)\n"
          "    --one|-o                        one line output per response "
          "descriptor (phy)\n"
          "    --phy=ID|-p ID                  phy identifier [or starting phy id]\n"
          "    --raw|-r                        output response in binary\n"
          "    --sa=SAS_ADDR|-s SAS_ADDR       SAS address of SMP "
          "target (use leading\n"
          "                                    '0x' or trailing 'h'). "
          "Depending on\n"
          "                                    the interface, may not be "
          "needed\n"
          "    --summary|-S                    output 1 line per active phy; "
          "typically\n"
          "                                    equivalent to: '-o -n 254' .\n"
          "                                    This option is assumed if '--phy=ID' "
          "not given\n"
          "    --verbose|-v                    increase verbosity\n"
          "    --version|-V                    print version string and exit\n"
          "Performs a SMP REPORT PHY ERROR LOG LIST function\n"
          "If '--phy=ID' not given then '--summary' is assumed.\n"
          );
}

static void
dStrRaw(const char* str, int len)
{
    int k;

    for (k = 0 ; k < len; ++k)
        printf("%c", str[k]);
}


/* Returns 0 when successful, -1 for low level errors and > 0
   for other error categories. */
static int
do_rep_phy_err_log_list(struct smp_target_obj * top, int sphy_id,
                        unsigned char * resp, int max_resp_len, int *resp_len,
                        struct opts_t * op)
{
    unsigned char smp_req[] = {SMP_FRAME_TYPE_REQ, SMP_FN_REPORT_PHY_ERR_LOG_LIST, 0, 3,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    struct smp_req_resp smp_rr;
    char b[256];
    char * cp;
    int len, res, k, dword_resp_len, act_resplen;

    dword_resp_len = (max_resp_len - 8) / 4;
    smp_req[2] = (dword_resp_len < 0x100) ? dword_resp_len : 0xff;
    smp_req[4] = sphy_id;
    smp_req[5] = op->do_num;
    
    if (op->verbose) {
        fprintf(stderr, "    Report PHY Error Log List request: ");
        for (k = 0; k < (int)sizeof(smp_req); ++k)
            fprintf(stderr, "%02x ", smp_req[k]);
        fprintf(stderr, "\n");
    }
    memset(&smp_rr, 0, sizeof(smp_rr));
    smp_rr.request_len = sizeof(smp_req);
    smp_rr.request = smp_req;
    smp_rr.max_response_len = max_resp_len;
    smp_rr.response = resp;

    res = smp_send_req(top, &smp_rr, op->verbose);
    if (res) {
        fprintf(stderr, "smp_send_req failed, res=%d\n", res);
        if (0 == op->verbose)
            fprintf(stderr, "    try adding '-v' option for more debug\n");
        return -1;
    }
    if (smp_rr.transport_err) {
        fprintf(stderr, "smp_send_req transport_error=%d\n",
                smp_rr.transport_err);
        return -1;
    }
    act_resplen = smp_rr.act_response_len;
    if ((act_resplen >= 0) && (act_resplen < 4)) {
        fprintf(stderr, "response too short, len=%d\n", act_resplen);
        return SMP_LIB_CAT_MALFORMED;
    }
    /*
    len = resp[3];
    if ((0 == len) && (0 == resp[2])) {
        len = smp_get_func_def_resp_len(resp[1]);
        if (len < 0) {
            len = 0;
            if (op->verbose > 0)
                fprintf(stderr, "unable to determine response length\n");
        }
    }
    len = 4 + (len * 4);        // length in bytes, excluding 4 byte CRC
    if ((act_resplen >= 0) && (len > act_resplen)) {
        if (op->verbose)
            fprintf(stderr, "actual response length [%d] less than "
                    "deduced length [%d]\n", act_resplen, len);
        len = act_resplen;
    }
    */
    len = act_resplen + 4;
    *resp_len = act_resplen;
    if (op->do_hex || op->do_raw) {
        if (op->do_hex)
            dStrHex((const char *)resp, len, 1);
        else
            dStrRaw((const char *)resp, len);
        if (SMP_FRAME_TYPE_RESP != resp[0])
            return SMP_LIB_CAT_MALFORMED;
        if (resp[1] != smp_req[1])
            return SMP_LIB_CAT_MALFORMED;
        if (resp[2]) {
            if (op->verbose)
                fprintf(stderr, "Report PHY Error Log List result: %s\n",
                        smp_get_func_res_str(resp[2], sizeof(b), b));
            return resp[2];
        }
        return 0;
    }
    if (SMP_FRAME_TYPE_RESP != resp[0]) {
        fprintf(stderr, "expected SMP frame response type, got=0x%x\n",
                resp[0]);
        return SMP_LIB_CAT_MALFORMED;
    }
    if (resp[1] != smp_req[1]) {
        fprintf(stderr, "Expected function code=0x%x, got=0x%x\n",
                smp_req[1], resp[1]);
        return SMP_LIB_CAT_MALFORMED;
    }
    if (resp[2]) {
        if (SMP_FRES_NO_PHY != resp[2]) {
            cp = smp_get_func_res_str(resp[2], sizeof(b), b);
            fprintf(stderr, "Report PHY Error Log List result: %s\n", cp);
        }
        return resp[2];
    }
    return 0;
}

int
main(int argc, char * argv[])
{
    int res, c, len, num_desc;
    int desc_len, k, j, no_more, off;
    long long sa_ll;
    char i_params[256];
    char device_name[512];
    unsigned char resp[1020 + 8];
    struct smp_target_obj tobj;
    int subvalue = 0;
    char * cp;
    int ret = 0;
    struct opts_t opts;

    memset(&opts, 0, sizeof(opts));
    memset(device_name, 0, sizeof device_name);
    while (1) {
        int option_index = 0;

        c = getopt_long(argc, argv, "hHI:n:op:rs:SvV",
                        long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 'h':
        case '?':
            usage();
            return 0;
        case 'H':
            ++opts.do_hex;
            break;
        case 'I':
            strncpy(i_params, optarg, sizeof(i_params));
            i_params[sizeof(i_params) - 1] = '\0';
            break;
        case 'n':
           opts.do_num = smp_get_num(optarg);
           if ((opts.do_num < 0) || (opts.do_num > 254)) {
                fprintf(stderr, "bad argument to '--num', expect value "
                        "from 0 to 254\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            ++opts.num_given;
            break;
        case 'o':
            ++opts.do_1line;
            break;
        case 'p':
           opts.phy_id = smp_get_num(optarg);
           if ((opts.phy_id < 0) || (opts.phy_id > 254)) {
                fprintf(stderr, "bad argument to '--phy', expect "
                        "value from 0 to 254\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            ++opts.phy_id_given;
            break;
        case 'r':
            ++opts.do_raw;
            break;
        case 's':
           sa_ll = smp_get_llnum(optarg);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad argument to '--sa'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            opts.sa = (unsigned long long)sa_ll;
            if (opts.sa > 0)
                ++opts.sa_given;
            break;
        case 'S':
            ++opts.do_summary;
            break;
        case 'v':
            ++opts.verbose;
            break;
        case 'V':
            fprintf(stderr, "version: %s\n", version_str);
            return 0;
        default:
            fprintf(stderr, "unrecognised switch code %c [0x%x] ??\n", c, c);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (optind < argc) {
        if ('\0' == device_name[0]) {
            strncpy(device_name, argv[optind], sizeof(device_name) - 1);
            device_name[sizeof(device_name) - 1] = '\0';
            ++optind;
        }
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == device_name[0]) {
        cp = getenv("SMP_UTILS_DEVICE");
        if (cp)
            strncpy(device_name, cp, sizeof(device_name) - 1);
        else {
            fprintf(stderr, "missing device name on command line\n    [Could "
                    "use environment variable SMP_UTILS_DEVICE instead]\n");
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if ((cp = strchr(device_name, SMP_SUBVALUE_SEPARATOR))) {
        *cp = '\0';
        if (1 != sscanf(cp + 1, "%d", &subvalue)) {
            fprintf(stderr, "expected number after separator in SMP_DEVICE "
                    "name\n");
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == opts.sa) {
        cp = getenv("SMP_UTILS_SAS_ADDR");
        if (cp) {
           sa_ll = smp_get_llnum(cp);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad value in environment variable "
                        "SMP_UTILS_SAS_ADDR\n");
                fprintf(stderr, "    use 0\n");
                sa_ll = 0;
            }
            opts.sa = (unsigned long long)sa_ll;
        }
    }
    if (opts.sa > 0) {
        if (! smp_is_naa5(opts.sa)) {
            fprintf(stderr, "SAS (target) address not in naa-5 format "
                    "(may need leading '0x')\n");
            if ('\0' == i_params[0]) {
                fprintf(stderr, "    use '--interface=' to override\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
        }
    }
    if ((0 == opts.do_summary) && (0 == opts.do_1line) &&
        (! opts.num_given) && (0 == opts.phy_id_given))
        ++opts.do_summary;
        
    if (opts.do_summary) {
        opts.do_1line = 1;
        opts.do_num = 254;
    } else if (! opts.num_given)
        opts.do_num = 1;
        

    res = smp_initiator_open(device_name, subvalue, i_params, opts.sa,
                             &tobj, opts.verbose);
    if (res < 0)
        return SMP_LIB_FILE_ERROR;

    no_more = 0;
    for (j = 0; (j < opts.do_num) && (! no_more); j += num_desc) {
        memset(resp, 0, sizeof(resp));
        if ((opts.phy_id + j) > 254) {
            ret = 0;    /* off the end so not error */
            break;
        }
        ret = do_rep_phy_err_log_list(&tobj, opts.phy_id + j, resp, sizeof(resp), &len,
                                      &opts);
        if (ret) {
            if (SMP_FRES_NO_PHY == ret)
                ret = 0;    /* off the end so not error */
            break;
        }
        num_desc = resp[5];
        if (num_desc == 0)
            no_more = 1;
        if (opts.do_hex || opts.do_raw)
            continue;
        //len = (resp[3] * 4) + 4;    /* length in bytes excluding CRC field */
        if ((0 == j) && (! opts.do_1line))
        {
            if (! opts.do_1line) {
                printf("Report PHY Error Log List response header:\n");
                printf("  starting phy id: %d\n", resp[4]);
                printf("  number of discover list descriptors: %d\n", resp[5]);
                printf("  descriptor length: %d bytes\n", resp[6]);
            }
        }

        desc_len = resp[6];
        if (len != (12 + (num_desc * desc_len))) {
            fprintf(stderr, ">>> Response length of %d bytes doesn't match "
                    "%d descriptors, each\n  of %d bytes plus a 12 byte "
                    "header and 4 byte CRC\n", len + 4, num_desc, desc_len);
            if (len < (12 + (num_desc * desc_len))) {
                ret = SMP_LIB_CAT_MALFORMED;
                break;
            }
        }
        for (k = 0; k < num_desc; ++k) {
            off = 12 + (k * desc_len);
            if (opts.do_1line) {
            
                if ((j == 0) && (k == 0))
                {
                    printf("PHY\n");
                    printf("ID\tInvWrdCnt\tDispErrCnt\tLossSyncCnt\tRstSeqFailCnt\tCRCErrCnt\tPHYChangeCnt\n");
                    printf("=====================================================================================================\n");
                }
                
                printf("%2d\t", opts.phy_id + j + k);
                printf("%d\t\t",
                       (resp[off] << 24) + (resp[off + 1] << 16) +
                       (resp[off + 2] << 8) + resp[off + 3]);
                printf("%d\t\t",
                       (resp[off + 4] << 24) + (resp[off + 5] << 16) +
                       (resp[off + 6] << 8) + resp[off + 7]);
                printf("%d\t\t",
                       (resp[off + 8] << 24) + (resp[off + 9] << 16) +
                       (resp[off + 10] << 8) + resp[off + 11]);
                printf("%d\t\t",
                       (resp[off + 12] << 24) + (resp[off + 13] << 16) +
                       (resp[off + 14] << 8) + resp[off + 15]);
                printf("%d\t\t",
                       (resp[off + 16] << 24) + (resp[off + 17] << 16) +
                       (resp[off + 18] << 8) + resp[off + 19]); 
                printf("%d\n",
                       (resp[off + 20] << 24) + (resp[off + 21] << 16) +
                       (resp[off + 22] << 8) + resp[off + 23]);
            } else {
                printf("descriptor %d:\n", j + k);
                printf("  phy identifier: %d\n", opts.phy_id + j + k);
                printf("  invalid dword count: %d\n",
                       (resp[off] << 24) + (resp[off + 1] << 16) +
                       (resp[off + 2] << 8) + resp[off + 3]);
                printf("  running disparity error count: %d\n",
                       (resp[off + 4] << 24) + (resp[off + 5] << 16) +
                       (resp[off + 6] << 8) + resp[off + 7]);
                printf("  loss of dword synchronization count: %d\n",
                       (resp[off + 8] << 24) + (resp[off + 9] << 16) +
                       (resp[off + 10] << 8) + resp[off + 11]);
                printf("  phy reset problem count: %d\n",
                       (resp[off + 12] << 24) + (resp[off + 13] << 16) +
                       (resp[off + 14] << 8) + resp[off + 15]);
                printf("  CRC error count: %d\n",
                       (resp[off + 16] << 24) + (resp[off + 17] << 16) +
                       (resp[off + 18] << 8) + resp[off + 19]); 
                printf("  phy change count: %d\n",
                       (resp[off + 20] << 24) + (resp[off + 21] << 16) +
                       (resp[off + 22] << 8) + resp[off + 23]);

            }
        }
    }

err_out:
    res = smp_initiator_close(&tobj);
    if (res < 0) {
        if (0 == ret)
            return SMP_LIB_FILE_ERROR;
    }
    if (ret < 0)
        ret = SMP_LIB_CAT_OTHER;
    if (opts.verbose && ret)
        fprintf(stderr, "Exit status %d indicates error detected\n", ret);
    return ret;
}
