#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "smp_lib.h"

#define SMP_FN_RESET_EXPANDER      0xC2

/* This is a Serial Attached SCSI (SAS) Serial Management Protocol (SMP)
 * utility.
 *
 * This utility issues a SET ENCLOSURE ID function and outputs its response.
 */

static const char * version_str = "0.01 20160607";

static struct option long_options[] = {
    {"address", required_argument, 0, 'a'},
    {"data", required_argument, 0, 'd'},
    {"help", no_argument, 0, 'h'},
    {"hex", no_argument, 0, 'H'},
    {"interface", required_argument, 0, 'I'},
    {"raw", no_argument, 0, 'r'},
    {"sa", required_argument, 0, 's'},
    {"verbose", no_argument, 0, 'v'},
    {"version", no_argument, 0, 'V'},
    {0, 0, 0, 0},
};


static void
usage(void)
{
    fprintf(stderr, "Usage: "
          "smp_reset_expander [--address=ADDR] [--data=DATA] [--help]\n"
          "                          [--hex] [--interface=PARAMS]\n"
          "                          [--raw] [--sa=SAS_ADDR] [--version]\n"
          "                          [--verbose] SMP_DEVICE[,N]\n"
          "  where:\n"
          "    --address=ADDR|-a ADDR          address\n"
          "    --data=DATA|-d DATA             write data\n"
          "    --help|-h                       print out usage message\n"
          "    --hex|-H                        print response in hexadecimal\n"
          "    --interface=PARAMS|-I PARAMS    specify or override "
          "interface\n"
          "    --raw|-r                        output response in binary\n"
          "    --sa=SAS_ADDR|-s SAS_ADDR       SAS address of SMP "
          "target (use leading\n"
          "                                    '0x' or trailing 'h'). "
          "Depending on\n"
          "                                    the interface, may not be "
          "needed\n"
          "    --verbose|-v                    increase verbosity\n"
          "    --version|-V                    print version string and exit\n\n"
          "Performs a SMP SET ENCLOSURE ID function.\n"
          );
}

int
main(int argc, char * argv[])
{
    int res, c, k;
    int do_hex = 0;
    int do_raw = 0;
    int verbose = 0;
    unsigned int address = 0;
    unsigned int data = 0;
    long long sa_ll;
    unsigned long long sa = 0;
    char i_params[256];
    char device_name[512];
    unsigned char smp_req[] = {SMP_FRAME_TYPE_REQ, SMP_FN_RESET_EXPANDER,
        0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    unsigned char smp_resp[8];
    struct smp_req_resp smp_rr;
    struct smp_target_obj tobj;
    int subvalue = 0;
    char * cp;
    int ret = 0;
    bool haveAddr = 0;
    bool haveData = 0;

    memset(device_name, 0, sizeof device_name);
    while (1) {
        int option_index = 0;

        c = getopt_long(argc, argv, "a:d:hHI:rs:vV",
                        long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 'a':
           address = smp_get_num(optarg);
           if (address > 0xFFFFFFFF) {
                fprintf(stderr, "bad argument to '--address'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            haveAddr = 1;
            break;
        case 'd':
            data = smp_get_num(optarg);
            if (data > 0xFFFFFFFF) {
                fprintf(stderr, "bad argument to '--data'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            haveData = 1;
            break;
        case 'h':
        case '?':
            usage();
            return 0;
        case 'H':
            ++do_hex;
            break;
        case 'I':
            strncpy(i_params, optarg, sizeof(i_params));
            i_params[sizeof(i_params) - 1] = '\0';
            break;
        case 'r':
            ++do_raw;
            break;
        case 's':
           sa_ll = smp_get_llnum(optarg);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad argument to '--sa'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            sa = (unsigned long long)sa_ll;
            break;
        case 'v':
            ++verbose;
            break;
        case 'V':
            fprintf(stderr, "version: %s\n", version_str);
            return 0;
        default:
            fprintf(stderr, "unrecognised switch code 0x%x ??\n", c);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (optind < argc) {
        if ('\0' == device_name[0]) {
            strncpy(device_name, argv[optind], sizeof(device_name) - 1);
            device_name[sizeof(device_name) - 1] = '\0';
            ++optind;
        }
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == device_name[0]) {
        cp = getenv("SMP_UTILS_DEVICE");
        if (cp)
            strncpy(device_name, cp, sizeof(device_name) - 1);
        else {
            fprintf(stderr, "missing device name on command line\n    [Could "
                    "use environment variable SMP_UTILS_DEVICE instead]\n");
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if ((cp = strchr(device_name, SMP_SUBVALUE_SEPARATOR))) {
        *cp = '\0';
        if (1 != sscanf(cp + 1, "%d", &subvalue)) {
            fprintf(stderr, "expected number after separator in SMP_DEVICE "
                    "name\n");
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == sa) {
        cp = getenv("SMP_UTILS_SAS_ADDR");
        if (cp) {
           sa_ll = smp_get_llnum(cp);
           if (-1LL == sa_ll) {
                fprintf(stderr, "bad value in environment variable "
                        "SMP_UTILS_SAS_ADDR\n");
                fprintf(stderr, "    use 0\n");
                sa_ll = 0;
            }
            sa = (unsigned long long)sa_ll;
        }
    }
    if (sa > 0) {
        if (! smp_is_naa5(sa)) {
            fprintf(stderr, "SAS (target) address not in naa-5 format "
                    "(may need leading '0x')\n");
            if ('\0' == i_params[0]) {
                fprintf(stderr, "    use '--interface=' to override\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
        }
    }
    
    if (haveAddr == 0) {
        fprintf(stderr, "    use '--address='\n");
        return SMP_LIB_SYNTAX_ERROR;
    }
    
    if (haveData == 0) {
        fprintf(stderr, "    use '--data='\n");
        return SMP_LIB_SYNTAX_ERROR;
    }

    res = smp_initiator_open(device_name, subvalue, i_params, sa,
                             &tobj, verbose);
    if (res < 0)
        return SMP_LIB_FILE_ERROR;

    smp_req[4] = (address & 0xFF000000) >> 24;
    smp_req[5] = (address & 0x00FF0000) >> 16;
    smp_req[6] = (address & 0x0000FF00) >> 8;
    smp_req[7] = address & 0xFF;
    
    smp_req[8] = (data & 0xFF000000) >> 24;
    smp_req[9] = (data & 0x00FF0000) >> 16;
    smp_req[10] = (data & 0x0000FF00) >> 8;
    smp_req[11] = data & 0xFF;
    
    if (verbose) {
        fprintf(stderr, "    Reset expander request: ");
        for (k = 0; k < (int)sizeof(smp_req); ++k)
            fprintf(stderr, "%02x ", smp_req[k]);
        fprintf(stderr, "\n");
    }

    memset(&smp_rr, 0, sizeof(smp_rr));
    smp_rr.request_len = sizeof(smp_req);
    smp_rr.request = smp_req;
    smp_rr.max_response_len = sizeof(smp_resp);
    smp_rr.response = smp_resp;
    smp_send_req(&tobj, &smp_rr, verbose);
    
    res = smp_initiator_close(&tobj);
    if (res < 0) {
        fprintf(stderr, "close error: %s\n", safe_strerror(errno));
        if (0 == ret)
            return SMP_LIB_FILE_ERROR;
    }
    if (ret < 0)
        ret = SMP_LIB_CAT_OTHER;
    if (verbose && ret)
        fprintf(stderr, "Exit status %d indicates error detected\n", ret);
    return ret;
}
